/*
 * V1.C
 *
 */

#define KEYDEFS 1       /* include key defs from v.h */
#include "v.h"

/*
 * GETKEY
 *
 * Purpose:
 *              Accept regular or function keys from the console.
 * Returns:
 *              Coded 16-bit integer for each single keystroke.
 *              Each distinct keystroke must be encoded to
 *              a unique integer in the range 0...65535.
 * Notes:
 *              o Ideas for most termcap terminals.
 *              o One character is returned if it is not a leadin
 *                character from one of the strings in ourkeys[].
 *              o No timer event flags are implemented.
 *              o strncmp(), strcmp() must be specially written
 *                to handle unsigned chars.
 */
static char safekeybuf[MKLEN+2];
char *getkeybuf=safekeybuf;
int getkeynchars;
int getkey() {
int lo,hi,i,j,c,flg;
int getbyte(),strncmp(),strcmp();
LONG ccBIOS();

  /* Repetition count on last key */
  if(repeats>1){--repeats; if(!ccBIOS(2)) return repkey; repeats=1;}
  c = getbyte();
#if MSDOS
/* 128 is function key leadin on a PC */
  if(c != 128 && ' ' <= c && c < 256) return c;
#else
/*  if(' ' <= c && c < 256) return c; */
#endif
  getkeynchars = 0;
  while(TRUE) {
    getkeybuf[getkeynchars] = c;
    getkeybuf[++getkeynchars] = EOS;
    if(getkeynchars < MKLEN){
      flg = FALSE;
      hi = MAXKEYS-1; lo = 0;
      while(lo <= hi){    /* requires the keys to be sorted */
        i = (lo+hi)/2;
        if((j=strncmp(ourkeys[i].seq,getkeybuf,getkeynchars)) == 0){
          if(strlen(ourkeys[i].seq)==getkeynchars) return (ourkeys[i].value);
          flg=TRUE; break;
        }
        if(j < 0) lo = i+1; else hi = i-1;
      }
      if(flg) {c=getbyte(); continue;}
      if(getkeynchars == 1) return c;
    }
    return NULLKEY;
  }
}

#if !MSDOS              /* MSDOS is in assembler elsewhere */
strncmp(s,t,n) char *s,*t; int n; {
  while(n-- && s[0]==t[0]){ if(!s[0] || !n) return 0; ++s; ++t;}
  return ((unsigned char)s[0]-(unsigned char)t[0]);
}

strcmp(s,t) char *s,*t; {
  while(s[0] && s[0]==t[0]){ ++s; ++t;}
  return ((unsigned char)s[0]-(unsigned char)t[0]);
}
#endif

sorttable(){
int n,gap,i,ig,j,t; char *tmp;
  n=MAXKEYS;
  for (gap = n/2 ; gap > 0 ; gap /= 2)
    for (j = gap ; j < n; j++)
      for (i = j-gap ; i >= 0 ; i -= gap) {
              ig = i + gap;
              if (strcmp(ourkeys[i].seq,ourkeys[ig].seq) <= 0)
                      break;
              tmp=ourkeys[i].seq; ourkeys[i].seq=ourkeys[ig].seq;
              ourkeys[ig].seq=tmp;
              t=ourkeys[i].value; ourkeys[i].value=ourkeys[ig].value;
              ourkeys[ig].value=t;
      }
  return 0;
}

/*
 * GGETSTR
 *
 * Purpose:
 *              Read in a line of text, terminate on RETURN
 *              or a special key. Delete key edits
 *              the input. TAB completes filename.
 *              Search keys recall the last search string.
 *              The string is NULL-terminated.
 * Returns:
 *              The integer value of the break character.
 */
int ggetstr(msg,p)
char *msg;  /* expect msg to be null terminated */
char *p;    /* expect p to be null terminated */
{
char buf[MAXFNAME];
int c,n;
int j;
char *q,*s,*r;
extern char *strrchr(),*toprint(),*strchr();
extern char *sob(),*fnb();
unsigned char *curchar();
extern int getkey(),getbyte();
extern LLONG atoiLL();
extern LLONG getatoiLL();
extern char *lltoa();

  q = (char *)curchar();
top:
  bbanner(msg,p);
  n = strlen(p);
  while(n<MAXLN) {
    if((c = getkey()) == CONTROLENTRY) c = getbyte();
    else {
      switch(c){
#if MSDOS && DOS386I
      case MOUSEPRESS: /* Mouse event */
               pressmouse(); goto killchars;
#endif
#if BSDunix || SYS5
      case MOUSEPRESS: /* Mouse click event. Kill trailing bytes. */
           getbyte(); getbyte(); getbyte(); goto killchars;
#endif
      case POPKEY:   /* Recopy last argument */
                     strcpy(p,lastarg);
                     goto killchars;
      case TABKEY:
        if((s=strrchr(p,' ')) != (char *)0) ++s; else s=p;
        strcpy(buf,s);
        if((j=dir(buf,-1,0,(char *)0,(char **)0)) >= 1) { /* complete file name */
          stripstars(buf);      /* Remove extra * on end */
          strcpy(s,buf);
          goto killchars;
        } else bell();
        goto loop;
      case EXTRIGHTARROW:
      case GRABTEXT:            /* Grab token from cursor position */
                     j=fnb(sob(q))-q;
                     if(j+n<MAXLN){
                       strnzcpy(p+n,q,j); q += j;
                       goto printend;
                     }
                     goto loop;
      case QPOPKEY:
        j=min(strlen(optbuf),MAXLN-n-1);
        strnzcpy(p+n,optbuf,j); goto killchars;
printend:
        ePrint(p+n); n=strlen(p); goto loop;
#if MATHFLOAT
      case OPTMATHFLOAT:
        convert2double(p);
        goto killchars;
      case OPTEVALEXPR:
        calculator(p);
        goto killchars;
#endif
      case OPTHEX:
        hexstrToBase10str(p);
        goto killchars;
      case OPTOCT:
        octstrToBase10str(p);
        goto killchars;
      case OPTMATH:
        convert2int(p);
        goto killchars;
      case RIGHTARROW: c = (q[0]? *q++ : ' '); break;
      case FWDSEARCH: /*  Search forward */
      case REVSEARCH: /*  Search reverse */
      case LFWDSEARCH: /*  Line Search forward */
      case LREVSEARCH: /*  Line Search reverse */
                     if(n==0){strcpy(p,srcharg); goto killchars;}
                     break;
      case DOWNARROW: getsrcharg(p,-1); goto killchars;
      case UPARROW:   getsrcharg(p,1); goto killchars;
      case REVTABKEY: /* Reverse tab key mungs prev file name to unix std */
                     if(n){
                       if((r=strrchr(p+n-1,' '))==(char *)0) r=p;
                       while((r=strchr(r,'\\'))!=(char *)0) r[0]='/';
                       goto killchars;
                     }
                     break;
      case DSKFILES: /* Insert file name */
                     if(1+n+strlen(sysfile)<MAXLN){
                       strcpy(p+n,sysfile);
                       goto printend;
                     }
                     goto loop;
      case REPEATKEY: /* Multiply argument by 4*/
                     if((j=(int)atoiLL(p))==0) j=1;
                     lltoa((LLONG)(4*j),p);
killchars:
                     goto top;
      case 8:case 127:case LEFTARROW:
                     if(n) { erachar(); p[--n]=EOS; } else bell();
                     goto loop;
      case ZEROTOKEN:   /* Insert begin or end of line token */
                     c=zero; break;
      }/* end switch */
      if(c > 1000) {p[n] = EOS; break;}
    }
    p[n++] = c;
    p[n] = EOS;
    ePrint(&p[n-1]);
loop: ;
  }/* End while() */
end:
  p[n] = EOS;
  if(p[0]) strcpy(lastarg,p);    /* Save whatever was typed in */
  return c;
}

/*
 * ENTKEY
 *
 * Purpose:
 *              Process keys after enter key.
 * Returns:
 *              Coded break key and system variables.
 * Notes:
 *              Independent of key definitions.
 */
int entkey() {
int c;
extern LLONG getatoiLL();

  c=getquery("ENTER: ");
  sysarg=(int)getatoiLL();    /* Collect numeric argument, if any */
  lcount();
  if(c==ENTERKEY && sysv[0]==0) return MENUKEY;
  if(c==RETURNKEY) return NULLKEY;
  return c;
}

/*
 * GETBYT
 *
 * Purpose:
 *              1. Get raw character from keyboard, no echo.
 *              2. If macrun <> 0, then get the char from macbuf[].
 *              3. If record <> 0, then add the char to macbuf[].
 * Returns:
 *              Integer c in the range 0...127, usually. The
 *              keyboard is expected to transmit 0...255.
 * Notes:
 *              o ccBIOS(3) is a library direct-bios call
 *                that reads the console, no-echo; ccBIOS(2) checks
 *                console status.
 *              o Any keyboard hit stops a running macro.
 *
 */
int getbyte() {
int c;
LONG ccBIOS();

  while(TRUE) {
    if(macrun) {
      if(macind>=macmax) {
        if(macind==0) statblank=TRUE;
        macind = 0;
        if(ccBIOS(2)) goto quit;
        if(macrep>1) --macrep;
        else goto quit;
      }
      else
      return (macbuf[macind++]);
      continue;
    }
    c = ccBIOS(3);
    if(c==0) c=128;     /* normally can't use null in a string */
    if(record) {
      if(macmax>=MAXMB) { bell(); record = 0;}
      else {
        macbuf[macmax++] = c;
        macbuf[macmax]=0;  /* Terminate the string */
      }
    }
    return c;
  }
quit: macabort(); return -1;
}

/*
 * MACABO
 *
 * Purpose:
 *              Abort a running macro sequence.
 */
void macabort() {
  record = macind = macrun = 0; macrep = 1;
}

/*
 * TESTDR
 *
 * Purpose:
 *              Test to see if we need to redraw the screen
 *              after a silent macro.
 * Notes:
 *
 */
void testdraw() {
extern int nodisp,halfdone;
  if(halfdone || nodisp) {
    if(!macrun) {
      nodisp = 0;
      drawall();
    }
  }
}

int getquery(s) char *s; {
  sysv[0]=0; return ggetstr(s,sysv);
}

int getnumquery(s,x,low,high) char *s; int *x; int low,high;{
int y; char v[128]; int c; char *endof();
extern LLONG getatoiLL();
  ssprintf(v,"%s [%u]: ",s,x[0]);
  c=getquery(v);
  if(sysv[0]){
    y=(int)getatoiLL();
    if(high<low) high=y;
    if(low<=y && y<=high) x[0]=y;
  }
  return c;
}

int getupper()
{
int toupper();
  return (toupper(getbyte()));
}

int getyesno(s) char *s; {
  bbanner(s," [y/n]? [N]: ");
  return (getupper() == 'Y');
}

int secondkill(){
extern unsigned lastkey;
  switch(lastkey){
  case WORDERASE:
  case ERASEKEY:
  case ERASELINE:
  case DELLINE:
  return 1;
  }
  return 0;
}

int secondpush(){
extern unsigned lastkey;
 return (lastkey==PUSHKEY);
}

struct optinfo {char *descrip; int *val;};
/* negative cursor index means display in col 30 */
/* line, string, data */
struct optinfo opt[]={
{PiVersion,      0},
{AUTHOR,         0},
{ADDRESSofAUTHOR,0},
{"",             0},
{"Max files    ",&maxinfo},
{"Active Files ",&sysarg},
{"Max file name",&MAXFN},
{"Blocks in use",&USEDBLOCK},
{"Blocks total ",&MAXBLOCK},
{"Block size   ",&SIZEBLOCK},
{"Max columns  ",&MAXLN},
{"Max rows     ",&MAXROWS},
{"Columns 0 to ",&COLMAX},
{"Rows 0 to    ",&ROWMAX},
{"",0},
};
#define SIZEOPT sizeof(opt)/sizeof(opt[0])

/*
 * OPTIONS
 *
 * Prints option toggles
 */
void options(){
register int i;
char s[128];
struct optinfo *p;
char *endof();
extern char *SRCDIR,*USERFILE;

clearscreen();
sysarg=usedfiles();
for(i=0;i<SIZEOPT;++i){
  strcpy(s,(p= &opt[i])->descrip);
  if(p->val) ssprintf(endof(s),": %u",*(p->val));
  xPrint(s,i,1);
}
crlf();
         displaymemoryleft();

#if PIINILOAD
#if TERMCAP
ePrint("TERM="); ePrint(termtype); crlf();
ePrint("Keymap="); ePrint(INIFILE); crlf();
ePrint("SrcDir="); ePrint(SRCDIR); crlf();
ePrint("SetUp="); ePrint(USERFILE); crlf();
#endif
#endif

         pressany("");
}

#if 0
struct rusage
   {
        struct timeval ru_utime; /* user time used */
        struct timeval ru_stime; /* system time used */
        long ru_maxrss;          /* maximum resident set size */
        long ru_ixrss;      /* integral shared memory size */
        long ru_idrss;      /* integral unshared data size */
        long ru_isrss;      /* integral unshared stack size */
        long ru_minflt;          /* page reclaims */
        long ru_majflt;          /* page faults */
        long ru_nswap;      /* swaps */
        long ru_inblock;         /* block input operations */
        long ru_oublock;         /* block output operations */
        long ru_msgsnd;          /* messages sent */
        long ru_msgrcv;          /* messages received */
        long ru_nsignals;        /* signals received */
        long ru_nvcsw;      /* voluntary context switches */
        long ru_nivcsw;          /* involuntary context switches */
   };
#endif

#if BSDunix || LINUX
#include <sys/time.h>
#include <sys/resource.h>

int my_getrusage(n,r)
int n; struct rusage *r;
{
#if 0
/proc/pid/status has these lines:
VmSize:     1376 kB
VmLck:         0 kB
VmRSS:       588 kB
VmData:      144 kB
VmStk:         8 kB
VmExe:        84 kB
VmLib:      1100 kB
structure names:
long ru_ixrss;      /* integral shared memory size */
long ru_idrss;      /* integral unshared data size */
long ru_isrss;      /* integral unshared stack size */
#endif
char s[512];
char *p;
FILEP fi;
extern char *sob(),*fnb(),*strstr();
extern long atol();
  sprintf(s,"/proc/%d/status",getpid());
  fi=fopenbread(s);
  if(fi==0) goto err;
  while(fgetln(s,fi)>0){
    if((p=strstr(s,"VmExe"))!=0)
        r->ru_ixrss=atol(sob(fnb(p)));  /* See usage below */
    if((p=strstr(s,"VmData"))!=0)
        r->ru_idrss=atol(sob(fnb(p)));
    if((p=strstr(s,"VmStk"))!=0)
        r->ru_isrss=atol(sob(fnb(p)));
  }
  closedfile(fi);
  return 0;
err:
  return -1;  /* Error */
}

int getusageBSD(s) char *s; {
struct rusage r;
  if(my_getrusage(RUSAGE_SELF,(struct rusage *)&r) == -1)
  s[0]=0; else
  sprintf(s,"Shared Text=%lu, Private Data=%lu, Stack=%lu",
              r.ru_ixrss, r.ru_idrss, r.ru_isrss);
}

long farcoreleft(){
  return MAXBLOCK*(long)SIZEBLOCK;
}
#endif

#if SYS5 && !LINUX
long farcoreleft(){
long ulimit();
long x,y;
  x=ulimit(3,0L);       /* hard limit */
  y=MAXBLOCK*SIZEBLOCK;
  return (x<y ? x : y);
}
#endif

/*
 * GETWDIR
 *
 * Get working directory string
 */

char * cwd(){
extern char *cd;
extern char *getwdir(),*getenv();

#if BSDunix || SYS5
   char *p;
   if(cd[0]==0){
     if((p=getenv("PWD")) != (char *)0) strcpy(cd,p);
     else getwdir(cd);
     if(cd[0]!=0 && tailchar(cd) != '/') strcat(cd,"/");
   }
#endif
#if MSDOS
   if(cd[0]==0)
   if(getwdir(cd) != (char *)0 && tailchar(cd) != '\\') strcat(cd,"\\");
#endif
  return cd;
}

/*
 * GETFNAME
 *
 * Get a file name from user.
 */
int getfname(msg,s) char *msg,*s;{
extern char * cwd();
  s[0]=0;
  if(cwd()) strcpy(s,cwd());
  if(isadir(s)) addslash(s);
#if POPUPS
  if(ggetstr(msg,s)!=RETURNKEY) s[0]=0;
  else
  if(isadir(s) || iswild(s)){
      strcpy(sysv,s);
      if(!FindFile(sysv,s,0)) s[0]=0;
  }
#else
  ggetstr(msg,s);
  if(cwd()) {
    if(strcmp(cwd(),s)==0) s[0]=0;
  }
#endif
  return s[0];
}

int queryreplace(insmode) int insmode;{
int fiddle,c,i; extern char *strchr();
int gotreplarg;
LONG ccBIOS();
  fiddle=0;
top:
  strcpy(sysv,srcharg);
  if(ggetstr("Search for: ",sysv)!=RETURNKEY || !sysv[0])
    goto quit;
  copysrcharg(sysv);
  if(search(1)){ /* Search forward */
    c='X';
    gotreplarg=0;
    while(1){
      nodisp=0;
     if(c=='A' && ccBIOS(2)) break;
     if(c!='A'){
        banner("REPLACE: Y=Yes, A=All, Q=Quit, P=Prev, N=Next [N]: ");
        if(strchr("YAQPN\003\025\r\n\t\040",c=getupper())==(char *)0)
          continue;
        if(strchr("Q\003\025",c)) goto erase;
        if(c=='P'){
          unshowmatch(); if(!rfind()) goto top; continue;
        }
        if(c<=' ') c='N';
      }
      if(c!='N'){
        if(!gotreplarg){
          if(ggetstr("Replace with: ",replarg) != RETURNKEY) goto erase;
          gotreplarg=1;
        }
        /* del match at cursor, and replace */
        if(!zap(sysx))break; /* see showmatch() */
        fiddle=1;
        if((i=addstring(replarg,line[row],TRUE,insmode))== -1) break;
        if(i) curfix(col+curoff+i,0); else {edleft(); curses();}
      } else
      unshowmatch();       /* Erase inverse video of showmatch() */
      if(!ffind()) {bell(); goto quit;}
    }
  } else goto quit;
erase:
  unshowmatch();       /* Erase the match */
quit:
  lcount();
  clrcon();            /*  clear type-aheads */
  return fiddle;
}

/* stack for macros so they can do recursion to 10 levels */
void setlevel(maclevel,poppush)
int maclevel;           /* level 0-9 to pop or push */
int poppush;            /* 1=push, 0=pop */
{
struct keep {int MACrep,MACind,MACmax,MACrun,RECord; char *MACbuf;};
static struct keep level[10];
  if(poppush){  /* push values  */
    level[maclevel].MACrep=macrep;
    level[maclevel].MACind=macind;
    level[maclevel].MACmax=macmax;
    level[maclevel].MACrun=macrun;
    level[maclevel].RECord=record;
    level[maclevel].MACbuf=macbuf;
  } else {      /* pop values */
    macrep=level[maclevel].MACrep;
    macind=level[maclevel].MACind;
    macmax=level[maclevel].MACmax;
    macrun=level[maclevel].MACrun;
    record=level[maclevel].RECord;
    macbuf=level[maclevel].MACbuf;
  }
}

#if 0  /* For LINUX */
#if 0
Example:
 int main(){
  ash_system ("date >/tmp/theDate");
  exit(0);
 }
#endif

#include <errno.h>

 int ash_system (command) char *command; {
           int pid, status;
           extern int errno;
           extern char **environ;

           pid = fork();
           if (pid == -1)
               return -1;
           if (pid == 0) {
               char *argv[4];
               argv[0] = "ash";
               argv[1] = "-c";
               argv[2] = command;
               argv[3] = 0;
               execve("/bin/ash", argv, environ);
               exit(127);
           }
#if 0
           do {
               if (waitpid(pid, &status, 0) == -1) {
                   if (errno != EINTR)
                       return -1;
               } else
                   return status;
           } while(1);
#else
           return 0;
#endif
       }

#endif

#if MSDOS
void runWinprogram(s) char *s; {
char t[256];
  ssprintf(t,"%s%s",(isWinEnhanced()? "start ":""),s);
  runprogram(t,0);
}
#endif

#if BSDunix || SYS5
int isXwindows(){
extern char *getenv();
  if(getenv("DISPLAY")) return 1;
  return 0;
}
#endif

#if BSDunix || SYS5
/* Inject system command pipe [READPIPE] */

static FILEP PIPEinput=0;

#include <signal.h>
void pipehandler(){
extern int PIPEopen;
 signal(SIGINT,SIG_DFL);
    PIPEopen=0;
    termraw();          /* Interrupt chars off */
    pclose(PIPEinput);
}

int readpipe(s) char *s;{
extern int PIPEopen;
int i;
  PIPEopen=0;
  PIPEinput = popen(s,"r");
  if(!PIPEinput){
    ssprintf(sysvv,"Can't pipe %s",s);  bellpressany(sysvv);
    return 1;
  }
  PIPEopen=1;
  termcanon();          /* Interrupt chars on */
  signal(SIGINT,pipehandler);
  return 0;
}

int getpipe(){
int i;
    if((i=getc(PIPEinput)) != EOF){
      if(i=='\r') return EXTLEFTARROW;
      if(i=='\n') return RETURNKEY;
      return i;
    }
quit:
    if(!PIPEopen) return;
    PIPEopen=0;
    termraw();          /* Interrupt chars off */
    pclose(PIPEinput);
}
#endif

/* Get the shell variable */
void getSHELL(){
extern char *SHELL;
extern char *getenv();
char *p;
   if((p=getenv("SHELL"))!=0) newcore(&SHELL,p);
}

/* Run a system program under unix, msdos */
void runprogram(s,flag) char *s; int flag; {
LONG ccBIOS();
  if(s[0]) {
    exitterminal();     /* Clear screen, set colors */
    termDoExit();
    system(s);          /* MSDOS: Could be swapped if swapswitch==1 */
    setup();
    if(flag){
      etype("\r\nPress any key to return to PI:");
      ccBIOS(0);
    }
  }
  initterminal();       /* reset terminal in case we lost keypad */
  clrcon();             /* clear out type-aheads */
}

/* set STOPS to power of two */
void setdetab() {
char s[100]; int x; extern int STOPS;
char *strchr();
  ssprintf(s,"Detab %s. New 2,4,8,16,32",
  ((DETABBER=1-DETABBER)!=0 ? " ON" : " OFF"));
  x=STOPS; getnumquery(s,&x,2,32);
  if(strchr("\2\4\10\20\40",x)!=(char *)0) STOPS=x;
}

void settabs(){
extern int margin,curoff,col,TSTOPS,MAXLN;
extern char *sysv;
  getnumquery("Tab key indent",&TSTOPS,0,MAXLN-2);
  if(sysv[0]=='.') TSTOPS=col;
  getnumquery("Right margin",&margin,0,MAXLN-2);
  if(sysv[0]=='.') margin=curoff+col;
}

void settoggle(msg,x) char *msg; int *x; {
  bbanner(msg,((x[0]= (x[0]? 0:1))!=0 ? " ON" : " OFF"));
}

void shrinkwindow() {
extern int ROWMAX,COLMAX,MAXROWS,MAXLN;
int rrow,ccol;
  rrow=ROWMAX+1; getnumquery("New ROWS",&rrow,1,MAXROWS);
  ccol=COLMAX+1; getnumquery("New COLUMNS",&ccol,1,MAXLN-1);
  setwindowsize(rrow-1,ccol-1);
}

void setwindowsize(rrow,ccol) int rrow,ccol; {
#if TURBOC
extern int curoff,row,col,ROWMAX,COLMAX,MAXROWS,FARSCROLL;
  if(PackScreenBuffer(0)) return;
  equipment();                  /* Have to erase old colors */
  FARSCROLL=1+(ROWMAX=rrow);
  COLMAX=ccol;
  UnPackScreenBuffer(0);
  row=min(row,ROWMAX); col=min(col,COLMAX);
  drawall(); killstatus();
#else
extern int curoff,row,col,ROWMAX,COLMAX,MAXROWS,HASSCROLL,FARSCROLL;
int NN,f,fiddle;
  if(PackScreenBuffer(0)) return;
  FARSCROLL=1+(ROWMAX=rrow);
  COLMAX=ccol;
  setdrawbuff();                /* unix network screen buffer */
  UnPackScreenBuffer(0);  if(HASSCROLL) setscroll(0);
  row=min(row,ROWMAX); col=min(col,COLMAX);
  drawall(); killstatus();
#endif                          /* for fast Xterm refresh */
}

#if 0
Foreign.
Unix iso8859-1 and DOS5.0 Translation table for Czech characters
Not all characters are correct. However, there is a character editor,
called fed, which may be able to alter a font for the purpose.
Newer versions of X already have the latin-2 fonts (eg, X11R6).
See www.biz.net.pl in Warsaw for free fonts 8859-2.
;==========================================================
;Unix iso8859-1 and DOS5.0 swap table for Czech characters
;==========================================================
;Chr  Oct  Dec      DOS5.0 Decimal
;==========================================================
;' ', 264   180     243     /* Hook on blank space */
;'a', 341   225     160     /* long vowel, lower case */
;'e', 351   233     130
;'i', 355   237     161
;'o', 363   243     162
;'u', 372   250     163
;'y', 375   253     236
;'c', 342   226     159     /* hook consonant, lower case */
;'d', 313   203     212
;'n', 361   241     229
;'r', 374   252     253
;'s', 344   228     231
;'t', 345   229     156
;'z', 366   246     167
;'A', 301   193     181     /* long vowel, upper case */
;'E', 311   201     144
;'I', 315   205     214
;'O', 323   211     224
;'U', 332   218     233
;'Y', 335   221     237
;'C', 302   194     172     /* hook consonant, upper case */
;'D', 310   200     210
;'N', 321   209     213
;'R', 334   220     252
;'S', 304   196     230
;'T', 333   219     155
;'Z', 326   214     166
;372, 371   249     163, 133     /* long u to u-circle */
;371, 'u'           133, 'u'     /* u-circle to u */
;351, 352   234     130, 216     /* long e to e-hook */
;352, 'e'           216, 'e'     /* e-hook to e */
;311, 312   202     144, 183     /* long E to E-hook */
;312, 'E'           183, 'E'     /* E-hook to E */
;
#endif

#if 0

8859-2 Codes from unix

base    hex     dec     oct     description
===========================================
a       e1      225     341     a-long
A       c1      193     301     A-long
c       e8      232     350     c-hook
C       c8      200     310     C-hook
d       ef      239     357     d-soft
D       cf      207     317     D-hook
e       e9      233     351     e-long
e       ec      236     354     e-hook
E       c9      201     311     E-long
E       cc      204     314     E-hook
i       ed      237     355     i-long
I       cd      205     315     I-long
n       f2      242     362     n-hook
N       d2      210     322     N-hook
o       f3      243     363     o-long
O       d3      211     323     O-long
r       f8      248     370     r-hook
R       d8      216     330     R-hook
s       b9      185     271     s-hook
S       a9      169     251     S-hook
t       bb      187     273     t-soft
T       ab      171     253     T-hook
u       fa      250     372     u-long
u       f9      249     371     u-circle
U       da      218     332     U-long
U       d9      217     331     U-circle
y       fd      253     375     y-long
Y       dd      221     335     Y-long
z       be      190     276     z-hook
Z       ae      174     256     Z-hook

Cycle table constructed by hand from the above info:
a\341      /*   a ==>  a-long */
A\301      /*   A ==>  A-long */
c\350      /*   c ==>  c-hook */
C\310      /*   C ==>  C-hook */
d\357      /*   d ==>  d-soft */
D\317      /*   D ==>  D-hook */
i\355      /*   i ==>  i-long */
I\315      /*   I ==>  I-long */
n\362      /*   n ==>  n-hook */
N\322      /*   N ==>  N-hook */
o\363      /*   o ==>  o-long */
O\323      /*   O ==>  O-long */
r\370      /*   r ==>  r-hook */
R\330      /*   R ==>  R-hook */
s\271      /*   s ==>  s-hook */
S\251      /*   S ==>  S-hook */
t\273      /*   t ==>  t-soft */
T\253      /*   T ==>  T-hook */
y\375      /*   y ==>  y-long */
Y\335      /*   Y ==>  Y-long */
z\276      /*   z ==>  z-hook */
Z\256      /*   Z ==>  Z-hook */
/* Special cases of three chars on one base */
e\351      /*   e ==>  e-long */
\351\354   /*   e-long ==> e-hook */
E\311      /*   E ==>  E-long */
\311\314   /*   E-long ==> E-hook */
u\372      /*   u ==>  u-long */
\372\371   /*   -long ==> u-circle */
U\332      /*   U ==>  U-long */
\332\331   /*   U-long ==> U-circle */

Cycle table in input format:
a\341A\301c\350C\310d\357D\317i\355I\315n\362N\322o\363O\323r\370
R\330s\271S\251t\273T\253y\375Y\335z\276Z\256e\351
\351\354E\311\311\314u\372\372\371U\332\332\331

#endif

struct swapchar {unsigned char old,new;};
#if 0
static
struct swapchar czechchars[]={
{' ', 243},     /* Hook on blank space */
{'a', 160},     /* long vowel, lower case */
{'e', 130},
{'i', 161},
{'o', 162},
{'u', 163},
{'y', 236},
{'c', 159},     /* hook consonant, lower case */
{'d', 212},
{'n', 229},
{'r', 253},
{'s', 231},
{'t', 156},
{'z', 167},
{'A', 181},     /* long vowel, upper case */
{'E', 144},
{'I', 214},
{'O', 224},
{'U', 233},
{'Y', 237},
{'C', 172},     /* hook consonant, upper case */
{'D', 210},
{'N', 213},
{'R', 252},
{'S', 230},
{'T', 155},
{'Z', 166},
{163, 133},     /* long u to u-circle */
{133, 'u'},     /* u-circle to u */
{130, 216},     /* long e to e-hook */
{216, 'e'},     /* e-hook to e */
{144, 183},     /* long E to E-hook */
{183, 'E'}      /* E-hook to E */
};
#define NUMSWAP (sizeof(czechchars)/sizeof(czechchars[0]))
int NUMACCENT=NUMSWAP;
struct swapchar *accentchars= &czechchars[0];
#endif

int NUMACCENT=0;
struct swapchar *accentchars= (struct swapchar *)0;

unsigned foreign(c)
unsigned char c;
{
extern int NUMACCENT; extern struct swapchar *accentchars;
static struct swapchar *p;
int i;
  for(p=accentchars,i=0;i<NUMACCENT;++i,++p){
    if(c == p->old) return (unsigned)p->new;
  }
  for(p=accentchars,i=0;i<NUMACCENT;++i,++p){
    if(c == p->new) return (unsigned)p->old;
  }
  return 0;
}

#if 0
void dumpforeign()
{
extern int NUMACCENT; extern struct swapchar *accentchars;
unsigned char *p;
char s[256];
p = (unsigned char *)&accentchars[0];
strncpy(s,p,NUMACCENT); s[NUMACCENT]=0;
ePrint(s);
crlf(); fprintf(stderr,"NUMACCENT=%d\r\n",NUMACCENT);
}
#endif

unsigned char *curchar() {
int i; char *p;
  i=strlen(p=line[row]);
  while(i<curoff+col) p[i++]=' '; p[i]=0;
  return ((unsigned char *)&line[row][curoff+col]);
}

int makeupper(p) char *p; {
  if(islower(p[0])) {p[0] = toupper(p[0]); return 1;}
  return 0;
}

int makelower(p) char *p; {
  if(isupper(p[0])) {p[0] = tolower(p[0]); return 1;}
  return 0;
}

/*
 * CASECHANGE
 *
 * Purpose: Change character(s) at cursor to lower or upper case.
 */
int casechange(n)
int n;                  /* n=0 is lower, 1=upper */
{
int tag;
char *p,*q,*r; extern char *sobl();
  tag=0;
  if(strlen(r=line[row])<=col+curoff)  /* nothing to map */
    goto nextline;
  q=p= r+col+curoff;
  switch(n){                            /* Single character map */
  case CASELOWER: /* change to lower case */
          tag=makelower(p++);
          goto print;
  case CASEUPPER: /* change to upper case */
          tag=makeupper(p++);
          goto print;
  }/* end switch */

  if(*(p=sobl(p))==0) goto nextline;

  switch(n){                            /* map whole words */
  case CASECAPITAL: /* Capitalize: first char upper, rest lower */
          tag=makeupper(p++);
          if(p != sobl(p)) break;
  case CASELLOWER: /* change word to lower case */
          while(*p){
            tag |= makelower(p++);
            if(p != sobl(p)) break;
          }
          break;
  case CASELUPPER: /* change word to upper case */
          while(*p){
            tag |= makeupper(p++);
            if(p != sobl(p)) break;
          }
          break;
  }/* end switch */
print:
  n=(int)(p-q);
  if(tag) {
    strnzcpy(sysvv,q,n);
    xPrint(sysvv,row,col);
  }
  if(*sobl(p)!=0) {curfix(1+(int)(p-r),0); goto quit;}
nextline:
  homeit(); eddn();
quit:
  return tag;
}

void copydefaultfile(){
  strcpy(sysfile,finfo[curfile].fname);
}

#if 0 /* For Hilite mouse tracking and region information */
      /* This did not work very well with the current version of xterm */
      /* but someday it might, if they fix it, so I leave it here. */
void pressmouse(){
 char s[128];
 int a,b,c;
 extern char *endof();
 a=getbyte()&3; b=getbyte()-' '-1; c=getbyte()-' '-2;
 if(a == 1 || a == 2) { /* Press MB2 or MB3? Ignore hilite */
   etype("\033[0T");
 }
 if(a == 0) {           /* Press MB 1 */
   if(c>=0 && b>=0 && c<=ROWMAX && b<=COLMAX) { /* Enable Hilite */
     while(row >c) decrROW();
     while(row <c) incrROW();
     row=c; col=b;
     ssprintf(s,"\033[1;%u;%u;2;%u;T", b+1, c+2, ROWMAX+1);
     etype(s);
   }                                            /* Ignore hilite */
   else {
     etype("\033[0;0;0;0;0T");
   }
 }
/*
 banner("Mouse Press:");
 fprintf(stderr,"MB%d: b=%d, c=%d",a+1,b,c);
*/
}

void releasemouse(c) char c; {
 /* Release MB1, MB2, MB3 */
 /* Valid: ESC [ t Cx Cy */
 /* Invalid: ESC [ T Cx Cy Cx Cy Cx Cy */
 int startx, starty, endx, endy, mousex, mousey;
 /* startx, starty, endx, and endy give the starting and ending
    character positions of the region.  mousex and mousey give the
    location of the mouse at button up, which may not be over a
    character.
  */
 if(c == 't'){
  endx=getbyte(); endy=getbyte();
  /* banner("Mouse t:"); fprintf(stderr,"%d %d",endx,endy); */
 }
 if(c == 'T'){
  startx=getbyte(); starty=getbyte();
  endx=getbyte(); endy=getbyte();
  mousex=getbyte(); mousey=getbyte();
  /* banner("Mouse T:"); */
  /* fprintf(stderr,"%d %d %d %d %d %d",
                     startx,starty,endx,endy,mousex,mousey); */
 }
}

void setmouse(n){
char s[128];
static int wasdone=0;

 if(n==0 && wasdone==1){
   etype("\033[?1001l");
   wasdone=0;
   return;
 }
 identifykey(MOUSEPRESS,s);
 if(n==1 && strcmp(s,"Keymap") != 0 && mouseswitch !=0) {
   etype("\033[?1001h");
   wasdone=1;
 }
}

#endif

#if BSDunix || SYS5 || (MSDOS && DOS386I)
void pressmouse(){
 char s[128];
 int a,b,c;
 a=getbyte()&3; b=getbyte()-' '-1; c=getbyte()-' '-2;
 /* Press MB2 (a==1) or MB3 (a==2)? Ignore */
 if(a == 0) {           /* Press MB 1 (a==1)? */
   if(c>=0 && b>=0 && c<=ROWMAX && b<=COLMAX) { /* Valid position */
     while(row >c) decrROW();
     while(row <c) incrROW();
     row=c; col=b;
   }
 }
}
#endif

#if BSDunix || SYS5
void releasemouse() {
/* Release taken care of by pressmouse() above */
}

void setmouse(n){
 etype(n==0 ? "\033[?9l" :"\033[?9h");
 banner(n==0 ? "OFF" : "ON");
}

int mouseexists() {
extern int mouseswitch;
   return mouseswitch;
}

int mousekeys() {
char s[128];
   identifykey(MOUSEPRESS,s);
   if(strcmp("Keymap",s)==0) return 0;  /* Must be a key defined */
   return 1;
}
#endif

/* Replace dest by its integer computation */
void convert2int(dest) char *dest; {
LLONG x,y; char *p; int i;
char *sob(); LLONG atoiLL(); extern char *lltoa();

   x=atoiLL(p=sob(dest));
   while(p[0]){
     if((p[0]=='-' || p[0]=='+') && isdigit(p[1])) ++p;
     while(isdigit(p[0])) ++p; p=sob(p);
     i= *p++; y=atoiLL(p);
     switch(i){
     case '+': x += y; goto addone;
     case '-': x -= y; goto addone;
     case '/': x /= y; goto addone;
     case '*': x *= y;
     addone: p=sob(p); break;
     case '=': if(p[0]=='='){
                 if(x == atoiLL(p+1)) x=1; else x=0;
               }
     default: goto quit;
     }
   }
quit:
   lltoa(x,dest);
}

/* Replace hexadecimal string dest by its base 10 equivalent */
void hexstrToBase10str(dest) char *dest; {
LLONG x,y; char *p;
extern char *sob(); extern char *lltoa();
extern void cvupper();

   cvupper(dest);
   p=sob(dest);
   if(p[0]=='0' && p[1]=='X') p += 2;   /* 0x23fe is a hex number format */
   x=0;
   while((y=hexdigit(p[0]))!= -1){
     x=x*16+y; ++p;
   }
   (void)lltoa(x,dest);
}

/* Replace octal string dest by its base 10 equivalent */
void octstrToBase10str(dest) char *dest; {
LLONG x,y; char *p;
extern char *sob(); extern char *lltoa();
extern void cvupper();

   cvupper(dest);
   p=sob(dest);
   if(p[0]=='\\') ++p;  /* \033 is an octal number format */
   x=0;
   while((y=octdigit(p[0]))!= -1){
     x=x*8+y; ++p;
   }
   (void)lltoa(x,dest);
}


/* Warning: This strchr() must not match the null string terminator! */
/* See v13.c for the source to strchr() */
int isdigit(c) char c; {
extern char *strchr();
 return (strchr("0123456789",c) != (char *)0);
}

int hexdigit(c) char c; {
  if('0'<= c && c<= '9') return c-'0';
  if('A' <= c && c <= 'F') return c-'A'+10;
  return -1;
}

int octdigit(c) char c; {
  if('0'<= c && c<= '7') return c-'0';
  return -1;
}

void setSYSone(){
 sysv[0]= '1'; sysv[1]=0;
}

static int tmp;
void PUSHnodisp(){
 tmp=nodisp;
}
void POPnodisp(){
 nodisp=tmp;
}

LLONG absLL(lx) LLONG lx; {
 return (lx<0) ? -lx: lx;
}

LLONG getatoiLL() {
extern LLONG atoiLL(); extern char *sysv;
 return atoiLL(sysv);
}

void getbackupfname(s) char *s; {
/* variable BACKUPS: 0=FALSE,1=$HOME/PIbackups,2=./PIbackups */
char *p;
extern int BACKUPS;
extern char *PIBACKUPS;
extern char *getenv();
extern char *cwd();
  if(!BACKUPS) {s[0]=0; return;}
  p=0;
  if(BACKUPS==2) p=cwd();
  else
  if(BACKUPS==1) p=getenv("HOME");
  if(!p) p=".";
  strcpy(s,p);
  addslash(s);
  strcat(s,PIBACKUPS);
}

/* variable BACKUPS: 0=FALSE,1=$HOME/PIbackups,2=./PIbackups */
void setbackups(){
extern int BACKUPS;
char buf[MAXFNAME];
  if((++BACKUPS)==3) BACKUPS=0;
  getbackupfname(buf);
  ssprintf(sysv,"Backups->%s",(!BACKUPS ? "OFF" : buf) );
  banner(sysv);
}

void setupcolors(){
#if MSDOS
int x,i;
  x=(SYSATT >> 4);
  getnumquery("Background color (0-15) ",&x,0,15);
  i=(SYSATT & 15);
  getnumquery("Foreground color (0-15) ",&i,0,15);
  SYSATT = ((i & 15) + (x << 4));
  getnumquery("Window title attribute (19=blue+yell,15=mono) ",&titlattr,0,128);
  getnumquery("Window border attribute (31=bold,15=mono) ",&bordattr,0,128);
#endif
#if BSDunix || SYS5
  getnumquery("Screen attribute (0=std,7=mono,128=bold,112=rev video) ",&SYSATT,0,128);
  getnumquery("Window title attribute (0,7,128,112) ",&titlattr,0,128);
  getnumquery("Window border attribute (0,7,128,112) ",&bordattr,0,128);
  getnumquery("Text foreground color (0,30-37) ",&textfgcolor,0,37);
  getnumquery("Text background color (0,40-47) ",&textbgcolor,0,47);
  getnumquery("Menu foreground color (0,30-37) ",&menufgcolor,0,37);
  getnumquery("Menu background color (0,40-47) ",&menubgcolor,0,47);
  setansicolors();
  clearscreen();
#endif
}

/* kind of instr() for C */
char *strstr(a,b)
char *a,*b;
{
int n;
  n=strlen(b);
  while(a[0]){
   if(strncmp(a,b,n)==0) return  a;
   ++a;
  }
  return (char *)0;
}
