/*
 *  This file contains TERMCAP code for DOS and UNIX machines, and
 *  any other operating system that can support terminal types. No
 *  termcap databases are supplied, except for DOS, because it is
 *  preferable to get the termcap file from the local system.
 *
 *  This file contains KEYMAP code for mapping keys on startup and
 *  also on the fly.
 *
 */

#define KEYDEFS 1
#include "v.h"

#if PIINILOAD

#undef BUFSIZE
#define BUFSIZE 1536
extern char *getenv(),*strchr(),*strstr(),*tgetstr(),*core(),*skipto(),*sob();
extern char *endof(),*skipdelay(),*fpick(),*fnb();
static char    *eebuf = (char *)0;       /* pointer to entry buffer */

#if BSDunix || SYS5
#define TERMFILE  "/etc/termcap"
#ifndef PIEROOT
#define PIEROOT "/usr/local/lib/pie/"
#endif
char *SETEMACS=0;
char *SETWDSTAR=0;
char *SETWDPERF=0;
char * SRCDIR=0;
char * VT100key;
char * KEYMAP;
char * SYSUSERFILE;
char * USERFILE;
char * DUMPKEY;
char * PIMACROS;
char * PIHLPDOC;

#define PIDUMPKEY  "pi-dump.key"
#define PIMACROKEY "pi-macro.key"
#define PISET      ".pirc"
#define LOCALPISET "./.pirc"
struct dataset {char *name; char **addr;};
struct dataset unixnames[]={
{"emacs.keys", &SETEMACS},
{"wdstar.keys", &SETWDSTAR},
{"wdperf.keys", &SETWDPERF},
{"keymap.dat",&KEYMAP},
{"SYSpi.rc",&SYSUSERFILE},
{"pihelp.htm",&PIHLPDOC}
};
#define SIZEunixnames sizeof(unixnames)/sizeof(unixnames[0])

void getprogramname(s) char *s; {
extern char *progname;
extern char *realpath();
char *p;
  if(realpath(progname,s) ) return;
  if( realpath("/bin/pi",s) ) return;
  if( realpath("/usr/local/bin/pi",s) ) return;
  strcpy(s,PIEROOT); addslash(s); strcat(s,"pi");
}

/* Create file names with HOME as the directory */
/* Unix has the trouble of no write permission in various places */
/* So the files have to be kept in a read/write location */
/* File ./.pirc is special. If it exists, then assume it is read/write. */
/* Always, pi-macro.key is written to the current directory, but that */
/* can fail. */
void setfiledefaults(){
char *getenv(),*endof(),*ffname();
char *p;
int i;
char homedir[MAXFNAME];
  /* source ./pi-macro.key */
  newcore(&PIMACROS,"pi-macro.key");

  /* source $HOME/pi-dump.key */
  p=getenv("HOME");
  strcpy(homedir,(p? p:"."));
  addslash(homedir);
  p=endof(homedir);
  strcpy(p,PIDUMPKEY); newcore(&DUMPKEY,homedir);

  /* sources ./.pirc or  $HOME/.pirc */
  if(classify(LOCALPISET)==1) {strcpy(homedir,"./"); p=endof(homedir);}
  strcpy(p,PISET); newcore(&USERFILE,homedir);

  /* Find the SRCDIR by lookup of the program name. */
  getprogramname(homedir);       /* Worst case is PIEROOT */
  *ffname(homedir)=0;            /* Strip off the exe name */
  if(!SRCDIR) newcore(&SRCDIR,homedir);

  /* source $PIE/keymap.dat and others */
  strcpy(homedir,SRCDIR); addslash(homedir); p=endof(homedir);
  for(i=0;i<SIZEunixnames;++i){
    strcpy(p,unixnames[i].name);
    newcore(unixnames[i].addr,homedir);
  }
}

void queryLocalPiSet(){
char *p;
  ssprintf(sysvv,"./%s",ffname(USERFILE));
  if(classify(sysvv)==1) return;
  if(getyesno("Create local .pirc")){
    ssprintf(sysv,"cp %s %s",USERFILE,sysvv);
    system(sysv);
    if(classify(sysvv)==1) newcore(&USERFILE,sysvv);
  }
}
#endif

#if MSDOS
#define DUMPKEY   "pi-dump.key"
#define PIMACROS  "pi-macro.key"
char * SRCDIR=0;
char * KEYMAP;
char * USERFILE;
char *PIHLP="pihelph";

struct dataset {char *name; char **addr;};
struct dataset dosnames[]={
{".dat",&KEYMAP},
{".set",&USERFILE}
};
#define SIZEdosnames sizeof(dosnames)/sizeof(dosnames[0])

void setfiledefaults(){
char *ffname(),*endof(),*strchr();
char *p;
int i;
char homedir[MAXFNAME];
char dest[MAXFNAME];
  getprogramname(homedir);
  strcpy(dest,ffname(homedir));
  if((p=strchr(dest,'.'))!=(char *)0) p[0]=0;
  *ffname(homedir)=0;
  if(!SRCDIR){
    newcore(&SRCDIR,homedir);
  } else strcpy(homedir,SRCDIR);
  addslash(homedir);
  strcat(homedir,dest);
  p=endof(homedir);
  for(i=0;i<SIZEdosnames;++i){
    strcpy(p,dosnames[i].name);
    newcore(dosnames[i].addr,homedir);
  }
}

void queryLocalPiSet(){
char *p;
  ssprintf(sysvv,".\\%s",ffname(USERFILE));
  if(classify(sysvv)==1) return;
  if(getyesno("Create local .pirc")){
    ssprintf(sysv,"cp %s %s",USERFILE,sysvv);
    system(sysv);
    if(classify(sysvv)==1) newcore(&USERFILE,sysvv);
  }
}

#endif


/* find file name after first colon, if any */
char *
fpick(pat,s)
char *pat;      /* pattern to match */
char *s;        /* string to search */
{
   if(strncmp(s,pat,strlen(pat))==0){
        s=skipto(":",s);
        if(s[0] && s[1]) return s+1;
   }
   return "";
}

/* Use core() to build memory blocks not to be released by free() */
#if MSDOS
char *core(n) int n;{
char *p;
  if((p=(char *)sbrk(n))==(char *)-1){
    etype("Out of memory"); quit(100);
  }
  fillchr(p,0,n);
  return p;
}
#else                   /* DEC3100 needs word alignment - use malloc() */
char *core(n) int n;{
char *p;
  if((p=(char *)malloc(n))==(char *)0){
    etype("Out of memory"); quit(100);
  }
  fillchr(p,0,n);
  return p;
}
#endif

/* Get core, copy string */
void newcore(d,s) char **d; char *s; {
  strcpy(d[0]=core(strlen(s)+1),s);     /* Uses sbrk() under MSDOS */
}

int safecore(d,s) char **d; char *s; {
  if((d[0]=(char *)malloc(strlen(s)+1))==(char *)0) return 0;
  strcpy(d[0],s);
  return 1;
}

void fartonearstrcpy(d,s) char *d; char far *s;
{
  while((*d++ = *s++) != 0) ;
}


#if MSDOS
void far * farcore(n) unsigned long n; {
void far *p;
  if((p=farmalloc(n))==0){
    bellpressany("farmalloc failed");
    exit(0);
  }
  return p;
}

int farsafecore(d,s) char far **d; char *s; {
char far *p;
  if((p=(char far *)farmalloc((unsigned long)strlen(s)+1))==(char far
  *)0){
    bellpressany("farmalloc failed");
    return 0;
  }
  d[0]=p;
  while((*p++ = *s++) != 0) ;
  return 1;
}


#else
int farsafecore(d,s) char far **d; char far *s; {
  return safecore(d,s);
}

#endif

void getcorememblock(){
int i;
#if !MSDOS
    memmark=(char *)core(MAXBLOCK); /* core() nulls out the memory block */
    meminfo=(struct bucket *)core(sizeof(struct bucket)*MAXBLOCK);
    memblock=(char far **)core(sizeof(char far *)*MAXBLOCK);
#else
    memmark=farcore((unsigned long)MAXBLOCK);
    for(i=0;i<MAXBLOCK;++i) memmark[i]=0;
    meminfo=farcore((unsigned long)sizeof(struct bucket)*MAXBLOCK);
    memblock=farcore((unsigned long)sizeof(char far *)*MAXBLOCK);
#endif
}

#if MSDOS
#if defined(__SMALL__) || defined(__MEDIUM__) || defined(__LARGE__)
/* TURBOC. Used for debugging */

static char *watermark=0;

void nearmemorybaseinit(){
 if(watermark==0) watermark=(char *)sbrk(0);
}
#endif
#endif


#if MSDOS
void freeup(s) char *s; {
extern void free();
  /* Try not to "free" a memory address for compiled-in data */
  if(s!=(char *)0 && s[0]!=0 && (long)s >= (long)watermark) free(s);
}
#else
void freeup(s) char *s; {
extern void free();
  if(s!=(char *)0 && s[0]!=0) free(s);
}
void farfree(s) char far *s; {
  freeup(s);
}
#endif

/* copy string to 127 bytes, null terminate */
void grabtoken(d,s) char *d,*s;{
char *fnb();
  moveMEM(d,s,127); d[127]=0; *fnb(d)=0;
}

void strnzcpy(d,s,n) char *d,*s; int n; {
  moveMEM(d,s,n); d[n]=0;
}



#if MSDOS
#if defined(__SMALL__) || defined(__MEDIUM__)
/* TURBOC. Used for debugging */
void displaymemoryleft(){           /* Debugging memory */
char s[128];
extern unsigned coreleft();  /* small and medium models only TURBOC */
lltoa((LLONG)farcoreleft(),s);  /* LLONG is long for MSDOS */
ssprintf(sysv,"Far memory = %s, Low memory=%u",s,coreleft());
ePrint(sysv); crlf();
lltoa((LLONG)MAXBLOCK*SIZEBLOCK,s);  /* LLONG is long for MSDOS */
ssprintf(sysv,"File size max = %s bytes",s);
ePrint(sysv); crlf();
}
#endif
#endif
#if !MSDOS
void displaymemoryleft(){           /* Debugging memory */
#if BSDunix || LINUX
char s[256];
sprintf(s,"File size max = %lu bytes",(long)farcoreleft());
ePrint(s); crlf();
getusageBSD(s);
ePrint(s); crlf();
#endif
}
#endif

/* Load global variables file to set dynamic dimensions */
void loadinifile(argc,argv) int argc; char **argv; {
char *p,*q;
int i,flag,flagsrc;
char s[128];
extern char *ffname();
extern char *fnb();
    /* should be able to pass resource file -set: on command line */
    /* (-dir:xxx -key:xxx -set:xxxx -term:xxxxx). Environment=defaults. */
/*
 * Ready to load up the resource files
 */
   flag=flagsrc=0;
   for(i=1;i<argc;++i){ /* scan command line for -set:, -dir: */
     getsetups(argv[i],&flag,&flagsrc);
   }
   /* hit environment if void command line */
   if((p=getenv("PIEINIT"))==(char *)0) p="";
   while(p[0]){
     getsetups(p,&flag,&flagsrc);
     p=fnb(p);
   }
   setfiledefaults();   /* All file names determined */
   if(flag) return;
   /* Get all integer setup variables. No keys yet. */
#if MSDOS
   if(classify(ffname(USERFILE))==1)
     loadrc(ffname(USERFILE));  /* File in connected directory */
   else
     loadrc(USERFILE);          /* $HOME or connected directory */
#else      /* unix */
   if(classify(ffname(USERFILE))==1) /* Get setup from connected dir */
     loadrc(ffname(USERFILE));
   else
   if(!loadrc(USERFILE)){
     char buf[512];
     loadrc(SYSUSERFILE);   /* Otherwise use system default */
     /* Try to copy SYSUSERFILE to USERFILE */
     ssprintf(buf,"cp %s %s",SYSUSERFILE,USERFILE);
     system(buf);
   }
#endif
}

void getsetups(p,flag,flagsrc) char *p; int *flag,*flagsrc; {
char *q; char *fpick();
     if(*(q=fpick("-s",p))!=0) if(!flag[0]) flag[0] = loadrc(q);
     if(*(q=fpick("-d",p))!=0){
       newcore(&SRCDIR,q);
       flagsrc[0]=SRCDIR[0]==0 ? 0 : 1;
     }
}

/* Load termcap entry for user terminal */
#if !TERMCAP
void loadtcap(){}
#else
char mybuf[BUFSIZE];            /* Put termcap entry into here */
void loadtcap(argc,argv) int argc; char **argv; {
char *p,*q;
int i,stat;
char s[128];
    /* should be able to pass terminal type on command line */
    /* (-key:xxx -set:xxxx -term:xxxxx). Use environment for defaults. */

    s[0]=0;
    for(i=1;i<argc;++i) {       /* User switch -term: on command line? */
      if(*(q=fpick("-t",argv[i]))!=0) strcpy(s,q);
    }
    if(!s[0]){         /* look for -term: in environment */
      if((p=getenv("PIEINIT"))==(char *)0) p="";
      while(p[0]){
        if(*(q=fpick("-t",p=sob(p)))!=0){
          grabtoken(s,q);
        }
        p=fnb(p);
      }
    }
    eebuf = &mybuf[0]; eebuf[0]=0;
    if(!s[0]) { /* No TERM variable found on cmdline or in PIEINIT */
      if((p = getenv("TERMCAP")) != (char *)0) {
        /* flesh out TERM variable from TERMCAP */
        p=skipto("|:",sob(p));
        if(p[0]=='|'){
          strnzcpy(s,p+1,127);
          *skipto("|:",s)=0;
        }
      }
    }
    if(!s[0]) { /* No term variable found yet, look in TERM */
      if((p = getenv("TERM")) != (char *)0) strcpy(s,p);
    }
    /* Find TERMCAP entry ... */
gettermcap:
    stat=tgetent(eebuf, s);
    /* mybuf[] loaded with TERMCAP entry, ready to decode. */
    /* get cm,al,dl,cs,sf,sr,ce,cd,rs,si,so,se,md,me */
    /* May have to assume internal VT100 terminal */
    decodetermcap(stat,s);
#if BSDunix || SYS5
    resetWinSize(&ROWMAX,&COLMAX);     /* Use SIG to find window size parameters */
#endif
    return;    /* Success! */
}
#endif

#if !MSDOS
char *setkeydefault() {
     switch(KEYMODE){
     case 1: return SETEMACS;
     case 3: return SETWDPERF;
     default:
     case 2: return SETWDSTAR;
     }
}

void setMode() {
char *keyfile;
char *setkeydefault();
    keyfile=setkeydefault();
    loadkeymap(keyfile);
    safecore(&INIFILE,keyfile); /* Save last keyfile name */
    sorttable();
}

setEmacs(){ KEYMODE=1; setMode(); EMACSMODE=TABINSERT=1;}

setWdstar(){ KEYMODE=2; setMode(); EMACSMODE=TABINSERT=0;}

setWdperf(){ KEYMODE=3; setMode(); EMACSMODE=TABINSERT=1;}
#endif

/* Load key files to set up user keys */
void loadkeys(argc,argv) int argc; char **argv; {
char *p,*q;
static char *keyfile="";
int i,flag;
char s[128];
char *ffname();
    /* should be able to pass keyboard type on command line */
    /* (-key:xxx -set:xxxx -term:xxxxx). Use environment for defaults. */

    flag=0;
/*
 * Load up key files.
 */
   for(i=1;i<argc;++i){ /* command line first choice for -key: */
     if(*(q=fpick("-k",argv[i]))!=0) flag += loadkeymap(keyfile=q);
   }
   if(!flag){   /* command line eliminates environment */
    /* scan environment for -key: */
    if((p=getenv("PIEINIT"))==(char *)0) p="";
     while(p[0]){
       if(*(q=fpick("-k",p=sob(p)))!=0){
         grabtoken(s,q);
         if(loadkeymap(s)){keyfile=s; flag=1;}
       }
       p=fnb(p);
     }
   }
#if TERMCAP
   { extern char *setkeydefault();
   if(!flag) {   /* No keyfile found, choose system default */
     keyfile=setkeydefault();
     if(KEYMODE==2) {flag=1; keyfile="Internal WS Keys";} else
     flag = loadkeymap(keyfile);
   }/* end if */
   }
#endif
   if(keyfile==0 || keyfile[0]==0) keyfile="Internal WS keys";
   safecore(&INIFILE,keyfile);            /* Save last keyfile name */
   loadkeymap(USERFILE);                  /* Load pi.set or ~/.pirc */
   loadkeymap(ffname(PIMACROS));          /* Load macros from current dir */
   /* Sort table before use! */
   /* Call sorttable() after this routine. */
}

/* get a key sequence definition */
int getkeyseq(prompt,buf) char *prompt,*buf; {
char d[128];
char *p;
int x;
  bell();
  macabort();
  p=buf;
  ssprintf(d,"%s: ",prompt);
  while(1){
    banner(d);
    x=getbyte();
    if(x=='\r') break;
      if(x=='P'-'@') *p++ = getbyte();
        else
      if((x==127 || x==8) && p>buf) --p;
        else
      *p++ =x;
    p[0]=0;
    /* translation printed if s[0]!=' ' */
    printseq("End with RETURN: ",buf,d);
  }
  if(p==buf) return 0;
  return 1;
}

int getkeyslot(){
int i;
  for(i=0;i<MAXKEYS;++i) {
    if(ourkeys[i].value== -1) return i;
  }
  return -1;
}

/* Find a keyslot matching macro text in macbuf[], else an empty slot */
/* Return -1 on error. */
int getmacroslot(slot) int *slot; {
int j;
extern int MACNUM,*macromax;
for(j=1;j<=MACNUM;++j){
  if(macromax[j]==0) continue;
  if(strcmp(macbuf,macrobuf[j])==0){slot[0]=j; return 0;}
}
for(j=1;j<=MACNUM;++j){
  if(macromax[j]==0){slot[0]=j; return 1;}
}
return -1;      /* return -1 on error */
}

/* Print key alias list to string d[] */
int lastidentifykey;

int identifykey(key,d) unsigned key; char *d; {
int i,j; char *endof();
  strcpy(d,"Keymap");
  lastidentifykey=-1;
  for(j=i=0;i<MAXKEYS;++i){
    if(ourkeys[i].value == key) {
        printseq(":",ourkeys[i].seq,endof(d)); ++j; /* translation printed */
        lastidentifykey=i;
    }
  }
  return j;
}

/* kill duplicates in identify key string */
#if MSDOS
void killduplicates(){}
#endif
#if !MSDOS
void killduplicates(s) char *s; {
char *strchr(),*strstr(),*skipto(); char *p,*r;  char t[128]; int i;
  r=skipto(":",s);
  while(r[0]){
      p=strchr(r+1,':');
      if(p==(char *)0) break;
      p[0]=0; strcpy(t,r); p[0]=':';
      i=strlen(t);
      while((p=strstr(r+1,t))!=(char *)0) fillchr(p,' ',i);
      r += i;
  }

  p=s;
  while(p[0]){ /* remove blanks */
    s[0]=p[0];
    ++p;
    if(s[0]!= ' ') ++s;
  }
  s[0]=0;
}
#endif

/* Insert key sequence and copy the current macro. Then sort the table */
void copymacro(){
char buf[256];
  if(getkeyseq("Macro key",buf)) keyinsert(buf,1,0);
}

int addkey(value) int value; { /* Attach a key to an existing command */
char buf[256];
  if(value){
    if(getkeyseq("New key",buf)){
      return(keyinsert(buf,0,value)); /* return -1 on failure */
    }
  }
  return -1;
}


/* Remove a key sequence from the key table. Does not remove functions. */
int killkey(seq) char *seq; {
int i,x,xx;
char buf[256];
extern char *endof();
  xx=keyconflict(seq,&x,&i,sysv);  /* xx==0 for no conflict */
  if(xx == -3){ /* one key definition exists */
    strcpy(buf,ourkeys[i].seq);
    goto killkey;
  }
  if(xx != -4) goto err;
  bell();
  strcpy(sysv,"Delete which ");
  identifykey(x,endof(sysv));
  if(!getkeyseq(sysv,buf)) goto errminus1;
killkey:
  return(keyinsert(buf,2,x)); /* return -1 on failure */
err:
  bellpressany("Some error in killkey");
errminus1:
  return -1;
}

void showkeys(){
  getkeyseq(" Press keys",sysv);
}

int isinside(s,t) char *s,*t; { /* 0=Ok string, 1=s inside t, 2=s equals t */
int xx;
  xx=0;
  if(strncmp(s,t,strlen(t))==0) ++xx;
  if(strncmp(t,s,strlen(s))==0) ++xx;
  return xx;
}

static void nulloutkey(i) int i; {
/* Macro keys have extra text elsewhere, in macbuf[] */
/* Tries to do memory freeup, but some key sequences are compiled in */
#if MSDOS
  freeup(ourkeys[i].seq);       /* What will happen with this? */
#else                           /* Macro keys not compiled in */
  if(ourkeys[i].value>MACROKEY) freeup(ourkeys[i].seq);
#endif
  ourkeys[i].value= -1;
  ourkeys[i].seq="";             /* Nullstring is important, because freeup */
                                /* will not try to free it. */
}

int keyinsert(buf,kind,keyvalue) /* Returns index into ourkeys[] or -1 on fail */
char *buf;    /* chars in key sequence */
int kind;     /* Regular (0) or Macro (1) or Kill (2) key possible */
int keyvalue; /* Key code to insert for new key sequence, kind==0 */
{
char d[256],dd[256];
int i,j,x,xx,flag,slot;
char **q;
char *macrosfile();
top:
  xx=keyconflict(buf,&x,&i,d);  /* xx==0 for no conflict */
    /* char  *buf;  chars in key sequence */
    /*     int *x;  Keycode */
    /*     int *i;  Index i of key match or new key slot */
    /*    char *d;  Error message */
  strcpy(dd,"Can't define key. "); strcat(dd,d);
  if(kind==0){  /* Else, regular key slot request, kind==0 */
    if(xx < 0 || xx >0){
      /* xx == -4 */ /* "Can't define key. Key sequence in use" */
      /* xx == -3 */ /* "Can't define key. Key sequence in use" */
      /* xx == -1 */ /* "Can't define key. Key table full" */
      /* xx == -2 */ /* "Can't define key. Key conflict" */
      goto quit;
    }
    /* if(xx == 0) */
    x=keyvalue;
    ourkeys[i].value=x; safecore(&(ourkeys[i].seq),buf);
    sorttable();
    macrosfile(x,buf,2); /* document added key in PI.SET */
    goto sortexit;
  }
  if(kind==1){ /* Macro key slot request, kind==1 */
    /* Use old macro value or else hunt up a new one */
    flag=((xx == -3 || xx == -4) &&(x<MACROKEY)); /* keycode is not a macro */
    flag=(flag || xx == -1 || xx == -2);
    if(flag){
      /* xx == -1 */ /* "Can't define key. Key table full" */
      /* xx == -2 */ /* "Can't define key. Key conflict" */
      /* xx == -3 */ /* "Can't define key. Key sequence in use" */
      /* xx == -4 */ /* "Can't define key. Key sequence in use" */
      goto quit;
    }
    if(xx == -4){
      /* Then x is an active macro keycode with keys already assigned */
      /* If xx == -4, then one other key sequence is mapped. */
      /* It is a bad idea to redefine the other key; quit if xx == -4 */
      goto quit;
    }
    if(xx == -3){
      /* If xx == -3, then just this key sequence is mapped. */
      /* In this case (xx == -3) the request should delete the key */
      /* and then add it using the new macro text. The old macro text */
      /* will be saved in the file PI.SET; the new macro text */
      /* is located in the default macro buffer, macbuf[]. */

      j=x-MACROKEY;   /* Index into the macrobuf[] array */
      if(macromax[j]!=0){ /* Something in the buffer? */
        if(strcmp(macbuf,macrobuf[j])!=0){
          freeup(macrobuf[j]);  /* Throw away any existing macro text */
          macromax[j]=0;        /* to free up space for writing */
        } else { /* macro text and keymap already there */
          goto sortexit;
        }
      }
      /* j=Index into array macrobuf[], x=keycode for the macro */
      goto storemacro;
    }
    if(xx == 0){  /* Use new unused key slot required, key unmatched . */
    /* Store macro text in next macro slot numbered 1..MACNUM */
    /* But if macro text already exists, then just add a key to it */
    /*     int *i;  Index i of the new key slot */
      switch(getmacroslot(&j)){
      case 0:  /* Macro text matches macrobuf[j] */
        x=MACROKEY+j;
        goto storemacro1;
      case 1:  /* Macro text unmatched, got new slot j */
        x=MACROKEY+j;
        break;
      case -1:
      default:
        strcat(dd,"No macro slots");
        goto quit;
      }
      /* j=Index into array macrobuf[], x=keycode for the macro */

storemacro:
      /* A macro has a keycode 15XXX, a key sequence, and its TEXT */
      /* Other functions just have the keycode and key sequence */

      q=(char **) &macrobuf[j];
      macbuf[macmax]=0; /* terminate source string for safety */
      if(safecore(q,macbuf)){ /* safecore terminates the copied string */
         macromax[j]=macmax;
storemacro1:
         ourkeys[i].value=x;
         safecore(&(ourkeys[i].seq),buf);
         sorttable();
         macrosfile(x,buf,2);  /* rewrite the new macro into PI.SET */
      }
      goto sortexit;
    }
    goto quitminus1;
  }
  if(kind==2){ /* Delete the keymap from the table */
    if(xx == -3 || xx == -4){
      /* if(x != keyvalue) then there might be an error ... */
      /* don't know how to process this error, yet. */
      if(x != keyvalue){
        bell(); if(!getyesno("Kill this key")) goto quitminus1;
      }
      macrosfile(x,buf,1);  /* Document killed key in PI.SET */
      nulloutkey(i);
      sorttable();
      goto sortexit;
    }
    if(xx == 0 || xx == -1 || xx == -2){
      strcpy(dd,"Can't kill key. Key not found");
      goto quit;
    }
  }
  goto quit;
sortexit:
  return 0;
quit:
  bellpressany(dd);
quitminus1:
  return -1;  /* -1 on fail */
}

/*
Test for buf[] equal to a current key sequence
Returns:  0  Not found, new key slot available at index i
         -1  Not found and No new slot available for a key
         -2  Key conflict, substring error
         -3  It matches a definition. Keycode match has one definition.
         -4  It matches a definition. Keycode match has two definitions.
 key[0]     == keycode, on a match
 slot[0]    == Which array slot produced match, or else new slot
 err        == error message
*/
int keyconflict(buf,key,slot,err)
char  *buf;             /* chars in key sequence */
int *key;               /* Keycode, on a match */
int *slot;              /* Index i of key match or new key slot */
char *err;              /* Error message */
{
int i,j,x,y,xx;
char s[256];
char *p; int flag;
  for(y=flag=i=0;i<MAXKEYS;++i) {
    if((y=ourkeys[i].value)== -1) continue;
    xx=isinside(ourkeys[i].seq,buf);
    /* isinside(s,t): 0=mismatched, 1=s inside t or reverse , 2=s equals t */
    if(xx){
      /* */
      /* Count (x) how many keys use function code y */
      /* x==1 && xx==1 ==> key conflict, return -2; */
      /* x >1 && xx==1 ==> key conflict, return -2 */
      /* x==1 && xx!=1 ==> unique key in use, return -3 */
      /* x >1 && xx!=1 ==> key has two defs, return -4 */
      /* */
      x=identifykey(y,s);
      flag = (xx==1 ? -2 : (x==1 ? -3 : -4));
      break;
    } /* end if(xx) */
  } /* end for() */
  slot[0]=i;            /* Index of match, if any */
  key[0]=y;             /* keycode of match, if any */
  err[0]=0;
  switch(flag){
  case  0:        /*   0  Not found, see if new key slot is available */
        if((slot[0]=getkeyslot()) == -1) flag = -1;
        else break;
  case -1:        /*  -1  No new slot available for a key */
        strcpy(err,"Key table full"); break;
  case -2:        /*  -2  Key conflict, substring error */
        p="Key conflict"; goto keys;
  case -3:        /*  -3  Keycode has one key definition and it matches */
  case -4:        /*  -4  Keycode has two key definitions. One matches. */
        p="Key sequence in use";
keys:   ssprintf(err,"%s [%u]. %s",p,y,s);
        break;
  }
quit:
  return flag;
}

int readkeymap(){
  bell();
  if(getfname("Key map file: ",sysv)){
    loadkeymap(sysv); sorttable();
    return 1;
  } else return 0;
}


/* dump keys to a file. Returns 1=worked, 0=failed */
int pidumpkey()
{
FILEP fp; FILEP fi;
unsigned target;
int i,j,flag;
char *keyseq;
char buff[512];
LLONG atoiLL();
char *endof();

  flag=0;
  fi=fopenbread(KEYMAP);
  if(fi!=(FILEP)0){
    fp=fopenbwrite(DUMPKEY);
    if(fp!=(FILEP)0){
      while(fgetln(buff,fi)>0){
        fputsline(buff,fp);
        if(buff[0]==';' || buff[0]=='#') continue;
        target=(int)atoiLL(buff);
        if(target==0) continue;
        for(i=0;i<MAXKEYS;++i){
          if(target==ourkeys[i].value){ /* termcap sequence printed */
            printseq(" key=",keyseq=ourkeys[i].seq,buff);
            fputsline(buff,fp);
            printseq("#decoded=",keyseq,buff);  /* translation printed */
            if(!isSeqMatch(keyseq,buff+9)) fputsline(buff,fp);
          }
        }/* end for() */
      }/* end while */
      macrodump(fp);
      closedfile(fp);
      flag=1;
    }
    closedfile(fi);
  }

#if 0           /* DEBUG Set to 1 to dump all keys to "loadset.key" */
/* Write out the keys as currently sorted */

  {int i; char *endof();
  fp=fopenbwrite("loadset.key");
   for(i=0;i<MAXKEYS;++i){
        buff[0]='#';  buff[1]=0;
        target=ourkeys[i].value;
        ssprintf(buff,"%u #",target);
        if(target != (unsigned)-1) identifykey(target,endof(buff));
        fputsline(buff,fp);
        printseq(" key=",ourkeys[i].seq,buff); /* termcap sequence printed */
         fputsline(buff,fp);
   }
   closedfile(fp);
  }
#endif          /* DEBUG */

  return (flag);
}

int pimacrodump(){
FILEP fp;
  fp=fopenbwrite(PIMACROS);
  if(fp!=(FILEP)0){
    macrodump(fp);
    closedfile(fp);
    return 1;
  }
  return 0;
}

int isSeqMatch(keyseq,decoded)
char *keyseq;   /* Raw key sequence */
char *decoded;  /* decoded sequence */
{
char buf[512];
  printseq(" ",keyseq,buf);
  /* printseq emits text " " */
  if(strcmp(buf+1,decoded)==0) return 1;
  return 0;
}

void macrodump(fp) FILEP fp; {
char buff[512];
char s[128];
int i,j,k,keycode;
char *keyseq,*macseq;
extern char *endof();
  for(k=0,j=1;j<=MACNUM;++j){
    if(macromax[j]==0) continue;  /* No macro defined */
    keycode=j+MACROKEY;
    macseq=macrobuf[j];

    getMacroTitle(keycode,s);
    ssprintf(buff,"%u %s [MACROKEY+%u]",keycode,s,j);
    fputsline(buff,fp);
    printseq(" macro=",macseq,buff);  /* termcap sequence printed */
    fputsline(buff,fp);
    printseq("#decoded=",macseq,buff); /* translation printed */
    if(!isSeqMatch(macseq,buff+9)) fputsline(buff,fp);

    for(i=0;i<MAXKEYS;++i){
      if(ourkeys[i].value != keycode) continue;
      ++k;                              /* Count keys assigned */
      keyseq=ourkeys[i].seq;
      printseq(" key=",keyseq,buff); /* termcap sequence printed */
      fputsline(buff,fp);
      printseq("#decoded=",keyseq,buff); /* translation printed */
      if(!isSeqMatch(keyseq,buff+9)) fputsline(buff,fp);
    }
    strcpy(buff,"#Multiple ");
    if(identifykey(keycode,endof(buff))>1) fputsline(buff,fp);
  }
  ssprintf(buff,"# Defined %d key sequences [MAXKEYS=%d,MACNUM=%d]",k,MAXKEYS,MACNUM);
  fputsline(buff,fp);
}


int moveFileToPIbackups(src,dest)
char *src;              /* Existing file to be moved */
char *dest;             /* filename for PIbackups directory */
{
char buff[MAXFNAME];
char *ffname();
int n=0;
    if(BACKUPS){
      getbackupfname(buff);     /* $HOME\PIbackups or .\PIbackups */
#if MSDOS
      mkdir(buff);              /* Usually this does nothing */
#else
      mkdir(buff,0777);         /* Usually this does nothing */
#endif
      addslash(buff);
      strcat(buff,ffname(dest));
      maketempfilename(buff);   /* New name in subdirectory */
      n=rename(src,buff);      /* Move old file to directory */
      /* All of these can fail, no space, CD rom, protection error */
    }
    return n;
}

char *macrosfile(keycode,buf,flag)
int keycode;    /* keycode, >15000 a macro keycode */
char *buf;      /* key sequence */
int flag;       /* 0=remake file, 1=document kill key, 2=document add key */
{
static int test=0;
FILEP fp;
FILEP fi;
char s[MAXFNAME];
char buff[512];
extern LLONG atoiLL();
  if(!test){
     test=1;
     queryLocalPiSet();
  }
  strcpy(s,USERFILE);    /* Make a temp name for PI.SET file */
  maketempfilename(s);
  /* Insert killed key or added key info into temp PI.SET */
  documentAddKill(keycode,buf,flag,s);
  /* Move old PI.SET to a safe place */
  if(s[0]) {
     moveFileToPIbackups(USERFILE,USERFILE);
     rename(s,USERFILE);
     /* If this fails, then pi.set is unchanged */
     /* and the new one has a backup name. */
  }
  return USERFILE;
}

/* Test a PI.SET file for a comment line */
/* If a line starts with ; or # then it is a comment line */
int isacommentline(buff)
char *buff; {
  return (strchr(";#",buff[0])!=0) ? 1 : 0;
}

/* Test a PI.SET file for a header line */
int isaheaderline(buff,keycode)
char *buff;
int *keycode;
{
  return ((keycode[0]=(int)atoiLL(buff)) != 0);
}

static void putKeyLines(keyseq,token,fo)
char *keyseq;  /* Key sequence */
char *token;   /* Either " key=" or " kill=" */
FILEP fo;      /* Output file */
{
char bigbuf[512];
  printseq(token,keyseq,bigbuf);
  fputsline(bigbuf,fo);
  printseq("#decoded=",keyseq,bigbuf);
  if(!isSeqMatch(keyseq,bigbuf+9)) fputsline(bigbuf,fo);
}

static void putMacroLines(macrotext,fo)
char *macrotext;  /* macro text */
FILEP fo;      /* Output file */
{
char bigbuf[512];
  printseq(" macro=",macrotext,bigbuf);  /* termcap sequence printed */
  fputsline(bigbuf,fo);
  printseq("#decoded=",macrotext,bigbuf);  /* termcap sequence printed */
  if(!isSeqMatch(macrotext,bigbuf+9)) fputsline(bigbuf,fo);
}

void getMacroTitle(keycode,s)
int keycode;
char *s;     /* Buffer of size 512 or larger */
{
char buff[512];
FILEP fi;
int j;
extern char *fnb(),*sob();
extern char *USERFILE;
  if((fi=fopenbread(USERFILE)) == (FILEP)0) goto notitle;
  while(fgetln(buff,fi)>0){
     j=(int)atoiLL(buff)-MACROKEY;
     if(1 <= j && j <= MACNUM) {
       if(j+MACROKEY != keycode) continue;
       strTruncate(buff,'[');
       strncpy(s,sob(fnb(buff)),MACROTRUNC);
       if(*sob(s)!=0) goto quit;  /* Got a title */
       s[0]=0;
     }
  }
notitle:
  strcpy(s,"Untitled");
quit:
  if(fi) closedfile(fi);
}

void getToolTitle(keycode,title)
int keycode; char *title;
{
char buff[512];
int j,KEYcode;
FILEP fi;
char *p;
extern char *strstr();
extern char *USERFILE;
  title[0]=0;
  if((fi=fopenbread(USERFILE)) == (FILEP)0){
     banner("can't open usefile "); ePrint(USERFILE); pressany("");
     return;
  }
  while(fgetln(buff,fi)>0){
     if( isacommentline(buff) ) continue;
     if(isaheaderline(buff,&j) != 0) KEYcode=j;
     if(KEYcode==keycode){
       if((p=strstr(buff,"tool="))!=0){
         strcpy(title,p+5); title[MACROTRUNC]=0;
         break;
       }
     }
  }
  closedfile(fi);
}




/* Return 0 for no title entered */
int newmacrotitle(keycode) int keycode;
{
   if(keycode<=MACROKEY) return 0;
   return getTitle(keycode,0);
}

/* Return 0 for no title entered */
int newtooltitle(keycode) int keycode;
{
   if(keycode<=0) return 0;
   return getTitle(keycode,1);
}

int getTitle(keycode,titlekind)
int keycode,titlekind; /* titlekind=0 for macro, =1 for tool */
{
char title[2+MACROTRUNC];
char buff[512];
char tmp[128];
char s[MAXFNAME];
char *p;
int gotit,KEYcode;
FILEP fo;
FILEP fi;
extern LLONG atoiLL();
   if(titlekind==0) getMacroTitle(keycode,title);
   if(titlekind==1) getToolTitle(keycode,title);
   strcpy(buff,title);
   ggetstr("[40 char] New title:",buff);
   if(strcmp(title,buff)==0) return 0;                    /* No title entered */
   strncpy(title,buff,MACROTRUNC);
   if((fi=fopenbread(USERFILE))==0) goto fileError;
   gotit=0;
   strcpy(s,USERFILE);    /* Make a temp name for PI.SET file */
   maketempfilename(s);
   if((fo=fopenbwrite(s))==0) goto writeError;
   while(fgetln(buff,fi)>0){
     if( isaheaderline(buff,&KEYcode)!=0 && KEYcode==keycode){
       gotit=1;
       if(titlekind==0) /* Macro header */
         ssprintf(buff,"%u %s [MACROKEY+%d]",keycode,title,keycode-MACROKEY);
       if(titlekind==1){  /* Tools header */
         fputsline(buff,fo); goto getTools;
       }
     }
loop:
     fputsline(buff,fo);
loop1:;
   }
goto outerexit;
getTools:
   while(fgetln(buff,fi)>0){ /* Goto outerexit on break */
     if( isacommentline(buff) ) goto write1;
     if((p=strstr(buff,"tool="))!=0){
        strcpy(p+5,title); gotit=2;
        if(title[0]) goto loop;
        goto loop1;
     }
     if(isaheaderline(buff,&KEYcode)!=0 && KEYcode!=keycode){
loop2:
        ssprintf(tmp," tool=%s",title);
        if(title[0]) fputsline(tmp,fo); gotit=2;
        goto loop;
     }
write1:
     fputsline(buff,fo);
   }
   if(gotit != 2) goto loop2;
outerexit:
   closedfile(fi);
   closedfile(fo);
   if(!gotit){
     remove(s);
   } else {    /* Move old PI.SET to a safe place */
     moveFileToPIbackups(USERFILE,USERFILE);
     rename(s,USERFILE);
     /* If this fails, then pi.set is unchanged */
     /* and the new one has a backup name. */
  }
  goto err1;
writeError:
   closedfile(fi);
   ssprintf(buff,"Can't write %s",s);
   goto err;
fileError:
   ssprintf(buff,"%s not found",USERFILE);
err:
   bellpressany(buff);
err1:
   return 1;
}

void documentAddKill(keycode,buf,flag,s)
int keycode; /* keycode, >15000 is a macro code */
char *buf;   /* key sequence */
int flag;    /* 1=kill, 2=add */
             /* e.g. " kill=" plus "\E[1~" becomes " kill=\E[1~" */
char *s;     /* Temp file name for writing new copy of PI.SET */
             /* On return, s[0]==0 if something failed */
{
char bigbuf[512];
char buff[512];
FILEP fi;
FILEP fo;
int j;
int macrocode;
int macroDone;
int keyDone;
int killDone;
int KEYcode;
char *p;
extern LLONG atoiLL();
extern char *endof();

  macrocode=keycode-MACROKEY;
  if(macrocode<1 || macrocode > MACNUM) macrocode=0;
  keyDone=killDone=macroDone=0;

  fi=fopenbread(USERFILE);

  fo=fopenbwrite(s);
  if(fo==(FILEP)0){
    if(fi!=(FILEP)0) closedfile(fi);
    ssprintf(buff,"Write error on %s",USERFILE);
    bellpressany(buff); return;
  }

  if(fi==(FILEP)0) goto notdocumented;

  /* Read and write lines until the keycode matches on a header */
top:
  while(fgetln(buff,fi)>0){
    fputsline(buff,fo);
    if( isaheaderline(buff,&KEYcode) ){
      if(KEYcode==keycode) goto keymap;
    }
  }
  closedfile(fi); fi=0; /* No header matched */
  goto notdocumented;     /* Then file at end, maybe no hit. */

keymap:
  /* Examine lines after header until next header */
  while(fgetln(buff,fi)>0){
    if( isacommentline(buff) ) goto write;
    if(isaheaderline(buff,&KEYcode)){
      if(keycode == KEYcode) continue;       /* Duplicate header, skip it */
      if(flag==2 && keyDone==0){
        putKeyLines(buf," key=",fo);
        keyDone=1;
      }
      if(flag==1 && killDone==0){
        putKeyLines(buf," kill=",fo);
        killDone=1;
      }
    if(flag==2 && macrocode && macroDone==0){
      putMacroLines(macrobuf[macrocode],fo);
    }
      fputsline(buff,fo);                    /* Done with this block */
      goto top;
    }
    /* The buffer does not contain a header nor a comment */
    /* Test for "key=" or "kill=" */
    if((p=strstr(buff,"key=")) != 0){
      tcap_to_string(p+4,bigbuf);
      if(strcmp(buf,bigbuf)==0){
        if(flag==1) killDone=1;
        continue; /* Don't document this key */
      }
    }
    if((p=strstr(buff,"kill=")) != 0){
      tcap_to_string(p+5,bigbuf);
      if(strcmp(buf,bigbuf)==0){
         continue; /* Don't document this key */
      }
    }
    if(!macrocode) goto write;

macrotest:
    /* If its a macro header, then test for "macro=" */
    if(macromax[macrocode]==0) goto write;  /* No macro in memory */
    if((p=strstr(buff,"macro=")) == 0) goto write;
    tcap_to_string(p+6,bigbuf);
    /* This macro might be the same as the one in memory. */
    if(strcmp(macrobuf[macrocode],bigbuf) == 0){
      if(flag==2) goto write0;
      goto write1;
    }
    /* No, it's a different macro definition */
    buff[0]='#';  /* Comment out this macro, in case it's needed */
    fputsline(buff,fo);
write0:
    if(macroDone) continue;
    printseq(" macro=",macrobuf[macrocode],buff);  /* termcap sequence printed */
write1:
    macroDone=1;
write:
    fputsline(buff,fo);
  } /*  while(fgetln(buff,fi)>0) */

notdocumented:
  if(getHeaderLine(keycode,buff)){ /* Got a header */
    if(flag==1 && killDone==0){
      fputsline(buff,fo);
      putKeyLines(buf," kill=",fo);
    }
    if(flag==2 && keyDone==0){
      fputsline(buff,fo);
      putKeyLines(buf," key=",fo);
    }
    if(macrocode && macroDone==0){
      putMacroLines(macrobuf[macrocode],fo);
    }
  }
alldone:
  closedfile(fo);
  if(fi) closedfile(fi);   /* Can be that fi==0 */
}

int getHeaderLine(keycode,buff)
int keycode;
char *buff;
{
char s[2+MACROTRUNC];
int macrocode;
FILEP fi;
int gotit;
int KEYcode;
  gotit=0;
  macrocode=keycode-MACROKEY;
  if(macrocode<1 || macrocode>MACNUM) macrocode=0;
  if(macrocode){
    getMacroTitle(keycode,s);
    ssprintf(buff,"%u %s [MACROKEY+%u]",keycode,s,macrocode);
    gotit=1;
  } else {
    /* Find the first match in the KEYMAP file to the given keycode */
    fi=fopenbread(KEYMAP);
    if(fi!=(FILEP)0){
       while(fgetln(buff,fi)>0){
         if( isaheaderline(buff,&KEYcode) ) {
           if(keycode != KEYcode) continue;
           gotit=1;  break;
         }
       }
       closedfile(fi);
    }
  }
  return gotit;
}
/* Print termcap key sequence if msg[0]==' ' else translate function keys */
void printseq(msg,p,s) char *msg; char *p; char *s; {
int x,i,j,wasoctal; char *endof();
  strcpy(s,msg); i=strlen(s);
  wasoctal=0;
  while(p[0]){
    x=p[0]; ++p;
    if(0<x && x<' ' && x!='\33') {s[i++]='^'; s[i++]=x+'@';}
    else if(' '<=x && x <= '~' && !(wasoctal && '0'<=x && x <= '9')){
      s[i++]=x;
      if(x=='\\') s[i++]=x;
    }
    else if(msg[0] != ' ' && (j=functionkey(p-1,s+i))!=0) {
      p += j-1;         /* skipover used chars */
      i = strlen(s);    /* skip to end of string */
    }
    else if(x == '\33'){s[i++]='\\'; s[i++]='E';}
    else {
     wasoctal=1;
     s[i++]='\\';
     s[i++]=((x&0300) >> 6)+'0';
     s[i++]=((x&0070) >> 3)+'0';
     s[i++]=((x&0007))+'0';
     continue;
    }
    wasoctal=0;
  }
  s[i]=0;
}

int istrchr(s,x) char *s; int x;{
char *p; char *strchr();
  if((p=strchr(s,x))==(char *)0) x = 0; else x = 1+(int)(p-s);
  return x;
}

#if MSDOS
int functionkey(pp,s)
char *pp;  /* source string address */
char *s;   /* dest string */
{
char v[32];
char *p; int i,j,x;
char *endof(),*lltoa();
  x=pp[0];      /* leadin char of escape sequence? */
  if(x == '\200'){ /* IBM function keys */
    x=pp[1];    /* look for match to second char in string */
    p="";
    if((i=istrchr("\017TUVWXYZ[\\]\207\210",x))!=0)  /* shift TAB, F1 to F12 */
      p="Shift-";
    else
    if((i=istrchr("\224^_`abcdefg\211\212",x))!=0)   /* ctrl TAB, F1 to F12 */
      p="Ctrl-";
    else
    if((i=istrchr("\225hijklmnopq\213\214",x))!=0)   /* alt TAB, F1 to F12 */
      p="Alt-";
    else
    if((i=istrchr("\001;<=>?@ABCD\205\206",x))!=0)   /* placeholder, F1 to F12 */
      p="";

    if(i){
copynumber:
      lltoa((LLONG)(i-1),v);
      if(i==1) v[0]=0;
      ssprintf(s,"{%s%s%s}",p,(v[0]? "F" : "Tab"),v);
      goto quit;
    }
    /* alt 0 to 9 */
    if((i=istrchr("\201\170\171\172\173\174\175\176\177\200",x))!=0){
      j='0'+i-1;
copyALT:
      ssprintf(s,"{Alt-%c}",j);
      goto quit;
    }
    if((i=istrchr("\036\060\056\040\022\041\042\043\027\044\045\046\062\061\030\031\020\023\037\024\026\057\021\055\025\054",x))!=0){   /* alt a to z */
      j='A'+i-1; goto copyALT;
    }
    if((i=istrchr("\202\203\053\032\033\047\050\063\064\065",x))!=0){
    /* alt -=\[];',./ */
      {
      char *s0="-=\\[];',./";
      j=s0[i-1]; goto copyALT;
      }
    }
    {   /* special IBM function keys */
#if 0
    KEY    norm ctrl
    Home   71   119
    Up     72   141
    PgUp   73   132
    Left   75   115
    Right  77   116
    End    79   117
    Down   80   145
    PgDn   81   118
    Ins    82   146
    Del    83   147
#endif

      static char *ibmf[]={ "Home", "Up", "PgUp", "Left", "Right",
                            "End", "Down", "PgDn", "Ins", "Del" };
      if((i=istrchr("GHIKMOPQRS",x))!=0) p="";
      else
      if((i=istrchr("w\15\04stu\21v\22\23",x))!=0) p="Ctrl-";
      if(i){
        ssprintf(s,"{%s%s}",p,ibmf[i-1]); goto quit;
      }
    }
  }/* end if(x == '\200') */

return 0;       /* tell caller it failed */
quit:
return 2;       /* tell caller it worked */
}
#endif
#if !MSDOS
struct alias {char *string; char *key;};
static
struct alias sunterm[]={
{"R8}:{UP",    "A"},
{"R14}:{DOWN",  "B"},
{"R10}:{LEFT",  "D"},
{"R12}:{RIGHT", "C"},
{"L1",  "192z"},
{"L2}:{F12",  "193z"},
{"L3}:{F13",  "194z"},
{"L4}:{F14",  "195z"},
{"L5}:{HELP",  "196z"},
{"L6}:{DO",  "197z"},
{"L7}:{F17",  "198z"},
{"L8}:{F18",  "199z"},
{"L9}:{F19",  "200z"},
{"L10}:{F20", "201z"},
{"F1",  "224z"},
{"F2",  "225z"},
{"F3",  "226z"},
{"F4",  "227z"},
{"F5",  "228z"},
{"F6",  "229z"},
{"F7",  "230z"},
{"F8",  "231z"},
{"F9",  "232z"},
{"F10", "233z"},
{"R1",  "208z"},
{"R2",  "209z"},
{"R3",  "210z"},
{"R4",  "211z"},
{"R5",  "212z"},
{"R6",  "213z"},
{"R7",  "214z"},
{"R8",  "215z"},
{"R9",  "216z"},
{"R10", "217z"},
{"R11", "218z"},
{"R12", "219z"},
{"R13", "220z"},
{"R14", "221z"},
{"R15", "222z"},
/* SUN TERMINAL KEY CODES "xterm +sf" or "xterm" */
{"L1",  "23~"},
{"L2",  "24~"},
{"L3",  "25~"},
{"L4",  "26~"},
{"L5",  "28~"},
{"L6",  "29~"},
{"L7",  "31~"},
{"L8",  "32~"},
{"L9",  "33~"},
{"L10", "34~"},
{"F1",  "11~"},
{"F2",  "12~"},
{"F3",  "13~"},
{"F4",  "14~"},
{"F5",  "15~"},
{"F6",  "17~"},
{"F7",  "18~"},
{"F8",  "19~"},
{"F9",  "20~"},
};
#define SUNTERMSIZE sizeof(sunterm)/sizeof(sunterm[0])

/* DEC 3100 KEYBOARD "xterm -sf" */
/* F11 \33 */
static
struct alias decterm[]={
{"FIND",         "1z"},
{"INSERT",       "2z"},
{"REMOVE",       "3z"},
{"SELECT",       "4z"},
{"PREV",         "5z"},
{"NEXT",         "6z"},
/* change to \33OXXXX */
{"R8}:{UP",    "A"},
{"R14}:{DOWN",  "B"},
{"R10}:{LEFT",  "D"},
{"R12}:{RIGHT", "C"},
{"PF1",  "P"},
{"PF2",  "Q"},
{"PF3",  "R"},
{"PF4",  "S"},
{"KP0", "p"},
{"PERIOD",  "n"},
{"ENTER",  "M"},
{"COMMA",  "l"},
{"MINUS",  "m"},
{"KP1", "q"},
{"KP2", "r"},
{"KP3", "s"},
{"KP4", "t"},
{"KP5", "u"},
{"KP6", "v"},
{"KP7", "w"},
{"KP8", "x"},
{"KP9", "y"},
};
#define DECTERMSIZE sizeof(decterm)/sizeof(decterm[0])

int functionkey(pp,s)
char *pp;  /* source string address */
char *s;   /* dest string */
{
char v[32];
char *p; int i,x;
char *endof();
  /* pp[0] == leadin char of escape sequence? */
  if(pp[0] == '\33'){ /* SUN, DEC function keys */
    if(pp[1] == '['){
      for(i=0;i<SUNTERMSIZE;++i){
        if(strcmp(pp+2,sunterm[i].key)==0){
          strcpy(s,"{"); strcat(s,sunterm[i].string); strcat(s,"}");
          return (strlen(pp));
        }
      }
    }
    else
    if(pp[1] == 'O'){
      for(i=0;i<DECTERMSIZE;++i){
        if(strcmp(pp+2,decterm[i].key)==0){
          strcpy(s,"{"); strcat(s,decterm[i].string); strcat(s,"}");
          return (strlen(pp));
        }
      }
    }
  }/* end if(x == '\33') */

return 0;       /* tell caller it failed */
}
#endif

char *
skipto(charset,p) char *charset,*p; {
extern char *strchr();
 while(p[0] && strchr(charset,p[0])==(char *)0) ++p;
 return p;
}

int loadkeymap(keyfile) char *keyfile; {
int target,x;
FILEP fp;
char buff[512]; char *p; char *q;
char *sob(),*endof();
LLONG atoiLL();
struct swapchar {unsigned char old,new;};
extern struct swapchar *accentchars;
extern int NUMACCENT;

  fp=fopenbread(keyfile);
  if(fp!=(FILEP)0){
  /* Do the dirty work of reading a keymap source file */
  /* and changing the key dispatch table. */

    while(fgetln(buff,fp)>=0){
      if(buff[0]==';' || buff[0]=='#') continue;
      x=(int)atoiLL(sob(buff));
      if(x != 0){
        target=x;
      } else {
        if((p=strstr(buff,"key="))!= (char *)0) tblinsert(target,p+4);
        else
        if((p=strstr(buff,"kill=")) != (char *)0) tblkill(target,p+5);
        else
        if((p=strstr(buff,"macro=")) != (char *)0) {
          x=target-MACROKEY;
          if(x>0 && x<=MACNUM){
            tcap_to_string(p+6,buff);
            macromax[x]= safecore(&(macrobuf[x]),buff) ? strlen(buff):0;
            /* safecore terminates the copied string */
          }
        }
        else
        if((p=strstr(buff,"foreign=")) != (char *)0) {
            if(p[8]==0){
              freeup(accentchars); NUMACCENT=0;
            } else {
              tcap_to_string(p+8,buff);
              if(NUMACCENT){strcat(buff,(char *)accentchars);
              freeup(accentchars);}
              NUMACCENT= safecore(&accentchars,buff) ?
                       strlen(buff)/sizeof(struct swapchar) : 0;
            }
        }
        else
        if((p=strstr(buff,"browser=")) != (char *)0) {
            if(p[8]!=0){
              safecore(&BROWSER,p+8);
            }
        }/* else do nothing*/
      }
    }
    closedfile(fp);
    return 1;
  }
  return 0;
}

int loadrc(userfile) char *userfile; {
/* Load user options from resource file, if found. */
#if 0
Here is what we can change on the fly at run time:
; Default resource file "~/.PIrc" for PI editor. Comment lines start with ';'.
; Do not change text in this file - only the numbers after '='!
; No leading white space on any line!
select.emulation=2;   None==0, Emacs==1, WordStar==2, WordPerfect==3. (1=default)
auto.indent.toggle=1; Toggle auto-indent word wrap (1=on,0=off).
hang.indent.toggle=1; Toggle hanging indent mode (1=on,0=off)
word.wrap.toggle=1;   Toggle word wrap at right margin (1=on,0=off).
bit.strip.toggle=0;   Toggle stripping of 128-bit on input file (1=on,0=off).
detab.toggle=1;       Toggle auto-detab of input during scrolling (1=on,0=off).
fold.case.toggle=1;   Toggle case-insensitive search mode (1=on,0=off).
insert.toggle=1;      Toggle insert mode (1=on,0=off).
backup.toggle=1;      Toggle backup (0=off,1=~/PIbackups,2=./PIbackups)
; numerical settings
status.line.toggle=1; Print status line during edit (2=all,1=no numbers,0=off).
screen.tab.stop=8;    Screen tab stop increment for TAB key (8=normal).
file.detab.stop=8;    Tab stop increment for detabbing a file (8=default).
trailing.space=0;     Remove trailing space from lines (0=off, 1=on).
import.type=1;        Import text line delimiter code (1=MSDOS,2=MAC,3=UNIX).
export.type=1;        Export text line delimiter code (1=MSDOS,2=MAC,3=UNIX).
drawing.style=1;      Default line/box drawing style (1 to 5,1=default).
tab.insert.mode=0;    Tab key inserts spaces and/or TAB (1=on,0=off)
search.token.lf=-128; Token in search string used for begin or end line
return.insert.crlf=0; EMACS-like split on RETURN (1=on,0=off)
right.margin=72;      Right margin for word wrap and centering (72=normal).
max.macro.size=256;   Size in chars of a macro recording buffer (256=normal).
available_macros=25;  Maximum number of key slots for macros (25=normal)
; COLOR INFO for IBM-PC
; dark colors: BLACK=0, BLUE=1, GREEN=2, CYAN=3, RED=4, MAGENTA=5,
;              BROWN=6, LIGHTGRAY=7
; light colors: DARKGRAY=8, LIGHTBLUE=9, LIGHTGREEN=10, LIGHTCYAN=11,
;               LIGHTRED=12, LIGHTMAGENTA=13, YELLOW=14, WHITE=15
; Example: foreground=BLACK, background=LIGHTCYAN uses attribute 48
;          because (foreground+(background)*16) - 128 = 48.
;          Warning: Attributes are in the range 0 to 127.
;                   Add 128 to an attribute to make it blink.
other.attributes=7;   Screen attribute for edit screens (7=mono,48=color)
title.attribute=15;   Window title attribute flag (31=boldface,15=mono).
window.attribute=7;  Window text attribute (19=blue /w yellow text,7=mono).
border.attribute=31;  Window border attribute flag (31=boldface,15=mono).
box.characterset=0;   Character set for box border (0=Text,1=vga1,2=vga2,3=iso1,4=iso2)
ansi.mode.unix=1;     Ansi text mode for the window, 1=on, 0=off
tfg.text.fg.color=30; Foreground text color ansi unix (30)
tbg.text.bg.color=46; Background text color ansi unix (46)
mfg.menu.fg.color=37; Foreground menu color ansi unix (37)
mbg.menu.bg.color=44; Background menu color ansi unix (44)
killed.line.max=500;  Maximum killed lines that will be saved (500=normal).
line.max.chars=384;   Maximum allowed line length (160=normal).
most.lines.count=25;  Most lines per screen >= 1+(row-max-index) (40=normal).
cga.noretrace.video=1; Set noretrace mode for CGA video (1=mono/ega/vga).
; Printer functions next six lines
printer.number=0;     set printer 0-6:prn,com1,com2,com3,lpt1,lpt2,lpt3
lpp.printer=55;       lines printed per page, 55 normal (printer)
left.spaces.printer=0; number of spaces printed before line (printer)
top.lines.printer=5;  number of blank lines at top (printer)
bot.lines.printer=5;  number of blank lines at bottom (printer)
formfeed.printer=1;   use formfeed between pages, 1=on,0=off (printer)
; The abort key interrupts macros and stops type-aheads (editor halt).
abort.editor.char=21;  Bailout/abort key is normally ctrl-U (21 decimal)
; Next two handled by TERMCAP - changes declare you an expert! (Win3)
row.max.index=23;     Rows numbered 0..ROWMAX (ROWMAX=23 is normal).
col.max.index=79;     Columns numbered 0..COLMAX (COLMAX=79 is normal).
xmouse.toggle=0;      X-windows mouse for Xterm. 0=off, 1=on.
horizontal.mouse.sens=0; Mouse sensitivity. Normal=8, Off=0. MSDOS
vertical.mouse.sens=0;   Mouse sensitivity. Normal=16, Off=0. MSDOS
chigh.cursor.shape=0;  Set IBM cursor to vga (1) or cga (6) or mono (12).
clow.cursor.shape=15;  Set IBM cursor to vga (13) or cga (7) or mono (13)
xms.maximum.blocks=1024; Block maximum msdos xms (512) or unix (1024)
block.buffer.size=16; Block size msdos [2k->10k] (4) or unix [2k->32k] (16)
string.reserve.blocks=6; Blocks reserved msdos for internal line copying (6)
action.on.startup=1; Menu on initial startup (1) or UnTitled (0).
suspend.on.exit=1;  Suspend PI on SAVE THEN EXIT ctrl-KX.
; other set options
; browser=iexplorer
; foreign=
#endif
int i;
FILEP fp; char *p;
char buff[512];
struct setopts {char *option; int *setting;};
static struct setopts setup[]={
{"sel", /*select-emulation     */   &KEYMODE},
{"aut", /*auto-indent-toggle   */   &INDENAUTO},
{"han", /*hanging indent toggle*/   &HangingIndent},
{"wor", /*word-wrap-toggle     */   &WRDWRP},
{"rig", /*right-margin         */   &margin},
{"bit", /*bit-strip-toggle     */   &BITSTRIP},
{"det", /*detab-toggle         */   &DETABBER},
{"fol", /*fold-case-toggle     */   &FOLDCASE},
{"ins", /*insert-toggle        */   &INSMD},
{"bac", /*backup-toggle        */   &BACKUPS},
{"sta", /*status-line-toggle   */   &DOSTATUS},
{"scr", /*screen-tab-stop      */   &TSTOPS},
{"fil", /*file-detab-stop      */   &STOPS},
{"dra", /*drawing-style        */   &drawstyle},
{"imp", /*import-type          */   &intexttype},
{"exp", /*export-type          */   &outtexttype},
{"max", /*max-macro-size       */   &MAXMB},
{"inf", /*info-max-size        */   &maxinfo},
{"ava", /*available_macros     */   &MACNUM},
{"oth", /*other-attributes     */   &SYSATT},
{"tit", /*title-attribute      */   &titlattr},
{"win", /*window-attribute     */   &windattr},
{"bor", /*border-attribute     */   &bordattr},
{"box", /*box-characterset     */   &boxcharset},
{"lin", /*line-max-chars       */   &MAXLN},
{"mos", /*most-lines-count     */   &MAXROWS},
{"row", /*row-max-index        */   &ROWMAX},
{"col", /*col-max-index        */   &COLMAX},
{"ret", /*return-insert-crlf   */   &EMACSMODE},
{"cga", /*cga-noretrace-video  */   &_egavga},
{"abo", /*abort-editor-char    */   &abortkey},
{"lpp", /*lpp-printer          */   &PRTlpp},
{"lef", /*left-spaces-printer  */   &PRTleft},
{"top", /*top-lines-printer    */   &PRTtop},
{"bot", /*bottom-lines-printer */   &PRTbot},
{"for", /*formfeed-printer     */   &PRTff},
{"tab", /*tab-insert-mode      */   &TABINSERT},
{"sea", /*search-token-lf      */   &zero},
{"kil", /*killed-line-max      */   &MAXKILLER},
{"xms", /*xms-maximum-blocks   */   &MAXXMS},
{"blo", /*block-buffer-size    */   &SIZEKBLOCK},
{"act", /*action-on-startup    */   &ACTIONSTART},
#if BSDunix || SYS5
{"xmo", /*X-windows Xterm mouse*/   &mouseswitch},
{"ans", /*ansi-mode-unix       */   &ansimode},
{"tfg", /*tfg-text-fg-color    */   &textfgcolor},
{"tbg", /*tbg-text-bg-color    */   &textbgcolor},
{"mfg", /*mfg-menu-fg-color    */   &menufgcolor},
{"mbg", /*mbg-menu-bg-color    */   &menubgcolor},
{"tra", /*trailing.space       */   &BLANKSTRIP},
#endif
#if MSDOS  /* Mapping of GREY keys on the keypad to other keycodes */
           /* Value is the msdos keycode replacement for the key */
           /* The code appears in routine ccBIOS (v8.c) */
{"plu", /*plus.grey.keycode    */   &GREYplus},
{"ast", /*asterisk.grey.keycode*/   &GREYast},
{"ent", /*enter.grey.keycode   */   &GREYent},
{"str", /*Mem blocks for str   */   &STASHmin},
{"sus", /*Suspend on ctrl-KX   */   &SUSPENDexit},
{"hor", /*horizontal-mouse-sens*/   &hsens},
{"clo", /*IBM cursor shape CL  */   &cursorCL},
{"chi", /*IBM cursor shape CH  */   &cursorCH},
{"ver", /*vertical-mouse-sens  */   &vsens},
#endif
};
int *x;
LLONG atoiLL();
#define OPSIZE sizeof(setup)/sizeof(setup[0])
#define KEYSIZE 3               /* size of key string in search */

  fp=fopenbread(userfile);
  if(fp!=(FILEP)0){
    while(fgetln(buff,fp) >= 0){
      if(buff[0]==';' || buff[0]=='#') continue;
      for(i=0;i< (OPSIZE);++i){
        if(strncmp(buff,setup[i].option,KEYSIZE)==0){
          p=strchr(buff,'=');
          x=setup[i].setting;
          if(p != (char *)0) x[0] = (int)atoiLL(p+1);
        }
      }/* end for() */
    }/* end while() */
    closedfile(fp);
    return 1;
  }
  return 0;
}

void tblkill(x,buff) unsigned x; char *buff; {
int i;
char buf[128];
  /* Wipe out all references that match buff[] for code x */
  tcap_to_string(buff,buf);
  for(i=0;i<MAXKEYS;++i) {
    if(ourkeys[i].value == -1) continue;
    if(strcmp(buf,"*all")==0) nulloutkey(i);
    else
    if(ourkeys[i].value == (int)x){
      if(buff[0]=='*') nulloutkey(i); /* wildcard match keys to code */
      else
      if(strcmp(ourkeys[i].seq,buf)==0){ nulloutkey(i); return; }
    }
  }
}

void tblinsert(x,buff) unsigned x; char *buff; {
int i,u;
char buf[128]; char s[128]; char *p;
  tcap_to_string(buff,buf);
  /* A key cannot have two actions - check for conflicts */
  for(i=0;i<MAXKEYS;++i) {
    if(ourkeys[i].value == -1) continue;
    u=isinside(ourkeys[i].seq,buf);
    if(u==2 && ourkeys[i].value==x) return;     /* key already coded */
    if(u){ printseq("Bad key ignored: ",buf,s); goto fail; }
  }
  /* Add a new key into table */
  if((i=getkeyslot()) != -1) {
     if(safecore(&(ourkeys[i].seq),buf)) ourkeys[i].value = (int)x;
     return;
  }
  strcpy(s,"Key table full");
fail:
  bellpressany(s);
}

void fputsline(buf,fp) char *buf; FILEP fp; {
  ewrite(fileno(fp),(char far *)buf,strlen(buf));
#if CRLFBOX
  ewrite(fileno(fp),(char far *)"\r\n",strlen("\r\n"));
#else
  ewrite(fileno(fp),(char far *)"\n",strlen("\n"));
#endif
}

/*
 * FGETLN
 *
 * Purpose:
 *              Read in a line of text, binary mode.
 * Returns:
 *              -1 for end of file
 *              Number of chars read, otherwise
 */
int fgetln(buf,fp)
char *buf;
FILEP fp;
{
int i; extern int fgetc(); int x;
  for(i=0;i<511;++i){
    if((x=fgetc(fp)) == -1) goto err;
    buf[i]=x;
    if(x=='\n') break;
  }
  buf[i]=0;
  if(i>0 && buf[i-1]=='\r') buf[--i]=0;
  return (i+1);
err:
  buf[0]=0;
  return -1;
}

#if 0  /* Not used anymore */
int getln(p,m) char *p; int m; {
int c,n;
LONG ccBIOS();

  n = 0;
  while(n<m) {
    c = ccBIOS(0);
    if(c == '\r' || c == '\n') break;
    if(c == 8 || c == 127) {
      if(n) {
        erachar(); --n;
      } else bell();
      continue;
    }
    p[n++] = c;
    p[n] = 0;
    etype(&p[n-1]);
  }
  p[n] = 0;
  return n;
}
#endif

void tcap_to_string(s,buf)   /* convert termcap string to C string */
char *s;                /* termcap-style string, no white space */
unsigned
char *buf;              /* destination string */
{                       /* PI uses termcap strings from ini-files */
int i,x;

  /* remove leading and trailing spaces */
  while(s[0]==' ') ++s;
  i=0;
  while((x=s[0]) != 0){
    ++s;
    if(x=='\\'){
      if('0'<=s[0] && s[0]<='7'){
        x=0;
        while('0'<=s[0] && s[0]<='7') {x = 8*x + (s[0] - '0'); ++s;}
      } else {
        x=s[0];
        ++s;
        switch(x){
        case 'E': x='\033'; break;
        case 'r': x='\r'; break;
        case 'n': x='\n'; break;
        case 't': x='\t'; break;
        case 'b': x='\b'; break;
        case 'f': x='\f'; break;
        }
      }
    } else {
      if(x == '^'){
        x=s[0]; if('a'<=x && x<='z') x=x-'a'+'A';
        x=x-'@';  /* make a control char */
        ++s;
        if(x==0) x=128;
      }
    }
    buf[i++]=x; if(x==0) break;
  }
  buf[i]=0;
}

#if TERMCAP
/* Read in termcap variables into editor system variables */
int decodetermcap(stat,s)
int stat; char *s;
{
/* get co,li,cm,al,dl,cs,sf,sr,ce,cd,rs,is,so,se,md,me,ks,ke,bl */
char *p;
char buf[512];
char *buff;
int x,error;

  error=0;
top:
  if(stat != 1){
    etype("Set TERM and TERMCAP variables to eliminate this message.\r\n");
    etype("Assuming VT100 terminal.\r\n");
#define VT100eebuf \
"vt100|vt100internal:cm=\\E[%i%d;%dH:co#80:li#24:\
al=\\E[L:cd=\\E[J:ce=\\E[K:cl=\\E[H\\E[2J:dl=\\E[M:\
md=\\E[1m:me=\\E[m:mr=\\E[7m:sc=\\E7:rc=\\E8:sf=\n:\
so=\\E[7m:se=\\E[m:sr=\\EM:ku=\\EOA:ks=\\E[?1h\\E=:ke=\\E[?1l\\E>:"
    strcpy(eebuf,VT100eebuf);
    strcpy(s,"vt100internal");
    error=1;
  }
  buff= &buf[0];
  if(!egetstr("cm",&cm_termcap)){
    etype("Bad TERMCAP cursor motion CM.\r\n");
    stat = -1;
    goto top;
  }
  egetstr("al",&al_termcap);
  egetstr("dl",&dl_termcap);
  egetstr("ce",&ce_termcap);
  egetstr("cd",&cd_termcap);
  if(ce_termcap[0]==0 && cd_termcap[0]==0){
    etype("Bad TERMCAP CE and CD.\r\n");
    stat = -1;
    goto top;
  } else {
    egetstr("cs",&cs_termcap);
    if(0==egetstr("sf",&sf_termcap)) sf_termcap="\n";
    egetstr("sr",&sr_termcap);
    if(0==egetstr("pc",&pc_termcap)) pc_termcap="\200";
    egetstr("rs",&rs_termcap);
    egetstr("is",&is_termcap);
    egetstr("ks",&ks_termcap);
    egetstr("ke",&ke_termcap);
    egetstr("so",&REVVIDEO);
    egetstr("se",&NORMVIDEO);
    egetstr("md",&BOLDF);
    if(!egetstr("bl",&bl_termcap)) bl_termcap="\7";
    if(egetstr("me",&buff) && buff[0]) newcore(&NORMVIDEO,buff);
  }
  /* Set rows and columns on screen, reset max line if needed */
  x = tgetnum("co"); x -= 2; if(x <= -1) x=79; COLMAX=x;
  x = tgetnum("li"); x -= 2; if(x <= -1) x=22; ROWMAX=x;
  if(MAXLN <= COLMAX) MAXLN=COLMAX+10;

  /* Define variables HASSCROLL, HASDELETELINE, HASINSERTLINE */
  HASINSERTLINE=HASDELETELINE=HASSCROLL=0;
  if(al_termcap[0]) HASINSERTLINE=1;
  if(dl_termcap[0]) HASDELETELINE=1;
  if(cs_termcap[0] != 0 && sf_termcap[0] != 0 && sr_termcap[0] != 0)
    HASSCROLL=1;
  if(error){
    etype("Press RETURN to continue (ctrl-C aborts): ");
    if(getbyte()==3) {crlf(); quit(100);}
  }
  if(s[0]) newcore(&termtype,s);
  return 0;
}

char * skipdelay(s) char *s; {
char *p;
  p=s;
  while(isdigit(p[0])) ++p;
  if(p[0]=='.') ++p;
  while(isdigit(p[0])) ++p;
  if(p[0]=='*' && p!=s) ++p;
  return p;
}

int egetstr(s,d) char *s; char **d; {
char *p;
char buff[512];
  p=buff;
  p=tgetstr(s,&p); if(p==(char *)0){d[0]=""; return 0;}
  newcore(d,skipdelay(buff));      /* 2/96 Skip over stinking delays */
  return 1;
}

/* Typical termcap entry: d0|vt100|vt100-am|vt100am|dec vt100:\ */
/* First use: exact match to the terminal name string (c=='|'). */
/* eg, "|vt100-am|" appears in the string and we want to match "vt100-am" */
/* Second use: match token in the termcap string (c==':'). */
/* eg, ":tc=vt100" is in the string and we want to pick off "=vt100" */

char *findName(buf,name,c)
char *buf;              /* null-terminated string from termcap database */
char *name;             /* terminal name to locate */
int  c;                 /* left delimiter */
{
int n,i; char *p;
  p=buf;
  do {
    if(p[0]==0) break;
    i=0;
    if(c == '|'){ /* Use exact token length in match */
      while(p[i] != 0 && p[i] != '|' && p[i] != ':') ++i;
    }
    n=max(i,strlen(name));
    if(strncmp(p,name,n)==0) goto gotit;
    if((p=strchr(p+1,c))!=(char *)0) ++p;
  } while(p!=(char *)0);
  return (char *)0;
gotit:
  if(c==':') return p+n;        /* Point past match */
  return p;                     /* Point to exact match */
}

/*   g e t E n t r y
 *  Gets the named entry from the already-opened termcap file into
 *  the buffer. The size of the buffer is 1024, and it is
 *  considered to be an error if the size of the entry equals or
 *  exceeds this. We place a terminating NULL character at the end
 *  of the entry. Call error() on any irregularities. Return 0 if
 *  the named entry not found, else 1. Removes terminal names and
 *  all newlines from the entry. If this is called for a 2nd time
 *  from tgetent(), then the length checking is useless.
 */
static int getEntry(fp,outbuf,name)
FILEP  fp;              /* Stream pointer for termcap file*/
char  *outbuf;          /* where we put the entry */
char  *name;            /* terminal type name to locate */
{
char    *p;             /* scrap pointer variable */
char    inbuf[BUFSIZE]; /* temporary termcap file buffer 1024 bytes */

  outbuf[0]=0;
  if(name==(char *)0 || name[0]==0) goto quit;
  while(fgetln(inbuf,fp)>=0){
     if(inbuf[0]==0) continue;
     if(strchr("#\t\r\n\f\b\040",inbuf[0])!=(char *)0) continue;
     /* must be a terminal line entry */
#if 0
 etype(inbuf); crlf(); /* this prints out all termcap terminal names! */
#endif
     if(findName(inbuf,name,'|')!=(char *)0){  /* found terminal name */
       do {
         if(inbuf[0]!='#' && *sob(inbuf)!=0){
            strcat(outbuf,sob(inbuf));
            p=endof(outbuf);
            if(p[-1]=='\\') p[-1]=0; /* continuation line found */
            else break;              /* Go check tc= capability */
         }
         if(fgetln(inbuf,fp)<0) goto quit;      /* file format error */
       } while(inbuf[0]==0 || strchr("#\t\r\n\f\b\040",inbuf[0])!=(char *)0);
       /* Check for tc= capability done by caller */
       goto success;    /* got termcap entry in outbuf */
     }/* end if(found terminal name) */
  }/* end while on reading stream=fp */
  return 0;     /* Could not find name in termcap file */
quit:           /* Get here on error, so set retval = -1 */
  return -1;
success:        /* Get here if outbuf[] successfully filled, retval=1 */
  return 1;
}

/*   t g e t e n t
 *  Extracts the entry for terminal name into the buffer at bp. Bp
 *  should be a character array of size 1024 and must be retained
 *  through all subsequent calls to tgetnum(), tgetflag(), and
 *  tgetstr(). Returns -1 if it cannot open the termcap file, 0 if
 *  the terminal name given does not have an entry, and 1 if all
 *  goes well. Looks in the environment for a TERMCAP variable. If
 *  found, and the value does not begin with a slash, and the
 *  terminal type name is the same as the environment string TERM,
 *  the TERMCAP string is used instead of reading the termcap file.
 *  If it does begin with a slash, the string is used as a pathname
 *  rather than /etc/termcap. This can speed up entry into programs
 *  that call tgetent(), as well as to help debug new terminal
 *  descriptions or to make one for your terminal if you can't write
 *  the file /etc/termcap.
 */
int tgetent(bp, name)
char  *bp;              /* pointer to buffer */
char  *name;           /* terminal type name string */
{
char  *termcap;         /* pointer to environment TERMCAP string */
FILEP fp;               /* file pointer for termcap file */
int   retval;           /* return value */
char *p;
char  newname[128];

  p = getenv("TERMCAP");
  termcap = TERMFILE;  /* in case :tc capability found */
  retval = -1;
  if(p != (char *)0){
    if(p[0]=='/' || (p[0]!=0 && p[1]==':')) termcap=p;
    else {
      if(findName(p,name,'|')!=(char *)0){
        /* use TERMCAP as the entry */
        strnzcpy(bp,p,min(BUFSIZE,strlen(p)+1));
        retval =1;          /* using TERMCAP variable, no files needed */
      }
    }
  }
  if(retval == -1) {              /* look for entry in termcap file */
    if((fp = fopenbread(termcap))!=(FILEP)0) {
      retval = getEntry(fp, bp, name); closedfile(fp);
    }
  }
  if(retval == 1){
    /* consider :tc= capability */
    p = findName(bp,"tc",':');
    if(p != (char *)0) {
      retval = -1;
      strnzcpy(newname,&p[1],127);
      *skipto(":",newname )=0;
      if((fp = fopenbread(termcap))!=(FILEP)0){
        retval = getEntry(fp,&p[-2],newname); closedfile(fp);
      }
    }
  }

  return retval;
}



/*   t g e t n u m
 *  Gets the numeric value of capability id, returning -1 if it is
 *  not given for the terminal.
 */
int tgetnum(id)
char  *id;                /* capability name */
{
char  *p;
LLONG atoiLL();

  p = findName(eebuf,id,':');
  return ((p==(char *)0 || p[0] != '#') ? (-1) : (int)atoiLL(p+1));
}

/*   t g e t f l a g
 *  Returns 1 if found, 0 otherwise
 */
int tgetflag(id)
char  *id;                /* capability name */
{
char  *p;

  p = findName(eebuf,id,':');
  return (p != (char *)0  &&  p[0] == ':');
}



/*   t g e t s t r
 *  Returns the string value of the capability id, places it in the
 *  buffer at area, and advances the area pointer [past the
 *  terminating '\0' char]. It decodes the abbreviations for this
 *  field described in termcap(5), except for cursor addressing and
 *  padding information. Returns NULL if the capability was not
 *  found.
 */
char *tgetstr(id, area)
char  *id;              /* capability name */
char  **area;           /* pointer to output pointer */
{
char    *retval;        /* return value */
char    *p;             /* pointer into capability string */
int     x;

  p = findName(eebuf,id,':');
  if(p == (char *)0 || p[0] != '=') retval = (char *)0;
  else {
    retval = area[0]; ++p;
    while(p[0] != ':') {
      if (p[0] == '\\'){
        ++p;
        if(strchr("01234567",p[0])!=(char *)0){
          x=0;
          while(strchr("01234567",p[0])!=(char *)0) x = 8*x + (*p++ - '0');
          x  =  (x & 255);
          --p;
          /** will \200 really end up as \000 like it should ? **/
          /** see termcap(5), top of page 6 **/
        } else {
          switch (p[0]) {
          case 'e': x  =  '\033';  break;
          case 'E': x  =  '\033';  break;
          case 'b': x  =  '\b';  break;
          case 'f': x  =  '\f';  break;
          case 'n': x  =  '\n';  break;
          case 'r': x  =  '\r';  break;
          case 't': x  =  '\t';  break;
          default:  x  =  (255&p[0]);  break;
          }
        }
      }
      else if(p[0] == '^')
        x  =  *++p - '@';  /* control */
      else
        x  =  p[0];        /* normal */
      *(*area)++  =  x;
      ++p;
    } /* end while */
    *(*area)++ = '\0';          /* NULL-terminate the string */
  } /* end else */
  return retval;
}



/*   t g o t o
 *  Returns a cursor addressing string decoded from cm to go to
 *  column destcol in line destline. It uses the external variables
 *  UP (from the up capability) and BC (if bc is given rather than
 *  bs) if necessary to avoid placing \n, ^D, or ^@ in the returned
 *  string. (Programs which call tgoto() should be sure to turn off
 *  the XTABS bit(s), since tgoto() may not output a tab. Note that
 *  programs using termcap should in general turn off XTABS anyway
 *  since some terminals use control I for other functions, such as
 *  non- destructive space.) If a % sequence is given which is not
 *  understood, then tgoto() returns "OOPS".
 *
 *  Output buffer is static - recursion will not work.
 *  No error checking here.
 */
char *tgoto(cm, destcol, destline)
char  *cm;              /* cm capability string */
int   destcol;          /* destination column (left is 0) */
int   destline;         /* destination line (top is 0) */
{
char    *outp;          /* pointer into answer[] */
static char  answer[88];/* result stashed here */
int  reversed;          /* ==1 when should send col 1st */
int     value;          /* next value to output */
extern char *edigits();


  reversed = 0;
  value = destline;
  outp = answer;
  while(cm[0]){
    if (*cm == '%') {
      switch (*++cm) {
      case '%': *outp++ = '%';
            break;
      case 'd':
            strcpy(outp,edigits(value,1));      /* no lead zeros unless "0" */
            outp += strlen(outp);
            goto rev;
      case '2': strcpy(outp,edigits(value,2));
            outp += 2;
            goto rev;
      case '3': strcpy(outp,edigits(value,3));
            outp += 3;
            goto rev;
      case '.': *outp++ = (value==0 ? 128 : value);
            goto rev;
      case '+': *outp++ = value + *++cm;
rev:        value = (reversed) ? destline : destcol;
            break;
      case '>': if(value > *++cm) value += *++cm; else ++cm;
            break;
      case 'r': value = (reversed) ? destline : destcol;
            reversed ^= 1;
            break;
      case 'i': ++value; ++destline; ++destcol;
            break;
      case 'n': destcol  ^= 0140;
            destline ^= 0140;
            break;
      case 'B': value = (16 * (value / 10))  +  (value % 10);
            break;
      case 'D': value = (value - (2 * (value % 16)));
            break;
      default:  strcpy(outp,"OOPS");
            outp += 4;
            break;
      }
    } else *outp++ = *cm;
    ++cm;
  }
  *outp = '\0';
  return answer;
}



/*   e t p u t s
 *  Decodes the leading pad information of the string cp; affcnt
 *  gives the number of lines affected by the operation, or 1 if
 *  this is not applicable. Outc is a routine which is called with
 *  each character in turn. The external variable ospeed should
 *  contain the output speed of the terminal as encoded by stty(3).
 *  The external variable PC should contain a pad character to be
 *  used (from the pc capability) if a null (^@) is inappropriate.
 */
void etputs(cp, affcnt)
char  *cp;              /* leading pad information string */
int   affcnt;           /* number lines affected, or 1 */
{
char    *p;
int  delay;
/* milliseconds per character, for each of the possible baud rates of ospeed */
/* table is multiplied by 10 to eliminate decimal arithmetic */
static unsigned  delayFactor[] = {
     0,   /* B0 */  /* hang up  or network device */
  1920,   /* B50 */
  1280,   /* B75 */
   872,   /* B110 */
   730,   /* B134 */
   640,   /* B150 */
   480,   /* B200 */
   320,   /* B300 */
   160,   /* B600 */
    80,   /* B1200 */
    50,   /* B1800 */
    40,   /* B2400 */
    20,   /* B4800 */
    10,   /* B9600 */
     5,   /* EXTA (19200 here) */
     2,   /* EXTB (34800 here) */
};
LLONG atoiLL();

  etype(skipdelay(cp));               /* output string, delay later */
  /* calculate delay and output pad characters, if any */
  delay=(int)atoiLL(cp);
  if(delay){
    if(ospeed){                         /* nonzero tty speed? */
      delay = 10*delay;       /* delay factor is in 10ms units*/
      while(isdigit(cp[0])) ++cp;
      if(cp[0] == '.'){                 /* decimal point? */
        delay += (5 + (int)atoiLL(++cp)*100)/10;
        while(isdigit(cp[0])) ++cp;
      }
      if(cp[0] == '*') delay *= affcnt;
      /* output delay characters */
      while(delay>0){ etype(pc_termcap); delay -= delayFactor[ospeed];}
    }/* end if */
  }/* end if */
}

/*
 * EDIGITS
 *  Purpose:   special itoa() with zero-fill to n-digits.
 */
char *
edigits(r,n)
int r,n;
{
char d[11];
static char e[11];
register int i,j;
  if(n > 10) n=10;
  i = r; j = 0;
  do{
     d[j++] = '0' + i%10; i /= 10;
  } while(i);
  i=0;
  while(j<n) d[j++]='0';
  while(--j>=0) e[i++] = d[j];
  e[i]=0;
  return e;
}

#endif          /* TERMCAP */
#endif
