/*
 * This file contains routines for MS-DOS using the TURBOC
 * compiler version 1.5. Hardware dependencies are:
 *
 *              int 10h
 *              int 11h
 *              direct video for mono or color
 *              int 21h
 * Also contained here are portable C versions of all the IBM
 * routines, which are used on Unix hardware.
 *
 * See Peter Norton's Programmers Guide to the IBM PC for help on
 * understanding the routines below.
 *
 */
                                /* otherwise use direct video */
#include "v.h"

#if TURBOC              /* DOS hosts */
#if IBMHARDWARE         /* only /w IBM hardware */
#pragma inline
#define bios 10h

void inslin() {
  insORdel(0x0701);
}

void dellin() {
  insORdel(0x0601);
}

void insORdel(n) unsigned n; {
extern int ROWMAX,COLMAX,SYSATT;
/*      Insert or delelete a line at the current cursor position */
asm     cmp     nodisp,0
asm     jnz     end
asm     mov ah,3                /*  Get cursor position. */
asm     mov bh,0
asm     int bios
asm     push dx
asm     mov ch,dh
asm     xor cl,cl               /*  Start at beginning of row. */
asm     mov dh,byte ptr ROWMAX  /*  dh=number of rows minus 1 */
asm     inc dh                  /* make up for status line */
asm     mov dl,byte ptr COLMAX  /* dl=number of columns minus 1 */
asm     mov ax,n
asm     mov bh,byte ptr SYSATT
asm     int bios                /*  Scroll down one line. */
asm     pop dx
asm     xor dl,dl
asm     mov bh,dl
asm     mov ah,2
asm     int bios
end:
;
}

void fwdindex(){
extern int ROWMAX,COLMAX,SYSATT;
asm     cmp     nodisp,0
asm     jnz     end
asm     mov dx,0100h            /* putcursor (0,0) */
asm     xor bh,bh
asm     mov ah,2
asm     int bios
asm     mov ch,dh
asm     xor cl,cl               /*  Start at beginning of row. */
asm     mov dh,byte ptr ROWMAX
asm     inc dh                  /* make up for status line */
asm     mov dl,byte ptr COLMAX
asm     mov bh,byte ptr SYSATT
asm     mov ax,0601H            /*  Scroll up one line. */
asm     int bios
end:
;
}

void cls(){                          /* clear all screen positions */
        blankit(0x0000,7);
        home();
}

void clearscreen()
{
/* This routine blanks the screen and puts the cursor home [0,0] */
    blankit(0x0000,SYSATT);
    home();
    statblank=1;
}

void cleardisplay(){
/* This routine blanks the screen and puts the cursor on line 2 */
    blankit(0x0100,SYSATT);
    putcursor(0,0);
}

void blankit(corner,attr) unsigned corner; unsigned char attr; {
extern int ROWMAX,COLMAX,nodisp;
asm     cmp     nodisp,0
asm     jnz     end
asm     mov ah,0fh              /* Get current video state */
asm     int bios                /* ah=cols, al=mode, bh=active page */
asm     mov cx,corner           /* point=[chigh+1,clow+1] */
asm     mov dh,byte ptr ROWMAX  /* blank rectangle dh x dl */
asm     inc dh                  /* make up for status line */
asm     mov dl,ah               /* Number of columns */
asm     mov bh,attr             /* attribute to write */
asm     mov ax,600H
asm     int bios
end: ;
}

void eraseol()
{
extern int ROWMAX,COLMAX,SYSATT;
asm     cmp     nodisp,0
asm     jnz     end
asm     mov ah,3
asm     xor bh,bh
asm     int bios        /*  getcursor() */
asm     mov cx,dx
asm     mov dl,byte ptr COLMAX
asm     mov ax,0700h
asm     mov bh,byte ptr SYSATT
asm     int bios        /* scroll() */
end:
;
}

void killstatus()
{
extern int statblank;
statblank = 1;
home();
eraseol();
}

void putcursor(rrow,ccol)
int rrow,ccol;
{
extern int ROWMAX,COLMAX;
asm     cmp     nodisp,0
asm     jnz     end
asm     xor bh,bh       /* page 0 */
asm     mov ax,rrow
asm     cmp ax,ROWMAX       /* off screen ? */
asm     jle putc1
asm     mov ax,ROWMAX
putc1:
asm     inc ax          /* row - offset 1 in this application */
asm     xchg ax,dx
asm     mov ax,ccol     /* column */
asm     cmp ax,COLMAX   /* off screen ? */
asm     jle putc2
asm     mov ax,COLMAX
putc2:
asm     mov dh,dl
asm     mov dl,al
asm     mov ah,2        /* set cursor position */
asm     int bios
end:
;
}

void home()
{
asm     cmp     nodisp,0
asm     jnz     end
asm     xor dx,dx        /*      ; Go to top left corner of screen. */
asm     mov bh,dh
asm     mov ah,2
asm     int bios
end:
;
}

void erachar() {
extern int SYSATT;
asm     cmp     nodisp,0
asm     jnz     end
asm     mov ah,3
asm     xor bh,bh
asm     int bios        /*  getcursor() */
asm     cmp dl,0        /* on column 0? */
asm     jz end          /* can't erase on column 0 */
asm     dec dl          /* column-- */
asm     xor bh,bh       /* page 0 */
asm     mov ah,2        /* set cursor position */
asm     int bios
asm     mov bl,byte ptr SYSATT
asm     xor bh,bh
asm     mov cx,1
asm     mov al,' '      /* move left and print ' ' */
asm     mov ah,9
asm     int bios
end:
;
}

int getcursor() {
asm     xor     bh,bh
asm     mov     ah,3
asm     int     bios
asm     mov     ax,dx
return (_AX);
}

#if BIOSSRC
/* use this for a general version of PI for any clone */
/* Uses int 10h functions only - no direct video */
/* Good for machines like the Z100, and SUN DOS386I */

void ePrint(s) char *s; {
extern int nodisp;
  if(!nodisp){
    eBios(s,getcursor(),0);
  }
}

void eBios(s,n,flag) char *s; int n,flag; {
extern int invvideo();
extern int COLMAX;
unsigned char edge;
        edge= (unsigned char)COLMAX;
asm     mov     dx,n
asm     mov     si,s
bios1:
asm     mov     AL,[SI]
asm     inc     si
asm     cmp     al,0
asm     je      end
asm     mov     ah,al
asm     and     ah,96           /* test for inverse video attribute */
asm     je      bios3
asm     cmp     dl,edge         /* at edge of screen ? */
asm     jne     bios2           /* no, then write char and advance cursor */
asm     mov     ah,10           /* write video without cursor advance */
asm     mov     cx,1            /* because at right edge of screen */
asm     int     bios
asm     jmp     end             /* all done */
bios2:
asm     mov     ah,14           /* write tty */
asm     int     bios            /* with cursor advance */
asm     inc     dl
asm     jmp     bios1
bios3:
asm     push    dx
asm     push    bx              /* char needs to be in */
asm     push    ax              /* inverse video */
asm     call    invvideo
asm     pop     ax
asm     pop     bx
asm     pop     dx
asm     inc     dl
asm     cmp     dl,edge         /* but we might be at edge of screen */
asm     jle     bios1           /* if so, then quit SIGNED BYTE ERROR! */
end:
  if(flag) eraseol();
}
#endif


#if !BIOSSRC

/* use this for a version of PI for any clone with mono or color adapter */
void ePrint(s) char *s; {
extern int nodisp;
  if(!nodisp){
    if(iscga()) waitsync();
    setcursor(eDirect(s,getcursor(),0));
  }
}

/* write direct to video ram - returns new cursor position */

int eDirect(s,n,flag)
char *s;                        /* char buffer, null-terminated */
int n;                          /* dx=n, dl=col, dh=row cursor pos */
int flag;                       /* 0 for no fill, 1 for fill */
{
extern unsigned _screen;
extern int ROWMAX,COLMAX,SYSATT;
asm     mov ah,0fh              /* Get current video state */
asm     int bios                /* ah=cols, al=mode, bh=active page */
asm     mov     al,ah
asm     xor     ah,ah           /* ax== number of columns displayed */
asm     mov     si,s
asm     mov     bx,n            /* Cursor position */
asm     mov     cl,bh           /* row number */
asm     xor     ch,ch
asm     imul    cx              /* uses ax,cx,dx */
asm     mov     cl,bl           /* col number */
asm     xor     ch,ch
asm     add     ax,cx
asm     add     ax,ax           /* ax = offset in ram */
asm     mov     di,ax           /* video ram offset */
asm     mov     ax,_screen      /* video ram segment */
asm     mov     es,ax
asm     mov     cl,byte ptr flag        /* blank-fill flag for end of line */
asm     mov     dx,bx
asm     mov     bl,byte ptr COLMAX
asm     mov     bh,byte ptr SYSATT
loop:
asm     mov     AL,ds:[SI]

asm     cmp     al,0

asm     jne     loop1           /* end of string found */
asm     test    cl,1            /* Are we supposed to fill to end of line ? */
asm     jz      quit            /* no, then just quit */
asm     mov     al,' '          /* yes, then add in a blank */
asm     mov     ah,bh           /*      with reg attrib */
asm     jmp     short noattr
loop1:
asm     inc     si
asm     test    al,96           /* test for inverse video attribute */
asm     mov     ah,bh
asm     jnz     noattr          /* no, then write char with reg attribute */
asm     test    al,128          /* Is the high bit set? */
asm     jnz     noattr          /* yes, then write char with reg attribute */
asm     mov     ah,070h         /* no, change to inverse video */
asm     add     al,'@'          /* convert to printable */
noattr:
/*  mono adaptor processing - requires no port checks */
mono:
asm     mov     es:[di],ax
asm     inc     di
asm     inc     di
asm     cmp     dl,bl             /* at edge of screen ? */
asm     je      quit
asm     inc     dl
asm     jmp     short loop
quit:
end:
asm     mov     ax,dx           /* return new cursor */
return (_AX);
}

#endif

void invvideo(x)
int x;
{
extern int COLMAX;
asm     xor bh,bh               /* display page */
asm     mov ah,3
asm     int bios
asm     mov cx,1                /* write 1 char */
asm     mov ah,9                /* function 9, video services */
asm     mov al,x                /* char to print */
asm     test al,128             /* Signed? */
asm     jnz next
asm     cmp al,' '              /* control char? */
asm     jge next                /* No, then don't add '@' correction */
asm     add al,'@'
next:
asm     mov bl,70h              /* inverse video attribute */
asm     int bios
asm     cmp dl,byte ptr COLMAX
asm     je done                 /* can't move cursor */
asm     mov ah,14               /* overprint as tty */
asm     int bios
done:
;
}

void strinvvideo(s) char *s; {
  if(!nodisp){
    while(*s) invvideo(*s++);
  }
}

void xPrint(p,rrow,ccol) char *p; int rrow,ccol; {
extern int COLMAX,nodisp;
  if(!nodisp) {
    if(ccol > COLMAX+1) ccol = COLMAX+1;
    putcursor(rrow,ccol);
    ePrint(p);
  }
}

void setcursor(n) unsigned n; {
asm     mov     dx,n
asm     mov     ah,2            /* set cursor */
asm     xor     bh,bh
asm     int     bios
}

/* find out if mono or color card or neither */

unsigned _screen = 0xb000;
int oldbios=0;

void equipment() {
#define TEXT16colorHIGHRES80x25  3      /* cga high resolution text mode */
#define TEXTmonoHIGHRES80x25     7      /* mono high resolution text mode */
/* Set video modes AL=3 or 7, AH=0, int bios */
        cls();                  /* Erase screen before changing its size */
#if !BIOSSRC
asm     mov     ax,0500h        /* set active display page */
asm     int     bios
asm     int 11h                 /* do equipment check */
asm     and ax,30h
asm     mov _screen,0b000h
asm     cmp ax,30h
asm     je mono                 /* It's mono */
asm     mov _screen,0b800h      /* It's color (ax==20h)*/
mono:
#endif
asm     mov ax,40h
asm     mov es,ax
asm     test byte ptr [es:96h],10000b  /* enhanced keyboard installed */
asm     mov oldbios,10h
asm     jnz new
asm     mov oldbios,0h
new:
        setcursortype(1);       /* user cursor shape */
}

/*
 * setcursortype()
 *
 * Scan lines numbered 0-7 for CGA, 0-13 for VGA/MONO
 * n=0 : no cursor  ch=32, cl=0
 * n=1 : block cursor ch=1, cl=13 for vga, ch=6,  cl=7  for cga
 */
void setcursortype(n) int n; {
extern int cursorCH,cursorCL;
asm     xor ax,ax
asm     mov ch,32       /* n=0 is set no cursor */
asm     mov cl,al
asm     cmp ax,word ptr n
asm     je setct
asm     int 11h
asm     mov ch, byte ptr cursorCH       /* set cursor shape to user shape */
asm     mov cl,byte ptr cursorCL
setct:
asm     xor al,al
asm     mov ah,1
asm     mov bh,al
asm     int 10h
}

char vmodes[]={0x50,0x54,0x58,  /* 30 rows */
               0x51,0x55,0x59,  /* 43 rows */
               0x52,0x56,0x5a,0x00}; /* 60 rows */
char vmoderows[]={23,28,41,58};         /* 2 less than num rows */

int getvideomode(){
asm     mov ah,0fh              /* Get current video state */
asm     int bios                /* ah=cols, al=mode, bh=active page */
asm     xor ah,ah               /* AX==video mode */
return (_AX);
}

void setvideomode(x) unsigned x; {
unsigned i,j; unsigned char mode;
asm     xor ah,ah
asm     mov al,byte ptr x
asm     int bios
asm     mov ah,0fh              /* Get current video state */
asm     int bios                /* ah=cols, al=mode, bh=active page */
asm     mov byte ptr mode,al    /* Save mode */
/* If mode is 81,84,88 then use 29 rows */
/* If mode is 82,85,89 then use 42 rows */
/* If mode is 86,90 then use 59 rows */
/* AH==num columns, use 39, 79 or 131 by default */
asm     mov bl,39
asm     cmp ah,40               /* 40-col mode? */
asm     je done
asm     mov bl,79
asm     test ah,128             /* Signed ? */
asm     jz done
asm     mov bl,131
done:
asm     xor bh,bh
asm     mov word ptr j,bx
   i=(istrchr(vmodes,mode)+2)/3;
   setwindowsize(vmoderows[i],j);       /* Does equipment() too */
}

void cursorstyleset(hi,lo) char hi,lo; {
asm     mov ch,hi
asm     mov cl,lo
asm     mov ah,1
asm     xor al,al
asm     mov bh,al
asm     int 10h
}

void shortclick(lo,hi,duration) char lo,hi; int duration; {
/* LO and HI frequencies. See MBASIC program */
asm     mov     al,182
asm     out     67,al
asm     mov     al,lo
asm     out     66,al
asm     mov     al,hi
asm     out     66,al
asm     in      al,97
asm     mov     ch,al
asm     mov     cl,al
asm     or      ch,3
asm     mov     al,ch
asm     out     97,al
asm     push    cx
asm     mov     ax,duration

loop:
asm     dec     ax
asm     or      ax,ax
asm     jnz     loop

asm     pop     cx
asm     mov     al,cl
asm     out     97,al
}

void drawall()
{
int i,k;
extern char **line; extern char *toprint();
extern int ROWMAX,nodisp,halfdone;
  if(testkey()) {nodisp=halfdone=1; return;}
  if(!nodisp) {
    k=iscga();
    for(i=0;i<=ROWMAX;++i) {
        if(bioskey(1)) goto quit;
        /* about 460 characters can be written after waitsync() */
#if BIOSSRC
        putcursor(i,0);
        eBIOS(toprint(line[i]),(i+1) << 8,1); /* write 160 characters */
#else
        if(k) if(i % 2 == 0) waitsync();
        eDirect(toprint(line[i]),(i+1) << 8,1); /* write 160 characters */
#endif
    }
    lcount();
    halfdone=0;
  }
quit:;
}

int iscga() {
extern int _egavga;
extern unsigned _screen;
asm     xor     ax,ax
#if !BIOSSRC
asm     mov     dx,_screen
asm     cmp     dx,0b800h
asm     jne     short ismono
asm     cmp     ax, _egavga
asm     jne     short ismono
asm     inc     ax
#endif
ismono:
return (_AX);
}

void waitsync() {
#if !BIOSSRC
asm     mov     dx,03DAh
asm     mov     ah,09h          /* retrace mask. Al Williams, p 186 */
asm     cli
cga1:
asm     in      al,dx           /* wait for horizontal sync */
asm     rcr     al,1            /* Adaptor slower than monitor retrace */
asm     jc      cga1            /* This came from AT bios source */
cga2:
asm     in      al,dx           /* wait for vertical retrace */
asm     and     al,ah           /* about every 63 microseconds */
asm     jz      cga2
asm     sti
#endif
cga3:
;
}


/* Return 0 if not in enhanced mode, 1 otherwise */
int isWinEnhanced(){
asm     mov ax,0x1600   /* See if Enhanced mode windows */
asm     int 2fh
asm     mov ah,0
asm     or al,al
asm     jz quit
asm     inc al
quit:
        return (_AX);
}

#endif /* IBMHARDWARE special code ends */


/* setargv -- setup argv with wild card expansion for Turbo-C 1.5 */
#ifdef __TURBOC__               /* only for turboc command line */
/*  Get Program Name - argv[0] - TurboC2.0 setargv.asm does it */
void getprogramname(buffer) char *buffer; {
extern unsigned _envLng,_envseg; extern unsigned char _osmajor;
asm     mov     si,buffer       /* ds:si is buffer */
asm     mov     di, _envLng
asm     add     di, 2           /* DI = Program name offset */
asm     mov     es, _envseg     /* ES = segment address */
asm     mov     cl, 07fh        /* buffer is 256 bytes */
asm     cld
asm     mov     al,_osmajor
asm     cmp     al,3            /* No program name for DOS 2 */
asm     jb      loopexit
loop:
asm     mov     al,byte ptr es:[di]     /* Copy string */
asm     mov     byte ptr ds:[si],al
asm     or      al,al
asm     jz      loopexit
asm     dec     cl
asm     or      cl,cl
asm     jz      loopexit
asm     inc     si
asm     inc     di
asm     jmp     loop
loopexit:
asm     xor     al,al
asm     mov     byte ptr ds:[si],al     /* Terminate string! */
}

#ifndef WILDCARDS       /* WILDCARDS defined in dir.h */
#include <dir.h>
#endif
#ifndef NFDS            /* NFDS defined in dos.h */
#include <dos.h>
#endif

void _setargv()
{
unsigned char far *farcmdtail;
char buffer[129],s[129];
char *q[256];
char *p,*cmdtail,*pathend;
int i,k,n,c,wild;
struct ffblk ffb;
extern char *core(),*sob(),*fnb();
extern int              _argc;
extern char             **_argv;
extern unsigned         _psp;

  farcmdtail = MK_FP(_psp, 0x81);
  n=farcmdtail[-1]; if(n<0 || n>0x80) {n=0; goto quit;}
  for(s[n]=i=0;i<n;++i) s[i]=farcmdtail[i];
  n=0;
  cmdtail=s;
  while(*(cmdtail=sob(cmdtail)) != 0){
    /* find word */
    p = pathend = buffer;
    k=fnb(cmdtail)-cmdtail;             /* how many chars to scan */
    for(wild=i=0;i<k;++i){
      if ((c=cmdtail[i]) == '/') c = '\\';
      wild |= (c == '*' || c == '?');
      *p++ = c;
      if (c == '\\' || c == ':') pathend = p;
    }
    p[0]=0;

    if(wild && cmdtail[0] != '-')
      for(c=findfirst((char *)buffer,&ffb,0);c==0;c=findnext(&ffb)){
        strcpy(pathend,ffb.ff_name);
        if(n < 256) newcore(&q[n++],buffer);
      }
    else
      if(n<256) newcore(&q[n++],buffer);
    cmdtail += k;
  }

quit:
  _argv = (char **)core((n+1)*sizeof(char *));
  for(i=1;i<=n;++i) _argv[i]=q[i-1];
  _argv[0] = "C";
  _argc=n+1;
}
#endif  /* turbo-c command line code only */

#ifdef __TURBOC__
#ifdef __SMALL__
/* malloc.c */
#define MEMINCR 1024                    /* memory to allocate */

#define PAD     1                       /* Extra needed in block */
#define SLOP    4                       /* Extra before split */

struct  header  {
        unsigned           h_size;         /* Size of this free block */
        struct  header  *h_ptr;         /* Pointer to next header */
        };

static  struct  header  Base = { 0, &Base };
                                        /* Empty list to get started */
static  struct  header  *Allocp = &Base;        /* Last allocated block */

extern  void    free();
extern  void    *sbrk();
extern  void    *ssbrk();

void *malloc(nbytes) unsigned nbytes; {
register struct header *p, *q;
register char *cp;

  nbytes += PAD + sizeof p->h_size;
  nbytes &= ~1;
  q = Allocp;
  for (p = q->h_ptr; ; q = p, p = p->h_ptr){
    if (p->h_size >= nbytes){
      if (p->h_size <= nbytes + SLOP) q->h_ptr = p->h_ptr;
      else {
        p->h_size -= nbytes;
        p = (struct header *)((char *)p + p->h_size);
        p->h_size = nbytes;
      }
      Allocp = q;
      return((char *)(&p->h_ptr));
    }
    if (p == Allocp) break;
  }
  cp = ssbrk(nbytes);
  if ((int)cp == -1) return(0);
  p = (struct header *)cp;
  p->h_size = nbytes;
  return((char *)(&p->h_ptr));
}

void free(ap) char *ap; {
register struct header  *p, *q;

  p = (struct header *)(ap - sizeof p->h_size);
  for (q = Allocp; !(p > q && p < q->h_ptr); q = q->h_ptr)
    if (q >= q->h_ptr && (p > q || p < q->h_ptr)) break;

  if ((char *)p + p->h_size == (char *)q->h_ptr){
    p->h_size += (q->h_ptr)->h_size;
    p->h_ptr = (q->h_ptr)->h_ptr;
  }
  else p->h_ptr = q->h_ptr;
  if ((char *)q + q->h_size == (char *)p){
    q->h_size += p->h_size;
    q->h_ptr = p->h_ptr;
    p = q;
  }
  else q->h_ptr = p;

  /* If the block just freed is at the top of memory, use brk */
  if((char *)p + p->h_size == ssbrk(0)){ /* Find the new end of list */
    for (q = p; q->h_ptr != p; q = q->h_ptr) ;
    q->h_ptr = p->h_ptr;
    brk(p);
  }
  Allocp = q;
}

void *ssbrk(n) unsigned n; {
char *p;
/* Could kill the stack, since it's here also */
  p=sbrk(0);
  if(n==0) return p;
  if((unsigned)p > 63488) return (char *)-1;
  if(n>(unsigned)63488 - (unsigned)p) return (char *)-1;
  return sbrk(n);
}
#endif

#undef getwdir
char *getwdir(s) char *s; {
asm     mov     ah,25           /* get current disk */
asm     int     21h
asm     xor     ah,ah
asm     add     ax,'A'
asm     mov     si,s
asm     mov     byte ptr [si],al
asm     inc     si
asm     mov     byte ptr [si],58        /* colon ':' */
asm     inc     si
asm     mov     byte ptr [si],92        /* backslash '\\' */
asm     inc     si
asm     xor     dl,dl
asm     mov     byte ptr [si],dl        /* null */
asm     mov     ah,71
asm     int     21h
return (char *)(_AX);
}
#if 0
char * getenv(name) char *name; {
        int size,i; unsigned envb;
        unsigned psp; unsigned far *tmp;
        char far *env; char far *e; char *v;

        /* _psp is our PSP. */

        psp = _psp;

        /* Now get the address of environment block
           from the PSP, offset 2ch. */

        tmp = MK_FP(psp,44);
        envb=tmp[0];  /* Fetch word at psp+2ch */
        if (!envb) {/* Can't find environment */
            return (char *)0;}

        env = MK_FP(envb,0);

        /* Get the env block size from the MCB which immediately
           preceeds the environment block.  This is a paragraph count,
           which must be multiplied by 16 to get a byte count. */

        tmp = MK_FP(FP_SEG(env)-1,3);
        size = 16 * tmp[0];

        /* The outer loop below searches for an existing environment
           variable of the same name as we have in name[]. */

        while (*env) {  /* 2 zero bytes marks end of the env */
                /* Compare current env variable with name[]. */
                e = env; v = name; i=size;
                while (e[0] == v[0] && i--){ ++v; ++e;}
                if (e[0]=='=' && v[0]==0) {
                        ++e;  /* skip over equal sign */
                        v=sysx; /* Store it in system scratch */
                        while(e[0]!=0) *v++ = *e++;
                        *v=0;
                        return sysx;
                }
                while ((--size) && (*env++));
        }
        return (char *)0; /* failed */
}
#endif

int bioskey(n) char n; {
asm     mov     ah,n
asm     cmp     ah,1
asm     je      charReady
asm     cmp     ah,11h
asm     je      charReady
asm     int     16h
asm     jmp     end
charReady:
asm     int     16h
asm     jnz     end             /* return AX=char available */
asm     mov     ax,0            /* no char available */
end:
return (_AX);
}

int getdrive() {
asm mov ah,25
asm int 21h
return (_AX);
}

void setdrive(drv) int drv; {
asm mov dx,drv
asm mov ah,14
asm int 21h
}

#ifndef SWAPPER                    /* Use these when Ralf Brown's swapper */
long                            /* is disabled. Also works with SPAWNO40. */
farcoreleft(){                  /* Also with MSC version of SPAWNO30. */
asm     mov     bx,0efffh
asm     mov     ah,48h
asm     int     21h
asm     mov     ax,bx
        return (_AX*16L);
}

void far *farmalloc(n) long n; {
unsigned x;
        x=(unsigned)n;
        x = (unsigned)((x+15)/16);
asm     mov     bx,x
asm     mov     ah,48h
asm     int     21h
asm     jnc     end
asm     mov     ax,0
end: ;
asm     mov     x,ax
        return (char far *)MK_FP(x,0);
}

#endif
#endif


/* same as termraw for DOS386I */

#if DOS386I && !IBMHARDWARE
int nothing(){
extern int nodisp,halfdone;
LONG ccBIOS();
  bufput(ccBIOS(0)); return (halfdone=nodisp=1);
}

void ctrlbreak(n) int n;{
static int x;
extern void __int__();
  if(n==0){
     ctrlbrk(nothing);
    _AL=0; _AH=0x33; __int__(0x21); x= _DL;
    _DL=0;
  } else _DL=x;
  _AL=1; _AH=0x33; __int__(0x21);
}

#else
void
ctrlbreak(n) int n; {   /* Keeps ctrl-C from aborting HELP */
static int x;
extern void __int__();
  if(!n){ _AL=0; _AH=0x33; __int__(0x21); x= _DL; _DL=0; } else _DL=x;
  _AL=1; _AH=0x33; _DL=x; __int__(0x21);
}
#endif

#ifdef SWAPPER
/*
 * MYSYSTEM.C
 * by GBG, May 1991
 *
 * Special system() swaps the application to disk. Thanks to Ralf Brown
 * at cs.cmu.edu for the swapper __spawnv(). Since putenv() can fragment
 * the environment, it is necessary to build a private copy of the
 * environment and pass it to __spawnv(). Along the way it is cheap to
 * mung the PROMPT to emit an EXIT message (a la 5.1 WordPerfect). Calls
 * to putenv() are recognized but putenv() is not used.
 * This works with spawno20, spawno30, spawno40. Not tested with others.
 *
 * Env Variables:       PROMPT          dos prompt
 *                      COMSPEC         command.com path
 *                      SWAPDRV         where to swap application
 * Externals:
 *                      environ         Turbo-C enviroment **-pointer
 *                      malloc,free,    std library routines
 *                      strcpy,strcat,
 *                      strncmp,getenv,
 *                      strlen
 *                      __spawnv        Ralf Brown's library function
 *
 * The function __spawnv() is Copyright 1990 by Ralf Brown. Ralf does not
 * supply source for the swapper, but he does make the library available
 * free. Get it from wuarchive.wustl.edu anonymous FTP, /mirrors/msdos/c,
 * in file spawno30.zip. I am using the main function __spawnv() from
 * SPAWNS.LIB, in SPAWN_TC.ZIP, extracted as SPAWNVS.OBJ (renamed from
 * the original _spawnvs.obj). Note: MS-C also supported in SPAWNO30.ZIP.
 * For the medium model use _spawnvm.obj. I prefer to use spawno20,
 * because it does not use EMS, only XMS and disk. On modern hardware,
 * EMS exists only because of EMM386, which is generally loaded because
 * of windows, and already himem.sys is there to allow it to happen.
 * In short, XMS is the first thing installed.
 */
/*#include <dos.h>*/
#include <stdarg.h>

extern int pascal
__spawnv(const char *overlay_path,const char *name,va_list args,int env) ;

#define PROMPT "PROMPT=Enter EXIT to resume PI$_$n$g"

static char far *mem;

static void mv(p) char *p; {
   while((*mem++ = *p++)!=0) ;
}

#undef system
void mysystem(s) char *s; {
char **v;
char *p,*q;
char *env;
char file[128];
char tail[128];
char *argv[3];
int envseg;
unsigned int length;
extern char *getenv();
extern char **environ;          /* Turbo-C env pointer */
extern int swapswitch;
struct paramblock {unsigned envseg; char far * cmdline;
char far *fcb1; char far *fcb2; unsigned
childSP,childSS,childIP,childCS;};
struct paramblock pb;
unsigned char far *tmp; unsigned envv;

  if((q=getenv("COMSPEC"))==0) q=SHELL;
  strcpy(argv[0]=file,q);
  if(!swapswitch){ system((s[0]==' ' && s[1]==0) ? file : s); return;}

   argv[1]=argv[2]=(char *)0;
   length=16+1+sizeof(PROMPT);
   v=environ;
   while(v[0] && v[0][0]) length += strlen(*v++) + 1 ;
   env=(char *)malloc(length);
   envseg = FP_SEG(env) + ((FP_OFF(env)+15)>>4) ;
   mem = MK_FP(envseg,0) ;
   v=environ;
   while(v[0] && v[0][0]){
     if(strncmp(p = *v++,"PROMPT=",7)) mv(p);
   }
   mv(PROMPT);
   mem[0] = '\0' ;

  if(s[0] && s[0]!=' ') ssprintf(argv[1]=tail,"/c %s",s);
  if((p=getenv("SWAPDRV"))==0) p="c:\\";
  __spawnv(p,file,argv,envseg);
  free(env);
}

#endif

void strip(p) char *p; {
asm     mov     ax,ds
asm     mov     es,ax
asm     mov     di,p
asm     xor     al,al
asm     mov     cx,0ffffh
asm     cld
asm     repnz   scasb
asm     dec     di              /* point at null */
asm     mov     cx,di
asm     sub     cx,p            /* cx=string length */
L1:
asm     or      cx,cx
asm     jz      L2
asm     dec     cx
asm     dec     di
asm     cmp     byte ptr es:[di],32
asm     je      L1
asm     inc     di
L2:
asm     mov     byte ptr es:[di],0
}

/*
 * UNSETBITS
 *
 * Strips parity bit from a string.
 */
void unsetbits(s) char *s;{
register char *p;
  p=s;
  while(*p){ if((127 & p[0]) != 0) p[0] &= 127; ++p;}
}

int startof(s,k)
char far *s;
int k;
{
  asm   les     di,dword ptr s
  asm   mov     cx,word ptr k
  asm   xor     ax,ax
  asm   cmp     cx,ax
  asm   jle     quit
  asm   add     di,cx
mainloop:
  asm   dec     di
  asm   cmp     al,byte ptr es:[di]
  asm   jz      quitplus
  asm   loop    mainloop
quitplus:
  asm   mov     ax,cx
quit:
return (_AX);
}

/* Search source string s=ss+n[0] for a match to srcharg[] */
/* Returns string index of match >=1 or 0 for failure */
/* Array n[] is updated on failure to allow search of next string */
/* in the buffer (assume buffer full of 0-delimited strings). */
/* May 1995: Make search string ASCII 128 match end of line char */

#if 1
int midstr(sss,n)
char far *sss;                /* source string */
int *n;                       /* offset to end of string on fail */
{                             /* offset to start of string on entry */
char far *s;
int nn,mm,xl;
extern int taberr; extern int zeroleft,zeroright;
extern char *lbuf;
  s=sss+n[0]; nn=0;
  if(taberr && mydotabs(s)) {nn=farstrlen(s); sdetab(s,lbuf+1); s=lbuf+1;}
  if(zeroleft){ /* srcharg[] has a zero token at the start */
        xl=strc(s);  /* If xl==1, a match from beginning of line */
        mm= xl ? 0 : farstrlen(s); /* Offset to match or end of str */
  } else  xl=strd(s,&mm);
  n[0] += (nn ? nn+1 : mm+1);
  return (xl ? mm+1 : 0);
}
#endif

#if 0
int farstrlen(s) char far *s; {
register int i;
i=0; while(s[i++]); return i-1;
}
#endif

#if 1
int farstrlen(s) char far *s; {
asm     les     di,s
asm     xor     al,al
asm     xor     cx,cx
asm     dec     cx
asm     repnz   scasb   /* scan until cx=0 or [di] == al */
asm     mov     ax,-2
asm     sub     ax,cx
return (_AX);
}
#endif

#if 0
/* Return 1 if it matches, 0 if it fails to match */
/* Don't mix and match the ASM versions with these! */
int strc(d) char far *d;{
char *s; int x,z;
extern int zero,zeroleft; extern int FOLDCASE; extern char *srcharg;
s=srcharg;
if(zeroleft) ++s; /* First zero matches beginning of line */
while((z= *s++) != 0) {
  x = *d++;
  if(x==0){             /* Reached end of search string */
    if(z==zero) break;  /* Matched */
    /* Otherwise, z != 0 implies z != x and we exit below */
  }
  if(FOLDCASE) x = fold(x);
  if(x!=z) return 0;   /* Failed to match */
}
return 1;       /* It matched */
}
#endif

#if 1
int strc(d) char far *d; {
extern char *srcharg; extern int zero; extern int FOLDCASE;
asm     les     di,dword ptr d
asm     mov     dx,FOLDCASE
asm     mov     si,srcharg
asm     cmp     zeroleft,0
asm     je      loop
asm     inc     si
loop:
asm     mov     ah,byte ptr [si]
asm     inc     si
asm     or      ah,ah
asm     jz      matched                 /* End of search string */
strc5:
asm     mov     al,byte ptr es:[di]
asm     inc     di
asm     or      al,al                   /* End of string? */
asm     jnz     strc5x
asm     mov     al,ah
asm     cbw
asm     cmp     ax,zero                 /* z==zero? */
asm     je      matched                 /* Got a match */
asm     jmp     short mismatched        /* No match */
strc5x:
asm     or      dl,dl                   /* if(FOLDCASE==0) */
asm     jz      strc6
asm     test    al,128                  /* Signed? */
asm     jnz     strc6
asm     cmp     al,'a'
asm     jl      strc6
asm     cmp     al,'z'
asm     jg      strc6
asm     and     al,11011111B            /* fold case to upper */
strc6:
asm     cmp     al,ah                   /* Do they match? */
asm     jz      loop
mismatched:
asm     xor     ax,ax
asm     jmp     short quit
matched:
asm     mov     ax,1
quit:
return (_AX);
}
#endif

#if 0
/* Return 1 if it matches, 0 if it fails to match */
/* Updates n[0] to be the offset to the string match or end of string */
int strd(d,n) char far *d; int *n; {
char far *d1; int flag;
register int x,z; int xl;
extern int zero,zeroleft; extern int FOLDCASE; extern char *srcharg;
 flag=0; d1=d;
 x= zeroleft ? srcharg[1] : srcharg[0]; /* x is already uppercase */
 xl= (FOLDCASE && 'A'<=x && x<='Z') ? (x+' ') : x;
  /* First zero matches beginning of line */
 while((z=d1[0])!=0){
  if((x==z || xl==z) && (flag=strc(d1)) != 0) break; ++d1;
 }
 n[0]=(int)(d1-d);
 return flag;
}
#endif


#if 1
int strd(d,n) char far *d; int *n; {
extern char *srcharg; extern int FOLDCASE; extern int zero,zeroleft;
asm     mov     bx,srcharg
asm     cmp     zeroleft,0
asm     je      strd0
asm     inc     bx
strd0:
asm     mov     ch,byte ptr [bx]        /* Already in uppercase */
asm     mov     cl,ch                   /* x=ch, xl=cl */
asm     cmp     FOLDCASE,0
asm     jz      strd1
asm     test    cl,128                  /* Signed? */
asm     jnz     strd1
asm     cmp     cl,'A'
asm     jl      strd1
asm     cmp     cl,'Z'
asm     jg      strd1
asm     add     cl,' '                  /* xl==cl is in lowercase */
strd1:
asm     les     bx,dword ptr d
loop:
asm     mov     al,byte ptr es:[bx]
asm     or      al,al
asm     je      nomatch                 /* At end of string */
asm     cmp     al,cl
asm     je      strd2
asm     cmp     al,ch
asm     jne     strd3
strd2:
asm     push    es
asm     push    bx
asm     call    strc            /* Doesn't use cx or bx */
asm     pop     bx
asm     pop     es
asm     or      ax,ax           /* 1=match, 0=nomatch */
asm     jnz     match           /* Got  match */
strd3:
asm     inc     bx
asm     jmp     short loop
nomatch:
asm     xor     dx,dx
asm     jmp     short quit
match:
asm     mov     dx,1
quit:
asm     les     cx,d
asm     mov     ax,bx
asm     sub     ax,cx
asm     mov     bx,n
asm     mov     word ptr [bx],ax        /*  n[0]=(int)(d1-d); */
asm     mov     ax,dx                   /*  return flag; */
return (_AX);
}
#endif

int strccmp(d,s,n) char *d,*s; int n; {
int i;
int fold();
while((n-- > 0) && (fold(*d) == *s)) {
  if(*d == '\0' || n == 0) return 0;
  ++d; ++s;
}
return (fold(*d) - *s);
}

int fold(c) int c; {
return (('a' <= c && c <= 'z') ? (c - ' ') : c);
}

/*
 * MYDOTABS
 *
 * Returns 1 if the line needs detabbing, 0 otherwise.
 */
int mydotabs(buf) char far *buf; {
asm     les     di,dword ptr buf
asm     mov     ah,9
tabloop:
asm     mov     al,byte ptr es:[di]
asm     inc     di
asm     or      al,al
asm     jz      notab
asm     cmp     al,ah
asm     jnz     tabloop
return 1;
notab:
return 0;
}

/*
 * DOTABS
 *
 * Returns 1 if the line needs detabbing
 */
int dotabs(buf) char *buf; {
char *strchr();
 if(strchr(buf,'\t') != (char *)0) return 1;
 return 0;
}

void sdetab(s,d)
char far *s; char *d;
{
register int lc; int tc;
  lc=0;
  while(s[0]){
    if(s[0] != '\t') {*d++ = *s++; ++lc;}
    else {
      tc=(lc&(STOPS-1)); tc = tc ? STOPS-tc : STOPS;
      lc += tc; while(tc--) *d++ = ' '; ++s;
    }
  }
  *d=0;
}

/*
 * DETAB
 *
 * Converts tabs on the fly, writing over the input buffer as
 * needed when the conversion is incomplete. Returns the next
 * buffer location to be processed.
 */
char *detab(source,dest)
char *source,*dest;
{
register char *p,*q;
register int lc,tc;
  p=source; q=dest; lc=0;
  while(p[0]){
    if(p[0] != '\t') {if(++lc >= MAXLN-1) break; *q++ = *p++; }
    else {
      tc=(lc&(STOPS-1)); tc = tc ? STOPS-tc : STOPS;
      if(lc+tc>=MAXLN-1) break;
      lc += tc; while(tc--) *q++ = ' '; ++p;
    }
  }
  *q=0;
  if(p[0]) return p;
  return p+1;
}

int strstart(s) char *s; {

asm     mov     si,s
S3:
asm     dec     si
asm     mov     al,byte ptr [si]
asm     or      al,al
asm     jne     S3
asm     mov     ax,si
return (_AX);
}

char *endof(s) char *s; {
asm     mov     si,s
L1:
asm     mov     al,byte ptr [si]
asm     or      al,al
asm     je      L2
asm     inc     si
asm     jmp     SHORT L1
L2:
asm     mov     ax,si
return (char *)(_AX);
}

char * sob(p)
char *p;
{
asm     mov     si,p
N3:
asm     mov     al,byte ptr [si]
asm     cmp     al,32
asm     jne     N4
asm     inc     si
asm     jmp     short N3
N4:
asm     mov     ax,si
return (char *)(_AX);
}

char * fnb(p)
char *p;
{
asm     mov     si,p
N7:
asm     mov     al,byte ptr [si]
asm     cmp     al,32
asm     je      N8
asm     or      al,al
asm     je      N8
asm     inc     si
asm     jmp     short N7
N8:
asm     mov     ax,si
return (char *)(_AX);
}

int max(a,b)
int a,b;
{
asm     mov     ax,a
asm     cmp     ax,b
asm     jge     N9
asm     mov     ax,b
N9:
return (_AX);
}

int min(a,b)
int a,b;
{
asm     mov     ax,a
asm     cmp     ax,b
asm     jle     N11
asm     mov     ax,b
N11:
return (_AX);
}

#if 0
unsigned minu(a,b)
unsigned a,b;
{
asm     mov     ax,a
asm     cmp     ax,b
asm     jae     N12
asm     jmp     short N11
N12:
asm     mov     ax,b
N11:
return (_AX);
}
#endif

int abs(a)
int a;
{
asm     mov     ax,a
asm     cmp     ax,0
asm     jge     N13
asm     neg     ax
N13:
return (_AX);
}

void fillchr(d,c,n)
char *d; char c; unsigned n;
{
asm     mov     si,d
asm     mov     al,c
asm     mov     bx,n
N17:
asm     or      bx,bx
asm     je      N18
asm     mov     byte ptr [si],al
asm     inc     si
asm     dec     bx
asm     jmp     short N17
N18:
;
}

int strmov(d,s)         /* like strcpy. Copy null. ret count. */
char *d; char *s;
{
asm     mov     si,d
asm     mov     di,s
asm     mov     ax,0
loop:
asm     mov     cl,byte ptr [di]
asm     mov     byte ptr [si],cl
asm     inc     si
asm     inc     di
asm     inc     ax
asm     or      cl,cl
asm     jnz     loop
asm     dec     ax              /* ret count of non-null chars */
return (_AX);
}

int isupper(z) int z; {
asm     mov     bx,0
asm     mov     ax,z
asm     cmp     ax,'A'
asm     jl      quit
asm     cmp     ax,'Z'
asm     jg      quit
asm     inc     bx
quit:
asm     mov     ax,bx
return (_AX);
}

int islower(z) int z; {
asm     mov     bx,0
asm     mov     ax,z
asm     cmp     ax,'a'
asm     jl      quit
asm     cmp     ax,'z'
asm     jg      quit
asm     inc     bx
quit:
asm     mov     ax,bx
return (_AX);
}

int toupper(z) int z; {
asm     mov     ax,z
asm     mov     bx,ax
asm     cmp     ax,'a'
asm     jl      quit
asm     cmp     ax,'z'
asm     jg      quit
asm     sub     bx,' '
quit:
asm     mov     ax,bx
return (_AX);
}

int tolower(z) int z; {
asm     mov     ax,z
asm     mov     bx,ax
asm     cmp     ax,'A'
asm     jl      quit
asm     cmp     ax,'Z'
asm     jg      quit
asm     add     bx,' '
quit:
asm     mov     ax,bx
return (_AX);
}

int tailchar(s) char *s; {
register int j;
  return ((j=strlen(s))==0 ? 0 : s[j-1]);
}

char *toprint(p) char *p; {
 if(strlen(p)<curoff) p=""; else p += curoff;
 return p;
}

#if 0
char *strchr(s, c) char *s; char c; {
   while (*s){ if (*s == c) return(s); s++; } return(0);
}

char *strrchr(s, c) char *s; char c; {
int n;
  n=strlen(s);
  while(n){
    if(s[--n]==c) return (s+n);
  }
  return(0);
}
#endif

char *strchr(s, c) char *s; char c; {
/*   while (*s){ if (*s == c) return(s); s++; } return(0); */
asm     mov     ah,byte ptr c
asm     mov     si,word ptr s
loop:
asm     mov     al,byte ptr [si]
asm     or      al,al
asm     jz      fail
asm     cmp     al,ah
asm     je      success
asm     inc     si
asm     jmp     short loop
fail:
asm     xor     ax,ax
asm     jmp     short quit
success:
asm     mov     ax,si
quit:
return (char *)(_AX);
}

char *strrchr(s, c) char *s; char c; {
  _AX = strlen(s);
asm     mov     si,s
asm     add     si,ax
asm     mov     bl,c
/*  while(n){ if(s[--n]==c) return (s+n); } */
loop:
asm     or      ax,ax
asm     jz      fail
asm     dec     ax
asm     dec     si
asm     cmp     bl,byte ptr [si]
asm     jne     loop
succeed:
asm     mov     ax,si
asm     jmp     short quit
fail:
/*  return(0); */
asm     xor     ax,ax
asm     jmp     short quit
quit:
return (char *)(_AX);
}

strncmp(s,t,n) char *s,*t; int n; {
/*
  while(n-- && s[0]==t[0]){ if(!s[0] || !n) return 0; ++s; ++t;}
  return ((unsigned char)s[0]-(unsigned char)t[0]);
*/
asm     mov     cx,n
asm     mov     si,s
asm     mov     di,t
loop:
asm     cmp     cx,0
asm     jz      end
asm     dec     cx
asm     mov     al,byte ptr [si]
asm     cmp     al,byte ptr [di]
asm     jne     end
asm     cmp     cx,0
asm     je      zero
asm     cmp     al,0
asm     je      zero
asm     inc     si
asm     inc     di
asm     jmp     loop
zero:
asm     mov     ax,0
asm     jmp     exit
end:
asm     mov     al,byte ptr [si]
asm     mov     dl,byte ptr [di]
asm     mov     ah,0
asm     mov     dh,ah
asm     sub     ax,dx
exit:
return (_AX);
}

int strcmp(s,t) char *s,*t; {
/*
  while(s[0] && s[0]==t[0]){ ++s; ++t;}
  return ((unsigned char)s[0]-(unsigned char)t[0]);
*/

asm     mov     si,s
asm     mov     di,t
loop:
asm     mov     al,byte ptr [si]
asm     cmp     al,0
asm     je      end
asm     cmp     al,byte ptr [di]
asm     jne     end
asm     inc     si
asm     inc     di
asm     jmp     loop
end:
asm     mov     al,byte ptr [si]
asm     mov     dl,byte ptr [di]
asm     mov     ah,0
asm     mov     dh,ah
asm     sub     ax,dx
return (_AX);
}

char *lltoa(n,str) LLONG n; char *str;
{
LLONG i;
char *p;
char q[61];
char flag;
  *(p = &q[60]) = '\0';
  flag=(n<0);
  i = flag ? -n : n;
  do { *--p = '0' + i%10; } while( (i /= 10) > 0);
  if(flag) *--p = '-';
  strcpy(str,p);
  return str;
}

LLONG atoiLL(s) char *s; {
LLONG i;
char flag;
extern char *sob();
  s=sob(s);
  flag=i=0;
  if(s[0]=='+') ++s; else
  if(s[0]=='-'){++s; flag=1;}
  while(isdigit(s[0])) i = 10*i + (*s++) -'0';
  return (flag ? -i : i);
}

#endif          /* TURBOC */


#if !TURBOC || (DOS386I && !IBMHARDWARE)
/*
 * Routines below are entirely in C so that
 * they become portable across machines. Most
 * non-portable code for MS-DOS appears above.
 */

#if 1
int atoiLL(ss) char *ss; {
int i;
char flag;
extern char *sob();
extern int isdigit();
char *s;
  s=sob(ss);
  flag=i=0;
  if(s[0]=='+') ++s; else
  if(s[0]=='-'){++s; flag=1;}
  while(isdigit(s[0])) i = 10*i + (*s++) -'0';
  return (flag ? -i : i);
}
#endif

char *strchr(s, c) char *s; char c; {
   while (*s){ if (*s == c) return(s); s++; } return(0);
}

char *strrchr(s, c) char *s; char c; {
int n;
  n=strlen(s);
  while(n){
    if(s[--n]==c) return (s+n);
  }
  return(0);
}

char *lltoa(n,str) LLONG n; char *str;
{
LLONG i;
char *p;
char q[61];
char flag;
  *(p = &q[60]) = '\0'; flag=(n<0);
  i = (flag ? -n : n);
  do { *--p = '0' + i%10; } while( (i /= 10) > 0);
  if(flag) *--p = '-';
  strcpy(str,p);
  return str;
}

void ePrint(p)       /* no crlf pair issued */
char *p;
{
static
int x,i,j;
extern int COLMAX,nodisp;
extern int min();
char s[1024];
  if(!nodisp) {
    i=j = 0;
    while((p[j] & 96) != 0) ++j;
    if(p[j] && !(p[j] & 96)) j = min(strlen(p),COLMAX+1);
    else
    if(j>COLMAX) j = COLMAX+1;
    while(j--) {
      if((x = *p++) & 96) s[i++] = x; else {
        if((x & 128)!=0) s[i++] = x; else {
        strcpy(&s[i],REVVIDEO); i += strlen(REVVIDEO);
        s[i++]= x + '@';
#if BSDunix || SYS5
        if(UNIXtextfgbg[0]==0) {
          strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
        } else {
          strcpy(&s[i],UNIXtextfgbg); i += strlen(UNIXtextfgbg);
        }
#else
        strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
#endif
        }
      }
    }
    s[i]=0; etype(s);
  }
}

int eePrint(p,s)      /* issue crlf pair before string print */
register char *p; register char *s;
{
int x,i,j;
extern int COLMAX,nodisp;
extern int min();
extern char *skipdelay();
  if(!nodisp) {
    i = j = 0;
    s[i++]= '\r'; s[i++]= '\n';
    while((p[j] & 96) != 0) ++j;
    if(p[j] && !(p[j] & 96)) j = min(strlen(p),COLMAX+1);
    else
    if(j>COLMAX) j = COLMAX+1;
    while(j--) {
      if((x = *p++) & 96) s[i++] = x; else {
        if((x & 128)!=0) s[i++] = x; else {
        strcpy(&s[i],REVVIDEO); i += strlen(REVVIDEO);
        s[i++]= x + '@';
#if BSDunix || SYS5
        if(UNIXtextfgbg[0]==0) {
          strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
        } else {
          strcpy(&s[i],UNIXtextfgbg); i += strlen(UNIXtextfgbg);
        }
#else
        strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
#endif
        }
      }
    }
    strcpy(&s[i],skipdelay(ce_termcap));
    i += strlen(skipdelay(ce_termcap));
    s[i]=0;
  }
  return i;
}

void strinvvideo(s) char *s; {
  etype(REVVIDEO);
  etype(s);
#if BSDunix || SYS5
  if(UNIXtextfgbg[0]==0) etype(NORMVIDEO); else etype(UNIXtextfgbg);
#else
  etype(NORMVIDEO);
#endif
}

char *drawbuff=(char *)0;

void setdrawbuff(){
int n; extern char *core();
  freeup(drawbuff);
  n=(strlen(REVVIDEO)+strlen(NORMVIDEO)+1)*(3+COLMAX)*(1+ROWMAX);
  drawbuff=core(n);
}

#include <errno.h>
extern int errno;
void drawall()
{
register
int i,n;
extern char **line; extern char *toprint();
extern int nodisp,halfdone;
extern FILE *termout;

if(testkey()) {nodisp=halfdone=1; return;}
if(!nodisp){
  n=0;
  for(i=0;i<=ROWMAX;++i) {
    n += eePrint(toprint(line[i]),&drawbuff[n]);
  }
  lcount();
  home();
#if BSDunix || SYS5
    write(fileno(termout),drawbuff,n);
    fflush(termout);
#endif
#if MSDOS
    etype(drawbuff);
#endif
  halfdone=0;
  curses();
  }
}

void xPrint(p,rrow,ccol) char *p; int rrow,ccol; {
static
int i,x,j;
extern int COLMAX,nodisp;
extern int min();
char s[1024];
  if(!nodisp) {
    if(ccol > COLMAX+1) ccol = COLMAX+1;
    i=j = 0;
    while((p[j] & 96) != 0) ++j;
    if(p[j] && !(p[j] & 96)) j = min(strlen(p),COLMAX+1-ccol);
    else
    if(j>COLMAX-ccol) j = COLMAX+1-ccol;
    putcursor(rrow,ccol);
    while(j--) {
      x = *p++;
      if(x & 96) s[i++] = x; else {
        if((x & 128)!=0) s[i++] = x; else {
        strcpy(&s[i],REVVIDEO); i += strlen(REVVIDEO);
        s[i++]= x + '@';
#if BSDunix || SYS5
        if(UNIXtextfgbg[0]==0) {
          strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
        } else {
          strcpy(&s[i],UNIXtextfgbg); i += strlen(UNIXtextfgbg);
        }
#else
        strcpy(&s[i],NORMVIDEO); i += strlen(NORMVIDEO);
#endif
        }
      }
    }
    s[i]=0; etype(s);
  }
}

void setcursortype() {       /* do nothing for DOS386I */
}

void equipment(){            /* do nothing for DOS386I */
}

#if !DOS386I
int startof(s,k) char far *s; int k; {
  while(k){ if(s[k-1]==0) break; --k;}
return k;
}

void sdetab(s,d)
char *s; char *d;
{
register int lc,tc;
  lc=0;
  while(s[0]){
    if(s[0] != '\t') {*d++ = *s++; ++lc;}
    else {
      tc=(lc&(STOPS-1)); tc = tc ? STOPS-tc : STOPS;
      lc += tc; while(tc--) *d++ = ' '; ++s;
    }
  }
  *d=0;
}

/* Search source string s=ss+n[0] for a match to srcharg[] */
/* Returns string index of match >=1 or 0 for failure */
/* Array n[] is updated on failure to allow search of next string */
/* in the buffer (assume buffer full of 0-delimited strings). */
/* May 1995: Make search string ASCII 128 match end of line char */

int midstr(sss,n)
char far *sss;                /* source string */
int *n;                       /* offset to end of string on fail */
{                             /* offset to start of string on entry */
char far *s;
int nn,mm,xl;
extern int taberr; extern int zeroleft,zeroright;
extern char *lbuf;
  s=sss+n[0]; nn=0;
  if(taberr && dotabs(s)) {nn=strlen(s); sdetab(s,lbuf+1); s=lbuf+1;}
  if(zeroleft){ /* srcharg[] has a zero token at the start */
        xl=strc(s);  /* If xl==1, a match from beginning of line */
        mm= xl ? 0 : strlen(s); /* Offset to match or end of str */
  } else  xl=strd(s,&mm);
  n[0] += (nn ? nn+1 : mm+1);
  return (xl ? mm+1 : 0);
}

/* Return 1 if it matches, 0 if it fails to match */
int strc(d) char far *d;{
char *s; int x,z;
extern int zero,zeroleft; extern int FOLDCASE; extern char *srcharg;
s=srcharg;
if(zeroleft) ++s; /* First zero matches beginning of line */
while((z= *s++) != 0) {
  x = *d++;
  if(x==0){             /* Reached end of search string */
    if(z==zero) break;  /* Matched */
    /* Otherwise, z != 0 implies z != x and we exit below */
  }
  if(FOLDCASE && 'a'<=x && x<='z') x -= ' ';
  if(x!=z) return 0;   /* Failed to match */
}
return 1;       /* It matched */
}

/* Return 1 if it matches, 0 if it fails to match */
/* Updates n[0] to be the offset to the string match or end of string */
int strd(d,n) char far *d; int *n; {
register char far *d1; int flag;
register int x,xl,z;
 flag=0; d1=d;
 x= zeroleft ? srcharg[1] : srcharg[0]; /* x is already uppercase */
 xl= (FOLDCASE && 'A'<=x && x<='Z') ? (x+' ') : x;
  /* First zero matches beginning of line */
 while((z=d1[0])!=0){
  if((x==z || xl==z) && (flag=strc(d1)) != 0) break; ++d1;
 }
 n[0]=(int)(d1-d);
 return flag;
}

int strccmp(d,s,n) char *d,*s; int n; {
int fold();
while((n-- > 0) && (fold(*d) == *s)) {
  if(*d == '\0' || n == 0) return 0;
  ++d; ++s;
}
return (fold(*d) - *s);
}

int fold(c) int c; {
return (('a' <= c && c <= 'z') ? (c - ' ') : c);
}

char *strstart(q)
char *q;
{
  while(*--q) ;
  return (q);
}

char *endof(p)
char *p;
{
  return (p + strlen(p));
}

void strip(p) char *p; {
register
int i;
  i = strlen(p);
  while(--i >= 0) {
    if(p[i] != ' ') break;
  }
  p[i+1] = 0;
}

/*
 * DETAB
 *
 * Converts tabs on the fly, writing over the input buffer as
 * needed when the conversion is incomplete. Returns the next
 * buffer location to be processed.
 */
char *detab(source,dest)
char *source,*dest;
{
register char *p,*q;
register int lc,tc;
  p=source; q=dest; lc=0;
  while(p[0]){
    if(p[0] != '\t') {if(++lc >= MAXLN-1) break; *q++ = *p++; }
    else {
      tc=(lc&(STOPS-1)); tc = tc ? STOPS-tc : STOPS;
      if(lc+tc>=MAXLN-1) break;
      lc += tc; while(tc--) *q++ = ' '; ++p;
    }
  }
  *q=0;
  if(p[0]) return p;
  return p+1;
}

/*
 * DOTABS
 *
 * Returns 1 if the line needs detabbing, 0 otherwise.
 */
int dotabs(buf) char *buf; {
char *strchr();
 if(strchr(buf,'\t') != (char *)0) return 1;
 return 0;
}

/*
 * UNSETBITS
 *
 * Strips parity bit from a string.
 */
void unsetbits(s) char *s;{
register char *p;
  p=s;
  while(*p){ if((127 & p[0]) != 0) p[0] &= 127; ++p;}
}

char *toprint(p) char *p; {
 if(strlen(p)<curoff) p=""; else p += curoff;
 return p;
}

int strmov(d,s)         /* like strcpy. Copy null. ret count. */
register char *d,*s;    /* d=destination, s=source */
{
register int i;
i=0;
while((*d++ = *s++) != 0) ++i;
return i;     /* ret count of non-null chars */
}


int tailchar(s) char *s; {
int j;
  return ((j=strlen(s))==0 ? 0 : s[j-1]);
}

#endif          /* end !DOS386I */
#endif          /* end !TURBOC || (DOS386I && !IBMHARDWARE) */

#if !HASISUPPER
/*
 * ISUPPE
 *
 * Purpose:
 *              Test for upper case character.
 */
#undef isupper
#undef islower
int isupper(c)
int c;
{
return ('A' <= c && c <= 'Z');
}

int islower(c)
int c;
{
return ('a' <= c && c <= 'z');
}
#endif

#if !HASTOUPPER
/*
 * TOUPPE
 *
 * Purpose:
 *              Convert character to upper case.
 */
#undef toupper
int toupper(z)
int z;
{
return ( islower(z) ? z - ' ' : z);
}

int tolower(z)
int z;
{
return ( isupper(z) ? z + ' ' : z);
}
#endif

#if 0           /* No longer used. See ssprintf. */
/*
 * UTOA
 *
 * Purpose:
 *              Convert integer to ascii string.
 *              Standard C library function.
 */
char *utoa(n,str) LONG n; char *str;
{
LONG i;
char *p;
char q[61];
  i = n; *(p = &q[60]) = '\0';
  do { *--p = '0' + i%10; } while( (i /= 10) > 0);
  strcpy(str,p);
  return str;
}

#endif


#if !HASMAXMIN
/*
 * MAX
 *
 */
#undef max
int max(a,b)
int a,b;
{
return ( a < b ? b : a);
}

/*
 * MIN
 *
 */
#undef min
int min(a,b)
int a,b;
{
return ( a < b ? a : b);
}

#endif

#if !HASABS
/*
 * ABS
 *
 */
#undef abs
int abs(a)
int a;
{
return (a > 0 ? a : -a);
}
#endif

#if !HASFILLCHR
void fillchr(d,c,n)
char *d; char c; unsigned n;
{
     while(n--) *d++ = c;
}
#endif

#if !HASCVUPPER
void cvupper(s)
char *s;
{
while(*s) { *s = toupper(*s); ++s;}
}
#endif
