/*
 * V2.C
 *
 */

#define KEYDEFS 1       /* include key defs from v.h */
#include "v.h"

#if POPUPS
/* popup data and routines; see end of file for basics */
/*  popup menu without overlaps, looks like this, essentially: */
#if 0
Keymap:^K^L
 Pi Editor ͻ
  File  Keys  Tools  Search  Edit  Word  Block  Values  Other  Help  
ͼ
  ͻ
  Open            
  New             
  Reload          
  Rename          
  Save to disk    
  Save as ...     
  Save then exit  
  Save then close 
  Close           
  Inject file     
  Paste directory 
  Print file      
  Find window     
  Quit editor     
  ͼ
#endif
/* The Tools display can be 57 characters wide, leaving 54 for data */
/* Tools items after those compiled in come from PI.SET macros. */

enum COLORS {
        BLACK,                  /* dark colors, background, 0-7 */
        BLUE,
        GREEN,
        CYAN,
        RED,
        MAGENTA,
        BROWN,
        LIGHTGRAY,
        DARKGRAY,               /* light colors, 8-15 */
        LIGHTBLUE,              /* background cannot be 8-15 */
        LIGHTGREEN,             /* foreground can be 0-15 */
        LIGHTCYAN,
        LIGHTRED,
        LIGHTMAGENTA,
        YELLOW,
        WHITE
};

#define REVERSEVIDEO    0x70    /* reverse video attribute */

/*
 Pi Editor ͻ
  File  Keys  Tools  Search  Edit  Word  Block  Values  Other  Help  
ͼ
*/

struct menu { char *name; unsigned keycode; };

static struct menu File[]={
{"Open",            /*  "^KL", */          LOAD},
{"New",             /*  "^KU", */          NEWLOAD},
{"Reload",          /*  "^KE", */          RELOAD},
{"Rename",          /*  "^KN", */          NAMECHANGE},
{"Save to disk",    /*  "^KS", */          BACKUP},
{"Save as ...",     /*  "^KW", */          SAVEASKEY},
{"Save then exit",  /*  "^KX", */          SAVE_EXIT},
{"Save then close", /*  "^KD", */          SAVE_NEXT},
{"Close",           /*  "^KQ", */          SCRAPBUFFER},
{"Inject file",     /*  "^KR", */          INJECT},
{"Paste directory", /*  "^KF", */          DSKFILES},
{"Print file",      /*  "^KP", */          PRINTFILE},
{"Find window",     /*  "^K^A", */         SETBUFFER},
{"Quit editor",     /*  "Alt-X", */        TSRKEY},
{0,0}
};


static struct menu Edit[]={
{"Goto line # ",      /*  "^OG", */          GOTOKEY},
{"Focus at cursor",   /*  "^O p", */         CENTERREFRESH},
{"Goto bookmark",     /*  "^Q0", */          GOBOOKMARK},
{"Set Bookmark",      /*  "^K0", */          SETBOOKMARK},
{"Center line",       /*  "ALT-C", */        CENTERTEXT},
{"Justify line",      /*  "^B", */           REFORMKEY},
{"Split line",        /*  "^ON", */          SPLITLINE},
{"Insert line",       /*  "^N  or F3", */    INSERTLINE},
{"Delete line",       /*  "^Y  or F4", */    DELLINE},
{"Kill to end",       /*  "^QY", */          ERASEKEY},
{"Erase # lines",     /*  "^OY", */          DELMANYLINES},
{"Copy line",         /*  "^KC or F9", */    PUSHKEY},
{"Paste line",        /*  "^KV or F10", */   POPKEY},
{"Scroll ahead",      /*  "^Z or ^PgDn", */  LINEFORWARD},
{"Scroll back",       /*  "^W or ^PgUp", */  LINEBACK},
{"Recover lines",     /*  "Shift-F10", */    POPLIST},
{"Memory purge",      /*  "Shift-F10", */    PURGELIST},
{0,0}
};

static struct menu Word[]={
{"Next word",                 /*  "^F", */        WORDAHEAD},
{"Previous word",             /*  "^A", */        WORDBACK},
{"Delete word",               /*  "alt-G", */     WORDERASE},
{"Capitalize word",           /*  "^O,C", */      CASECAPITAL},
{"Lowercase word",            /*  "^O,L", */      CASELLOWER},
{"Uppercase word",            /*  "^O,U", */      CASELUPPER},
{"Fill phrase string",        /*  "^OC", */       QPUSHKEY},
{"Paste phrase string",       /*  "^OV", */       QPOPKEY},
{"Kill phrase match",         /*  "^OZ", */       QZAPKEY},
{"Copy phrase to notepad",    /* "^OP", */        OPTPUSHKEY},
{"IntMath on phrase",         /* "^O,i", */       OPTMATH},
{"HexToBase10 on phrase",     /* "^O,H", */       OPTHEX},
{"OctToBase10 on phrase",     /* "^O,H", */       OPTOCT},
#if MATHFLOAT
{"FloatMath on phrase",       /* "^O,j", */       OPTMATHFLOAT},
{"SciCalc on phrase",         /* "^O,k", */       OPTEVALEXPR},
{"FloatMath format",          /* "^OM", */        MATHFORMAT},
#endif
{0,0}
};


static struct menu Block[]={
{"Begin block set mark",      /* "^KB", */        SETMARK},
{"Swap mark with cursor",     /* "^KK", */        SWAPMARK},
{"Reveal block setup",        /* "^K,s", */       REVEALMARK},
{"Copy block to notepad",     /* "^K,c", */       COPYBLOCK},
{"Move block to notepad",     /* "^K,m", */       MOVEBLOCK},
{"Paste block from notepad",  /* "^K,p", */       PASTEBLOCK},
{"Erase the notepad",         /* "^OX", */        KILLNOTEPAD},
{"Dup line onto notepad",     /* "^OA", */        APPENDLINE},
{"Kill line, add on notepad", /* "^OK", */        ADDBUFFER},
{0,0}
};


static struct menu Search[]={
{"Set search string",   /* "^QN", */                 NEWSEARCH},
{"Forward search",      /* "^Of or F5", */           FWDSEARCH},
{" ... within Line",    /* "^OF or   ", */           LFWDSEARCH},
{"Reverse search",      /* "^Ob or F6", */           REVSEARCH},
{" ... within Line",    /* "^OB or   ", */           LREVSEARCH},
{"Query-replace",       /* "^QQ or Alt-S", */        QUERYREPLACE},
{"Fold case on-off",    /* "^OF", */                 FOLDSRCH},
{0,0}
};

static struct menu Tools[]={
{"Execute program",    /* "shift-F2", */      RUNMYPROGRAM},
{"Execute script",     /* "ctrl-OQ",  */      RUNBACKGROUND},
{"Command shell",      /* "F2", */            RUNPROGRAM},
#if !MSDOS
{"Inject command pipe",/*  "alt-p", */        READPIPE},
{"Suspend job",        /*  "^U", */           BAILOUT},
#endif
{0,0}
};

static struct menu Keys[]={
{"Memorize keys",                 /* "F8", */           RECORDING},
{"Replay keys",                   /* "F7", */           RUNMEMORIZE},
{"Repeat replay #",               /* "Shift-F7", */     MANYRUNMEM},
{"Save current macro",            /* "ALT-0", */        MACROKEY},
#if PIINILOAD
{"Keymap editor",                 /* "^O,h", */         COMPLETEKEY},
{"Macro editor",                  /* "ctrl-O,e", */     MACROLIBRARY},
#endif
{0,0}
};

#if 0  /* Removed March 2000 because of single interface COMPLETEKEY */
   /* Removed March 2000 because of single interface COMPLETKEY */
  {"Complete key list",             /* "^O,h", */         COMPLETEKEY},
  {"Define key",                    /* "^O,=", */         DEFINEKEY},
  {"Undefine key",                  /* "^O,=", */         KILLKEY},
#endif

static struct menu Values[]={
{"Word wrap",       /* "^OW", */          WORDWRAP},
{"Auto-indent",     /* "^OI", */          SETAUTO},
{"Hanging indent",  /* "^OU", */          SETHANGING},
{"Detab mode",      /* "^OD", */          SETDETAB},
{"Trailing space",  /* "   ", */          SETBLANKSTRIP},
{"Insert mode",     /* "Ins", */          INSMODE},
{"Highbit strip",   /* "   ", */          SETSTRIP},
{"Tab and margin",  /* "^OT", */          SETTABS},
{"Backup modes",    /* "^OB", */          BACKTOGGLE},
{"Status line",     /* "^OS", */          SETSTATUS},
{"Export-import",   /* "^KT", */          SETTEXTTYPE},
{"Printer",         /* "^OP", */          SETPRINTER},
#if SYS5 || BSDunix
{"Mouse enable",    /* none  */           SETMOUSE},
{"Mouse disable",   /* none  */           UNSETMOUSE},
#endif
{0,0}
};

static struct menu Other[]={
{"Ascii reveal codes",   /* none    */        ASCIITABLE},
{"Window resize",        /* "^O,-", */        SHRINKWINDOW},
{"Menu box setup",                            MENUBOXSET},
{"Color & menu setup",   /* none  */          COLORSET},
{"Box and line draw",    /* "^OL", */         LINEDRAWKEY},
#if MSDOS
{"Video mode",           /* none  */          VIDEOMODE},
{"Set cursorstyle",      /* none  */          CURSORSTYLESET},
{"Disk swap mode",       /* "^O,!",*/         SWAPSW},
#endif
#if PIINILOAD
{"Write pi-macro.key",   /* "ALT-9", */       DISPLAYMACRO},
{"Write pi-dump.key",    /* "^O,+", */        DUMPKEYS},
{"Read a keymap file",   /* "^O,@", */        KEYSLOADUP},
#if BSDunix || SYS5
{"Set Emacs keys",       /* "^O,@", */        SETEMACSMODE},
{"Set Wordstar keys",    /* "^O,@", */        SETWDSTARMODE},
{"Set WordPerfect keys", /* "^O,@", */        SETWDPERFMODE},
#endif
#endif
{0,0}
};

static struct menu Help[]={
{"Internal info",      /*  "^O^O", */         OPTIONSET},
{"Brief manual",       /* "^Oh", */           RUNBHELP},
{"HTML manual",           /*  */              HELPHTML},
{"Set browser",           /*  */              HELPBROWSER},
{"Text manual",           /* "^O^H", */       RUNHELP},
{"Quit editor",           /*  "Alt-X", */     TSRKEY},
{0,0}
};

enum {
HFile=   1+       sizeof("  ")-1,
HKeys=   HFile+   sizeof("File  ")-1,
HTools=  HKeys+   sizeof("Keys  ")-1,
HSearch= HTools+  sizeof("Tools  ")-1,
HEdit=   HSearch+ sizeof("Search  ")-1,
HWord=   HEdit+   sizeof("Edit  ")-1,
HBlock=  HWord+   sizeof("Word  ")-1,
HValues= HBlock+  sizeof("Block  ")-1,
HOther=  HValues+ sizeof("Values  ")-1,
HHelp=   HOther+  sizeof("Other  ")-1,
Htotal=  HHelp+   sizeof("Exit  ")-1
};

/*
   File  Edit  Word  Block  Search  Tools  Keys  Values  Other  Quit  |
*/

#define MaininfoTITLE " Pi Editor "
#define SELECTROW 1
#define MAINSTARTROW 2
#define MaininfoCOLS Htotal
#define MaininfoROWS 1
#define DROP (1+2+MaininfoROWS+1) /* how much to drop for secondary menu */

struct mainmenu {
char *name; unsigned offset; struct menu *pm;
char *title; unsigned items; unsigned hitrow;};

struct mainmenu Maininfo[]={
{"File",   HFile,   (struct menu *)&File[0],   "File Menu",        0, 1},
{"Keys",   HKeys,   (struct menu *)&Keys[0],   "Keyboard",         0, 1},
{"Tools",  HTools,  (struct menu *)&Tools[0],  "Custom Tools",     0, 1},
{"Search", HSearch, (struct menu *)&Search[0], "Search",           0, 1},
{"Edit",   HEdit,   (struct menu *)&Edit[0],   "Edit Line",        0, 1},
{"Word",   HWord,   (struct menu *)&Word[0],   "Word Operations",  0, 1},
{"Block",  HBlock,  (struct menu *)&Block[0],  "Block Operations", 0, 1},
{"Values", HValues, (struct menu *)&Values[0], "Set Values",       0, 1},
{"Other",  HOther,  (struct menu *)&Other[0],  "Other",            0, 1},
{"Help",   HHelp,   (struct menu *)&Help[0],   "Help",             0, 1}
};
#define MaininfoITEMS sizeof(Maininfo)/sizeof(Maininfo[0])


struct textinfo {
        int topcol;
        int toprow;
        int botcol;
        int botrow;
};

struct windows {
          char *title_center;                /* centered title */
          int top_row,top_col,               /* window coordinates*/
              bot_row,bot_col;
          struct textinfo oldw;              /* old window coordinates */
          char *info_center;                 /* centered information */
};

struct windows mainwind =
{MaininfoTITLE,
MAINSTARTROW,1,(MAINSTARTROW+1+MaininfoROWS),(1+MaininfoCOLS),
0,0,0,0,
""};

struct windows tmpwin =
{
"",               /* title */
DROP,             /* top_row */
0,                /* top_col, depends on screen width, start col */
(DROP+1+0),       /* bot_row, depends on # of items */
(0+0+1),          /* bot_col, depends on width of items */
0,0,0,0,          /* prev window coordinates */
""};              /* information line info_center */

/* Custom tools is a special case, where we add entries on the fly */
struct menu *UserTools= 0;  /* Pointer to Tools menu */
int UserToolsSize=0;        /* Menu entries in Tools menu */

void killUserTools(){
  if(UserTools) releaseMenu(UserTools,UserToolsSize);
  UserToolsSize=0; UserTools=0;
}

#if 0
struct windows {
          char *title_center;                /* centered title */
          int top_row,top_col,               /* window coordinates*/
              bot_row,bot_col;
          struct textinfo oldw;              /* old window coordinates */
          char *info_center;                 /* centered information */
};
#endif

void setupwindow(n) int n; {
int k;
int numlines,width;
struct menu *q;
extern void getUserTools();

  if(--n == 2){
    getUserTools(); /* Changes Maininfo */
    Maininfo[2].pm=UserTools;
  }
  /* Fix entries of tmpwin to match current data */
  width=strlen(tmpwin.title_center=Maininfo[n].title);
  numlines=0; q=Maininfo[n].pm;
  while(q->name) {++numlines; width=max(width,strlen(q->name)); ++q;}
  Maininfo[n].items=numlines;
  tmpwin.top_col=Maininfo[n].offset;          /* Might not work */
  tmpwin.bot_col=tmpwin.top_col+1 + 1+width;
  if(tmpwin.bot_col > Htotal){  /* window is too wide */
    tmpwin.top_col=Htotal+1-width-2;
    tmpwin.bot_col=Htotal+1;
  }
  /* The largest allowed ROWS size is ROWMAX-DROP because of screen size */
  tmpwin.top_row=DROP;
  k=min(numlines,ROWMAX-DROP);     /* Must fit on the screen */
  tmpwin.bot_row=DROP+1+k;
  tmpwin.info_center= (k == numlines) ?  "" : " + ";
}

static
int last=0;

unsigned menu()
{
unsigned retval;
int i,j,n,k;

  sysarg=sysv[0]=0;
  menumouse(1);
  retval=NULLKEY;
  setcursortype(0);
  setwindow(1,1,COLMAX+1,ROWMAX+1); /* using whole screen */
  if(winopen(&mainwind)){
     opnmain(); n=last+1;
     goto top1;
top:
     while(1){
       hilite(last,1);
       i=getkey();
       retval=NULLKEY;
       k=last;
       n=translate(i,&last,MaininfoITEMS);
       hilite(k,0);
       if(n==0) continue;
       if(n == -2) goto quit;
       if(n != -1) break;
     }
top1:
  while(1){
    if(n<1) n=MaininfoITEMS;
    if(n>MaininfoITEMS) n=1;
    last=n-1;
    hilite(last,1);
    setupwindow(n);
    if(winopen(&tmpwin)){
       j=opnselection(n,&retval);
       winclose(&tmpwin);
       killUserTools();
    } else {j = -1; bell();}
    if(j == 2) goto quit;
    hilite(last,0);
    if(j == -1) goto top;
    if(j == 3) --n;
    if(j == 4) ++n;
  }
quit:
  winclose(&mainwind);
  }
  setcursortype(1);
  lcount();
  menumouse(0);
  return retval;        /* NULLKEY if nothing happened */
}

void opnmain()
{
int i; extern char *endof();
  setattribute(windattr);
  strcpy(sysx,"  ");
  for(i=0;i<MaininfoITEMS;++i){
    strcat(sysx,Maininfo[i].name);
    strcat(sysx,"  ");
  }
    mycputs(1,SELECTROW,sysx);
}

/* Requires setupwindow to initialize variables in tmpwin */
int opnselection(n,retval) int n; int *retval; {
int j,k,hit,width,NUML;
/* struct menu { char *name; unsigned keycode; }; */
struct menu *p;
char *endof();

  p=Maininfo[--n].pm;
  k=Maininfo[n].items;     /* number of items */
  width=tmpwin.bot_col - tmpwin.top_col-2; /* number of columns*/
  NUML=tmpwin.bot_row - tmpwin.top_row - 1;     /* number of window rows */
  NUML=min(NUML,ROWMAX-DROP);     /* Must fit on the screen */
    hit =Maininfo[n].hitrow-1;
    j=scrollingwindow((char *)0,p,k,width,NUML,&hit,sysvv,1,0);
    Maininfo[n].hitrow=hit+1;
    if(j==2){ p += hit; retval[0]=p->keycode;}
    return j;
}

void hilite(n,flag) int n,flag; {
  setattribute(windattr);
  if(flag) setattribute(REVERSEVIDEO);
  mycputs(Maininfo[n].offset,SELECTROW,Maininfo[n].name);
}

unsigned tounsignedupper(i) unsigned i; {
  if('a' <= i && i <= 'z') i -= ' ';
  return i;
}

int translate(i,last,maxnum)
unsigned i; int *last,maxnum; {
int n;
  i=tounsignedupper(i);
  switch(i){
  case ' ':case BAILOUT:case ENTERKEY:case MENUKEY: n= -2;break;
  case LEFTARROW:case 8:case 127:
    --last[0]; if(last[0]<0) last[0]=maxnum-1; n=0; break;
  case RIGHTARROW:case TABKEY:
    ++last[0]; if(last[0]>=maxnum) last[0]=0; n=0; break;
  case RETURNKEY: case DOWNARROW: case UPARROW:
    n=last[0]+1; break;
#if BSDunix || SYS5
  case MOUSEPRESS: getbyte(); getbyte(); getbyte(); break;
#endif
  default:
    n=menusearch(i,last[0],maxnum); /* returns -1 on failure */
    break;
  } /* end switch */
  return n;
}

int menusearch(i,last,maxnum)
int i,last,maxnum;
{
int j;
int flg=0;
char *strchr();
twice:
  for(j=last;j<maxnum;++j){
    if(strchr(Maininfo[j].name,i) != (char *)0) return j+1;
  }
  if(!flg){flg=1; maxnum=last; last=0; goto twice;}
  return -1;
}

/* Show a key sequence definition */
/* buf[] contains keys already hit */
void DisplayKeySeq(prompt,buf) char *prompt,*buf; {
char d[256];
char *p;
int x;
extern char *endof();
  p=endof(buf);
  while(1){
    printseq(prompt,buf,d);
    banner(d);
    if((x = getbyte()) == '\r') break;
    if((x==127 || x==8) && p>buf) --p;
        else
    *p++ =x;
    p[0]=0;
  }
}

int statflag=0;
int statwidth=0;
int verbMode=0;
#define STATBYTECHR 'D'         /* for delete a file */

int scrollingwindow(s,qq,len,width,NUML,index,sysvv,keys,statmark)
char **s; struct menu *qq;
int len,width,NUML; int *index; char *sysvv; int keys; int statmark;
{
int row,page,oldpage,last,n,i,j,k,flg,tmp,keycode;
char *strchr(), *statbyte(); char *p,*r; struct menu *q; char *pp;
char *endof();
char bline[4];
extern int savcount;
extern char *getkeybuf;

      bline[0]=bline[2]=' '; bline[3]=bline[1]='\0';
      statflag=statmark;
      statwidth=width;
      last = index[0];
top:
      setcursortype(0);
      n=0;
      row=(last%NUML)+1;
      page=last/NUML; oldpage=page-1;
      while(n==0){
        if(page != oldpage){
          setattribute(windattr);
          for(oldpage=page, j=NUML*page,i=1;i<=NUML;++i,++j) {
            if(keys){q=qq+j; p = q->name; ++q;} else p = s[j];
            if(j>=len) p=bline;
            scrollcputs(i,p);
          }
        }
        setattribute(0);

        if(keys){  /* What to do with keymaps */
          q=qq+last; p = q->name; keycode=q->keycode;
          identifykey(keycode,sysvv);
          killduplicates(sysvv);
          if(strcmp(sysvv,"Keymap")==0) strcpy(sysvv,"No key mapping");
          if(keys==2){
            /* p is 2 characters longer than its string length */
            sprintf(sysv,"%s %s",(verbMode ? endof(p)+1:""),sysvv);
            strcpy(sysvv,sysv);
            if(keycode == 0){ /* No keymap on a header */
              strcpy(sysvv,(p[0]=='=')?"SPACEBAR quits":"");
            }
          }
          banner(sysvv);
        }
        else p = s[last];

        setattribute(REVERSEVIDEO);
        mycputs(1,row,p); /* mark target */
        setattribute(windattr);

        i=getkey();    /* Get one or more characters, return keycode */
        if(keys==2){
          /* getkey() stops on first mismatch of a key sequence */
          /* What to do with extra chars emitted by a key? */
          if(i< ' ') i=NULLKEY;
          if(i==NULLKEY) {
            bell();
            strcpy(sysvv,getkeybuf);  /* Copy what was read by getkey */
            DisplayKeySeq(" Press RETURN. Unmapped: ",sysvv);
            goto top;
          }
        }
#if BSDunix || SYS5
        if(i==MOUSEPRESS){
          int Row;
          getbyte(); getbyte(); Row=getbyte()-' '-DROP;
          if(!keys) goto top;  /* Eat mouse press */
          if(1<= Row && Row <= NUML){
             mycputs(1,row,p); /* unmark target */
             while(row>Row){ --row; --last;}
             while(row<Row){ ++row; ++last;}
          }
          goto loop;
        }
#endif

/* Working here ... */
        if(keys==2){
          setcursortype(1);
          switch(i){
/* debugging here .. */
#if 0
          case 'm':
             banner("");
             displaymemoryleft();           /* Debugging memory */
             goto loop;
#endif
          case '+': if(addkey(keycode) == -1) goto loop;
                    goto loopminus2;

          case '-': if(keycode == 0) goto loop;
                    { extern int lastidentifykey;
                      if(lastidentifykey!=-1){
                        if(killkey(ourkeys[lastidentifykey].seq) == -1)
                          goto loop;
                      }
                    }
                    goto loopminus2;

          case 'V': verbMode=1-verbMode; goto top;
          case 'S': /* S Show keypress */
                    sysvv[0]=0;
                    DisplayKeySeq(" RETURN quits: ",sysvv);
                    goto top;
          case 'M': if(!newmacrotitle(keycode)) goto top;
                    goto loopminus2;     /* "M  Macro title" */
          case 'T': if(!newtooltitle(keycode)) goto top;
loopminus2:
                    n= -2; goto loop;    /* "T  Tools menu item" */
          case '.': i=RETURNKEY;
                    /* strcpy(sysvv,p); */ /* Execute command */
                    /* n= 2; goto loop; */
          case ' ': /* Quit */
          case DOWNARROW:            /* Navigate */
          case UPARROW: goto navigate;
          default: /* match the key seq and refresh */
               for(j=0;j<len;++j){
                 if((qq+j)->keycode == i) {last=j; goto top;}
               }
          } /* End switch() */
          goto skipkeys;
        } /* end if(keys==2) */

navigate:
        switch(i){
          case INSERTLINE: tmp=' '; goto markit;
          case DELLINE:   tmp=STATBYTECHR;
markit:
                          if(statmark) *statbyte(p)=tmp;
dnarrow:
                          bline[0] = *statbyte(p);
                          mycputs(statwidth+1,row,bline);
                          bline[0]=' ';
          case DOWNARROW: if(last < len-1){
                            if(row<NUML) {scrollcputs(row,p); ++row;}
                            else {row=1; ++page;}
                            ++last;
                          }
                          break;   /* Has to goto loop */
          case UPARROW:   if(last>0) {
                            if(row>1){ scrollcputs(row,p); --row; }
                            else {row=NUML; --page;}
                            --last;
                          }
                          break;
          case PAGEFORWARD:
                          if((j=page*NUML+NUML)<len){row=1; ++page;last=j;}
                          break;
          case PAGEBACK:
                          if(page>0){
                            --page; row=NUML;
                            last = page*NUML+NUML-1;
                          }
                          break;
          case EXTLEFTARROW:
                              scrollcputs(row,p);
                              last=page*NUML;
                              row=1; break;
          case EXTRIGHTARROW:
                              scrollcputs(row,p);
                              last=page*NUML+NUML-1; row=NUML;
                              if(last>= len-1) {last=len-1; row=1+(last%NUML);}
                              break;
          case RETURNKEY: strcpy(sysvv,p);
                          n= 2; /* n= 2 is highlighted text */
                          break;
          case LEFTARROW: if(keys) n=3; break;
          case RIGHTARROW: if(keys) n=4; break;
          case ' ':
          case BAILOUT:
          case ENTERKEY:
          case MENUKEY:   n= -1; break;
          default: goto skipkeys;
        };
        goto loop;
skipkeys:

/* Search for the next text string that contains character i */

#if MSDOS      /* map to upper case */
        i=tounsignedupper(i);
#else
        if(keys) i=tounsignedupper(i);
#endif
        j=(last>=len-1) ? 0 : last+1;
        flg=0; k=len;
twice:
        for(;j<k;++j){
          if(keys){
            q=qq+j;
            if(strchr(q->name,i) == (char *)0) continue;
          } else {
            if(s[j][0] != i) continue;
          }
          flg=1; break;
        }
        if(!flg){flg=2; k=last+1; j=0; goto twice;}
        if(flg==1){
          /* new page? */ page = j/NUML;
          if(page==oldpage) scrollcputs(row,p);
          last=j; row = 1+(j%NUML);
        }  else {
          goto dnarrow;  /* Which eventually does goto loop */
        }
loop: ;
      }/* end while(n==0) */

     index[0]=last;     /* return index of selection */
quit1:
     setattribute(0);
     return n;
     /* n==  3: LEFT key */
     /* n==  4: RIGHT key */
     /* n== -2: [keys=2] special exit */
     /* n== -1: abort */
     /* n==  2: sysvv[] is highlighted text, index[0]=array index */
}

void scrollcputs(row,t) int row; char *t; {
int k; extern int statflag; char *statbyte();
  strcpy(sysx,t);
  k=strlen(t); while(k<statwidth) sysx[k++]=' ';
  sysx[k++]= ((statflag!=0) ? *statbyte(t) : ' ');
  sysx[k]=0;
  mycputs(1,row,sysx);
}



/* get directory list */
/* ===================================================================*/
/* need scrolling window, dynamically determined, with functions to
   scroll the window and select an item. If .. is selected, then open
   the mother directory. Ditto for subdirectories. Once a file is
   selected, then copy it to sysvv[] and return with load file code.
*/
void shell(s,n) char **s; int n; {
int gap,i,ig,j; char *tmp;
  for (gap = n/2 ; gap > 0 ; gap /= 2){
    for (j = gap ; j < n; j++){
      for (i = j-gap ; i >= 0 ; i -= gap) {
              ig = i + gap;
              if (strcmp(s[i],s[ig]) <= 0)
                      break;
              tmp=s[i]; s[i]=s[ig]; s[ig]=tmp;
      }
    }
  }
}


void removefile(ddir,fname,query) char *ddir,*fname; int *query; {
 if(query[0] != 'A' && query[0] != 'Q'){
   bbanner(fname,": Delete file [y/n/a/q]? ");
   query[0]=getupper();
 }
 if(query[0]=='Y' || query[0] == 'A'){
    strcpy(sysx,ddir); addslash(sysx); strcat(sysx,fname);
    unlink(sysx);
 }
}

#define MAXNUML    16   /* maximum number of lines in popup */
#define DirStartRow 2   /* line where window starts */
struct windows  DirList = {
"",
DirStartRow,HFile,DirStartRow+1+MAXNUML,HFile+MAXNUML,
0,0,0,0,""};

int FindFile(buf,sysvv,which) char *buf,*sysvv; int which; {
extern char *strchr(),*strrchr(), *statbyte();
char ddir[MAXFNAME];
char **s;
int v;
int i,j,n;
int query;
int width;              /* width of popup */
int NUML;               /* number of lines in popup */
int last=0;      /* offset into text array */
LLONG atoiLL();
top:
    if((s=(char **)malloc(MAXDIRENT*sizeof(char *)))==(char **)0) goto quitzero;
    if(!which){
      if((v=dir(buf,1,1,ddir,s))==0) {bell(); goto quitzero;}
      shell(s,v); last=0;
    } else {
      ddir[0]=0;
      v=getbufferlist(s,&last);
    }
    for(width=i=0;i<v;++i) width=max(width,strlen(s[i]));
    /* width = max strlen of file spec */
    width = max(2+strlen(ddir),width);
    NUML=min(v,ROWMAX-2-4); /* popup uses 2+NUML lines */
    DirList.title_center=ddir;
    DirList.info_center=(NUML==v) ? "" : " + ";
    DirList.bot_row = DirStartRow+1+NUML;
    DirList.bot_col = HFile+width+2;
    if(winopen(&DirList)){
      n=scrollingwindow(s,(struct menu *)0,v,width,NUML,&last,sysvv,0,1);
      winclose(&DirList);
    } else bell();
    for(query=j=0;j<v;++j){
      if(which==0 && *statbyte(s[j])==STATBYTECHR)
        removefile(ddir,s[j],&query);
    }

    for(j=0;j<v;++j) freeup(s[j]); freeup(s);
    setcursortype(1);

    if(n== -1) goto quitzero;
    if(which==1){
      sysarg=(int)atoiLL(sysvv); goto quit;
    }
    if(n==2) n=buildspec(ddir,sysvv);
    /* buildspec: n=2 is a dir, n=3 is a filespec */
    if(n==2) {
      dirsplit(sysvv,buf);
      goto top;
    }
quit:
    lcount();
    return 1;
quitzero:
    lcount();
    return 0;
}


#if BSDunix || SYS5 || MSDOS

#if MSDOS
#define BACKSL '\\'
#endif

#if SYS5 || BSDunix
#define BACKSL '/'
#endif

int buildspec(dir,s)
char *dir;      /* mother directory */
char *s;        /* target dir/file name, to be filled in */
{
char t[MAXFNAME];
int j; char *p; extern char *strrchr();
  /* is t=dir+s a directory or a file? */
  strcpy(t,dir);
  strcat(t,s);
  if(isadir(t)){        /* march up dir tree? Or down tree? */
    if(s[0]=='.' && s[1]=='.') {/* up tree, remove one dir level */
      /*  remove last BACKSL */
      strcpy(t,dir);
      if((j=strlen(t))>0 && t[j-1] == BACKSL) t[j-1]=0;
      if((p=strrchr(t,BACKSL))!=(char *)0){   /* Is there another? */
        p[1]=0;                               /* Then leave it, trunc. */
      } else strcpy(t,dir);                   /* Otherwise no change */
    }
    else {      /* down tree, add new level (all done in t[]) */
      /*  add last BACKSL */
#if BSDunix || SYS5
      addslash(t);
      addslash(dir);
#endif
    }
    j=2;        /* it's a dir */
  } else j=3;   /* it's a file */
  strcpy(s,t);
  return j;
}

#endif

/* End get directory list */

int reachedMAXDIRENT(v) int v; {
  return ((MAXDIRENT-4<v) ? 1 : 0);
}

#define HelpStartRow 2   /* line where window starts */

#if 0
struct windows {
          char *title_center;                /* centered title */
          int top_row,top_col,               /* window coordinates*/
              bot_row,bot_col;
          struct textinfo oldw;              /* old window coordinates */
          char *info_center;                 /* centered information */
};
#endif

struct windows  HelpList = {
"",
HelpStartRow,HFile,HelpStartRow+1+16,HFile+22,
0,0,0,0,
"Press SPACE to quit"};

/*#define MACROTRUNC 42*/       /* How wide before macro text is truncated */

int HelpKeys(KEYcode) int *KEYcode; { /* Sets keys=2 in scrollingwindow */
static int keeplast[2]={0,0};
extern char *sob(),*fnb(),*strchr(),*strrchr(),*endof();
extern char *KEYMAP;
FILEP fp;
char buff[512];
char title[128];
struct windows *wp;
struct menu *q,*qq;    /* struct menu { char *name; unsigned key; }; */
char *p,*r;
int j,k,keycode;
int n;                  /* error return code */
int v;                  /* Number of strings in menu */
int last=0;             /* offset into text array */
int width;              /* width of popup */
int NUML;               /* number of lines in popup */
int badcore;
extern char *strstr();
LLONG atoiLL();
top:
    if((qq=q=(struct menu *)malloc(MAXDIRENT*sizeof(struct menu)))==(struct menu *)0){
      goto errexit;   /* "Can't malloc, low memory" */
    }
    badcore=v=0;

    switch(KEYcode[0]){
    case MACROLIBRARY: HelpList.title_center="Macro Library Editor";
                       last=keeplast[1]; goto macrokeys;
    default:           HelpList.title_center="Key Map Editor";
                       last=keeplast[0];
    }

    /* Process system-wide editor function definition file KEYMAP */

    if((fp=fopenbread(KEYMAP))==(FILEP)0){
      ssprintf(buff,"%s not found",KEYMAP);
      bellpressany(buff);
      goto macrokeys;
    }
    /* using heavily the format of keymap.dat */
    while(fgetln(buff,fp)>=0){
      if(!isdigit(buff[0])) continue;  /* Skip all except digits */
      if( ( (keycode= (int)atoiLL(buff)) >= MACROKEY ) || keycode < 0) continue;
      if(!keycode) strcat(buff," []");
      else {
        if((p=strstr(buff," ["))==0 || !strchr(p,']')) strcat(buff," []");
        sprintf(endof(buff),"{%u}",(unsigned)(keycode));
      }
      /* The string now ends with ']' or '}' and has length >=2 chars */
      if(!safecore(&(q->name),sob(fnb(buff)))){
         badcore=1; break;
      }
      q->keycode=keycode;
      strTruncate(q->name,'[');         /* Truncate just before bracket */
      ++q; if(reachedMAXDIRENT(++v)){badcore=1;break;};
    }
    closedfile(fp);
    if(badcore) goto badCORE;


    /* Get the macros from the memory-resident macrobuf[] array */
macrokeys:
    for(j=1;j<=MACNUM;++j){  /* Search for active macros */
      if(macromax[j]==0) continue;
      keycode=j+MACROKEY;
      getMacroTitle(keycode,title);   /* Get the title */
      ssprintf(buff,"Macro%d: %s [%u]",j,title,keycode);
      if(!safecore(&(q->name),buff)) goto badCORE;
      strTruncate(q->name,'[');         /* Truncate just before bracket */
      q->keycode=keycode;
      ++q; ++v;

      getToolTitle(keycode,title);   /* Get the title */
      if(title[0]){
        ssprintf(buff,"Tools: %s [%u]",title,keycode);
        if(!safecore(&(q->name),buff)) goto badCORE;
        strTruncate(q->name,'[');         /* Truncate just before bracket */
        q->keycode=keycode;
        ++q; ++v;
      }

      strcpy(buff," macro=");
      printseq("",macrobuf[j],endof(buff)); /* keys might not be defined */
      buff[MACROTRUNC]=0; /* truncate it; only 42 chars show on screen */
      ssprintf(endof(buff)," [%u]",keycode);
      /* The string now ends with "]" and has length >=2 chars */
      if(!safecore(&(q->name),buff)) goto badCORE;
      strTruncate(q->name,'[');         /* Truncate just before bracket */
      q->keycode=keycode;
      ++q; if(reachedMAXDIRENT(++v)) goto badCORE;
    }
    last=min(last,v-1);           /* Upon reread, last could decrease */

    /* Display an information window */
    /* Don't worry about overlaps yet */
    {char *msg[]={
      "+  add a key sequence",
      "-  delete a sequence",
      ".  execute keycode",        /* Exit n= 2 */
      "M  Macro title",            /*  newmacrotitle(); */
      "T  Tools title",            /*  newtooltitle(); */
      "V  Verbose mode",
      "S  Show keypress",
      0};
      WarningMessage(msg); /* memory released; erase screen later */
    }



    /* Build the window data in which to display structure qq */
    wp= &HelpList;
    for(width=j=0;j<v;++j) width=max(strlen((qq+j)->name),width);
    width = max(1+sizeof(wp->title_center),width);
    width = max(1+sizeof(wp->info_center),width);
    NUML=min(v,ROWMAX-2-4); /* popup uses 2+NUML lines */
    wp->bot_row = HelpStartRow+1+NUML;
    wp->bot_col = HFile+width+2;
    n  = -1;
    if(winopen(wp)){
      n=scrollingwindow((char **)0,qq,v,width,NUML,&last,buff,2,0);
      winclose(wp);
    } else bell();
    if(KEYcode[0]==MACROLIBRARY) keeplast[1]=last;
    else keeplast[0]=last;
    if(n==2) KEYcode[0]=(qq+last)->keycode; /* Quit then execute keycode */
    closehelpsheet();                       /* Clean up the screen */
    releaseMenu(qq,v);                      /* free up memory */
    setcursortype(1);                       /* back to normal */
    lcount();                               /* Fix the status line */

    if(n== -2) goto top;
    return 0;
badCORE:
    releaseMenu(qq,v);
errexit:
    bellpressany("Can't malloc, low memory");
    return 1;
}


void releaseMenu(qq,v) struct menu *qq; int v; {
struct menu *q;
int j;
    for(q=qq,j=0;j<v;++j) {freeup(q->name); ++q; }
    freeup(qq);
}

/* Shorten string 's' one behind last char 'c' */
void strTruncate(s,c) char *s; int c; {
char *p;
extern char *strrchr();
  if((p=strrchr(s,c))!=(char *)0){
         if(p != s) p[-1]=0;
  }
}

void getUserTools() /* After call, releaseMenu(UserTools,UserToolsSize) */
{
char buff[512];
FILEP fi=0;
int j,v;
int keycode;
char *p;
struct menu *qq,*q;
extern char *strstr();
extern char *USERFILE;

  v=0;
  if((qq=q=malloc(MAXDIRENT*sizeof(struct menu)))==(struct menu *)0){
    bellpressany("Can't malloc, low memory");
    goto checkout;
  }
  for(j=0;(Tools[j].name != 0);++j){
    if(!safecore(&(q->name),Tools[j].name)) goto errexit;
    q->keycode=Tools[j].keycode; ++v; ++q;
    /* debug */
    /* printf("keycode=%d, title=%s\n",Tools[j].keycode,Tools[j].name); */
  }
  if((fi=fopenbread(USERFILE)) == (FILEP)0) goto checkout;
  while(fgetln(buff,fi)>0){
     if(isacommentline(buff)) continue;
     if(isaheaderline(buff,&j)) keycode=j;
     if((p=strstr(buff,"tool="))!=0){
       /* Save title and keycode as a Tools Menu item */
       if(!safecore(&(q->name),p+5)) goto errexit;
       q->keycode=keycode; ++q; if(++v >= MAXDIRENT-2) goto errexit;
       /* debug */ /* printf("keycode=%d, title=%s\n",keycode,p+5); */
     }
  }
  q->name=0; ++v;     /* Null marker for end of data */
checkout:
  if(fi) closedfile(fi);
  UserTools=qq;
  UserToolsSize=v;
  return;
errexit:
  bellpressany("Can't malloc, low memory");
  releaseMenu(qq,v);
  qq=0; v=0;
  goto checkout;
}


#endif  /* #if POPUPS */

#if POPUPS
/* window module for MSDOS/TURBOC and mainframes vt100 */
/*======================================================================*/

static
int topcol,toprow,botcol,botrow;

void setwindow(topc,topr,botc,botr) int topc,topr,botc,botr; {
  /* save window coordinates in global data set */
  topcol=topc; toprow=topr; botcol=botc; botrow=botr;
}

void mygotoxy(c,r) int c,r; {
  /* we use top row for status, row off by 1 */
  /* c,r numbered from 1, toprow,topcol numbered from 1 */
  /* putcursor row,col numbered from 0 */
  putcursor(r+(toprow-2)-1,c+(topcol-2));
}

void textreplace(topc,topr,botc,botr) int topc,topr,botc,botr; {
int c,r,i,j,width,height;  char *p; extern char *toprint();
  c = topc-1;  /* change to range 0..COLMAX */
  r = topr-1-1;  /* and range 0..ROWMAX */
  width = botc-topc+1+1;
  height = botr-topr+1;
  for(i=0;i<height;++i,++r){
    p=toprint(line[r]);
    if(strlen(p)<c) sysx[0]=0; else strcpy(sysx,p+c);
    j=strlen(sysx);
    while(j<width) sysx[j++]= ' ';
    sysx[width]=0;
    xPrint(sysx,r,c);
  }
  sysx[0]=0;
}

int winopen(w) struct windows *w;
{
int width,lastrow;

   if(w->bot_row > ROWMAX+1 || w->bot_col > COLMAX+1) return 0;
   if(w->bot_row - w->top_row - 1 <=0) return 0;
   w->oldw.topcol = topcol; w->oldw.toprow=toprow;
   w->oldw.botcol=botcol; w->oldw.botrow=botrow;
   setwindow(w->top_col,w->top_row,w->bot_col,w->bot_row+1);
   draw_box(w);
   setattribute(windattr);
   setattribute(titlattr);
   width  = w->bot_col - w->top_col +1;
   width -= 1+strlen(w->title_center);
   if(2<width) mycputs((width/2)+2,1,w->title_center);
   width  = w->bot_col - w->top_col +1;
   width -= 1+strlen(w->info_center);
   lastrow=w->bot_row-w->top_row+1;
   if(2<width) mycputs((width/2)+2,lastrow,w->info_center);
   setattribute(0);
   /* shrink window size 1 char on each edge */
   setwindow(w->top_col+1,w->top_row+1,w->bot_col-1,w->bot_row-1);
   return 1;
}

void winclose(w) struct windows *w;
{
  wintextreplace(w);
  winrestorewindow(w);
}

void wintextreplace(w) struct windows *w;
{
#if BSDunix || SYS5
  setattribute(0x00);
  textreplace(w->top_col,w->top_row,w->bot_col,w->bot_row);
#else
  textreplace(w->top_col,w->top_row,w->bot_col,w->bot_row);
  setattribute(windattr);
#endif
}

void winrestorewindow(w) struct windows *w;
{
  setwindow(w->oldw.topcol,w->oldw.toprow,w->oldw.botcol,w->oldw.botrow);
}

void draw_box(w) struct windows *w;
{
int x;
int height, width;
int horiz,vert;
int ulh,urh,llh,lrh;
char bar[2];

  if(macrun || nodisp) return;
  width     = w->bot_col - w->top_col +1;
  height    = w->bot_row - w->top_row +1;

  switch(boxcharset){
  case 1: horiz=196; vert=179;       /* single box */
          ulh=218; urh=191; llh=192; lrh=217;
          break;
  case 2: horiz=205; vert=186;       /* double box */
          ulh=201; urh=187; llh=200; lrh=188;
          break;
  case 3: horiz=232; vert=224;       /* single box iso font */
          ulh=229; urh=238; llh=226; lrh=235;
          break;
  case 4: vert=242; horiz=220;      /* double box iso font */
          llh=244; ulh=251; lrh=214; urh=163;
          break;
  default:
  case 0: horiz= '=';                /* text box */
          ulh=urh=llh=lrh=vert= ' ';
          break;
  }

  setattribute(windattr);
  setattribute(bordattr);
  /* Write the corners */
  mygotoxy(1,1);          aputc(ulh,1);
  mygotoxy(width,1);      aputc(urh,1);
  mygotoxy(1,height);     aputc(llh,1);
  mygotoxy(width,height); aputc(lrh,1);
  /* Verticals */
  for(x=2;x<height;++x) {
    mygotoxy(width,x); aputc(vert,1);
    mygotoxy(1,x);     aputc(vert,1);
  }
  /* horizontals */
  mygotoxy(2,1); aputc(horiz,width-2);
  mygotoxy(2,height); aputc(horiz,width-2);
  setattribute(0);
}

int sysattribute;

void setattribute(x) int x; {
if(x) sysattribute = x;
#if !MSDOS
switch(x){
case 0x00: etype(NORMVIDEO); setcolorsUNIX(1); break;
case 0x07: etype(NORMVIDEO); setcolorsUNIX(3); break;
case 0x70: etype(REVVIDEO); break;
case 0x80: etype(BOLDF); break;
}
#endif
}

/* output char=c with attribute=sysattribute for n times */
#if MSDOS
void aputc(c,n) int c,n; {
void __int__();
extern int sysattribute;
  _CX=n;
  _BH=0; _BL=sysattribute;
  _AH=0x9; _AL=c;
  __int__(0x10);
}
#endif

#if !MSDOS
void aputc(c,n) int c,n; {
char s[256]; int i=0;
  while(i<n) s[i++]=c; s[i]=0;
  ePrint(s);
}
#endif

#if MSDOS
void mycputs(col,row,s) int col,row; char *s; {
  if(macrun || nodisp) return;
  while(*s){
   mygotoxy(col++,row);
   aputc(*s++,1);
  }
}
#endif

#if !MSDOS
void mycputs(col,row,s) int col,row; char *s; {
  if(macrun || nodisp) return;
  mygotoxy(col,row);
  etype(s);
}
#endif

#endif /* POPUPS */

/* Line and Box drawing module, all systems */
/*======================================================================*/
/* Line and Box drawing module for IBM-PC graphics */
/* my line drawing set:   3  D    E   A   B   C   4   @   Y    Z    ?  */
/*                       179 196 197 193 194 195 180 192 217  218  191 */
enum {
VERT=1,HORIZ=2,CROSS=3,TEEup=4,TEEdn=5,TEEright=6,TEEleft=7,
BOXll=8,BOXlr=9,BOXul=10,BOXur=11
};

#if MSDOS
#define X 128
unsigned
char style[7][12] ={
/* single line */
' ','3'+X,'D'+X,'E'+X,'A'+X,'B'+X,'C'+X,'4'+X,'@'+X,'Y'+X,'Z'+X,'?'+X
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* double line */
' ',':'+X,'M'+X,'N'+X,'J'+X,'K'+X,'L'+X,'9'+X,'H'+X,'<'+X,'I'+X,';'+X
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* double horizontal, single vertical */
' ','3'+X,'M'+X,'X'+X,'O'+X,'Q'+X,'F'+X,'5'+X,'T'+X,'>'+X,'U'+X,'8'+X
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* double vertical, single horizontal */
' ',':'+X,'D'+X,'W'+X,'P'+X,'R'+X,'G'+X,'6'+X,'S'+X,'='+X,'V'+X,'7'+X
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Ascii */
   ' ', '|', '-', '+', '-', '-', '|', '|', '*', '*', '*', '*'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Ascii (+|=*) */
   ' ', '|', '=', '+', '=', '=', '|', '|', '*', '*', '*', '*'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Eraser */
   ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
};
#define STYLEMESSAGE "1=Single,2=Dbl,3=Dbl top,4=Dbl side,5=Asc-,6=Asc=,7=Erase"
#endif

#if !MSDOS
unsigned
char style[5][12] ={
/* Ascii (+|-) */
   ' ', '|', '-', '+', '-', '-', '|', '|', '+', '+', '+', '+'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Ascii (+|=) */
   ' ', '|', '=', '+', '=', '=', '|', '|', '+', '+', '+', '+'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Ascii (+|-*) */
   ' ', '|', '-', '+', '-', '-', '|', '|', '*', '*', '*', '*'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Ascii (+|=*) */
   ' ', '|', '=', '+', '=', '=', '|', '|', '*', '*', '*', '*'
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
,
/* Eraser */
   ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' ', ' '
/*  0    1    2    3    4    5    6    7    8    9    10   11 */
};
#define STYLEMESSAGE "1=Ascii(+|-),2=Ascii(+|=),3=Ascii(+|-*),4=Ascii(+|=*),5=Erase"
#endif

int stylesize = sizeof(style)/sizeof(style[0]);

/* subscripts which determine character of connector piece */
static char NORTH[]={1,3,5,6,7,10,11}; /* connect north */
static char SOUTH[]={1,3,4,6,7,8,9};   /* etc */
static char WEST[]={2,3,4,5,6,8,10};
static char EAST[]={2,3,4,5,7,9,11};

/* table arranged as follows: */
/*
       +----+
       | u  |
  +----+----+----+
  | l  | *  | r  |      * = current row,col
  +----+----+----+       1 = connect possible
       | d  |            0 = connect impossible
       +----+            u = 8-bit, r = 4-bit, d = 2-bit, l = 1-bit
*/
/*  8*u+4*r+2*d+1*l,  Rt, Lft, Up, Dn */
#define STATE1111 8*1+4*1+2*1+1*1
#define STATE1110 8*1+4*1+2*1+1*0
#define STATE1100 8*1+4*1+2*0+1*0
#define STATE1101 8*1+4*1+2*0+1*1
#define STATE1011 8*1+4*0+2*1+1*1
#define STATE1010 8*1+4*0+2*1+1*0
#define STATE1000 8*1+4*0+2*0+1*0
#define STATE1001 8*1+4*0+2*0+1*1
#define STATE0111 8*0+4*1+2*1+1*1
#define STATE0110 8*0+4*1+2*1+1*0
#define STATE0100 8*0+4*1+2*0+1*0
#define STATE0101 8*0+4*1+2*0+1*1
#define STATE0011 8*0+4*0+2*1+1*1
#define STATE0010 8*0+4*0+2*1+1*0
#define STATE0000 8*0+4*0+2*0+1*0
#define STATE0001 8*0+4*0+2*0+1*1
static char
transtable[16][5] = {
STATE1111, CROSS, CROSS, CROSS, CROSS,
STATE1110, TEEright, CROSS, TEEright, TEEright,
STATE1100, BOXll, TEEup, BOXll, TEEright,
STATE1101, TEEup, TEEup, TEEup, CROSS,
STATE1011, CROSS, TEEleft, TEEleft, TEEleft,
STATE1010, TEEright, TEEleft, VERT, VERT,
STATE1000, BOXll, BOXlr, VERT, VERT,
STATE1001, TEEup, BOXlr, BOXlr, TEEleft,
STATE0111, TEEdn, TEEdn, CROSS, TEEdn,
STATE0110, BOXul, TEEdn, TEEright, BOXul,
STATE0100, HORIZ, HORIZ, BOXll, BOXul,
STATE0101, HORIZ, HORIZ, TEEup, TEEdn,
STATE0011, TEEdn, BOXur, TEEleft, BOXur,
STATE0010, BOXul, BOXur, VERT, VERT,
STATE0000, HORIZ, HORIZ, VERT, VERT,
STATE0001, HORIZ, HORIZ, BOXlr, BOXur
};

/* choose default line style */
static
unsigned
char *boxpiece;

void stylechange(n) int n;{
extern int drawstyle,stylesize;
  if(n==0) {
    drawline=1-drawline;
    if(drawline) getnumquery(STYLEMESSAGE,&drawstyle,1,stylesize);
    killstatus();
  }
  else
  if(1<=n && n<=stylesize) drawstyle=n;
  boxpiece= style[drawstyle-1];
}

int findnewchar(x,y) int x,y; {
 int i;
 for(i=0;i<16;++i){
   if(transtable[i][0] == x) return (0xFF & boxpiece[transtable[i][y]]);
 }
 return 0;
}

/* see if x is one of the matching pieces in boxpiece[] */
int boxmatch(x,which) unsigned int x; char *which; {
int i;
  for(i=0;i<7;++i) if((unsigned char)x == boxpiece[which[i]]) return 1;
  return 0;
}

enum {RIGHT=1,LEFT=2,UP=3,DN=4};

int boxdraw(key) int key; {
extern int row,col,ROWMAX; extern char **line;
int x,y,r,c;
#define UC unsigned char
  switch(key){
  case UPARROW:    y=UP; break;
  case DOWNARROW:  y=DN; break;
  case LEFTARROW:  y=LEFT; break;
  case RIGHTARROW: y=RIGHT; break;
  default: return 0;
  }
/* got a valid key 1-4, build 4-bit key into table */
  x=0; c=col+curoff; r=row;
  if(0<r && c<strlen(line[r-1]) && boxmatch((UC)line[r-1][c],NORTH)) x += 8;
  if(c+1<strlen(line[r]) && boxmatch((UC)line[r][c+1],EAST)) x += 4;
  if(r+1<ROWMAX && c<strlen(line[r+1]) && boxmatch((UC)line[r+1][c],SOUTH)) x += 2;
  if(0<c && c-1<strlen(line[r]) && boxmatch((UC)line[r][c-1],WEST)) ++x;
  return findnewchar(x,y);
}

#if 1
/* A one-time warning window. */
#define HelpSheetROWS 3
#define HelpSheetCOLS 5

#if 0
struct windows {
          char *title_center;                /* centered title */
          int top_row,top_col,               /* window coordinates*/
              bot_row,bot_col;
          struct textinfo oldw;              /* old window coordinates */
          char *info_center;                 /* centered information */
};
#endif

struct windows helpsheet =
{"",
2,1,(2+1+HelpSheetROWS),(1+HelpSheetCOLS),
0,0,0,0,
""};
int ishelpsheetopen=0;

void WarningMessage(HlpText) char **HlpText; {
int i,width,len,spaces;
struct windows *w;
extern LONG ccBIOS();
extern int COLMAX;
extern int ishelpsheetopen;
extern char *strchr();
  w= &helpsheet;
  if(ishelpsheetopen==1){
    bell(); clrcon(); banner("More messages. Press a key:"); getkey();
    closehelpsheet(w);
  }
  if(!loadInfoText(w,HlpText)) goto err;
  if(winopen(w)){
    setattribute(windattr);
    len=w->bot_row-w->top_row-1;
    for(i=1;i<=len;++i) {
      mycputs(1,i,HlpText[i-1]);
      width=1+strlen(HlpText[i-1]);
      mygotoxy(width,i);
      spaces=w->bot_col-w->top_col-width;
      aputc(' ',spaces);
    }
    winrestorewindow(w);
    ishelpsheetopen=1;
    setcursortype(1);
    curses();
  }
err:;
}

#if 0
struct windows {
          char *title_center;                /* centered title */
          int top_row,top_col,               /* window coordinates*/
              bot_row,bot_col;
          struct textinfo oldw;              /* old window coordinates */
          char *info_center;                 /* centered information */
};
#endif
int loadInfoText(w,HlpText) struct windows *w; char **HlpText; {
int len,nrows,k;

  len=2+max(strlen(w->title_center),strlen(w->info_center));
  for(nrows=0;HlpText[nrows]!=0;++nrows){
    len=max(len,strlen(HlpText[nrows]));
  }
  /* nrows=number of matrix rows, len=number of matrix columns */
  k=COLMAX-len;
  w->bot_col=COLMAX;              /* Justify matrix right side of screen */
  w->top_col= (k>0) ? k-1: 0;
  w->top_row=2;                   /* And just below the status line */
  w->bot_row=2+1+nrows;
  if(2+len>COLMAX || 3+nrows>ROWMAX) return 0;
  return 1;
}



void closehelpsheet() {
extern int ishelpsheetopen;
    if(ishelpsheetopen==1){
      ishelpsheetopen=0;
      wintextreplace(&helpsheet);
      setcursortype(1);
      curses();
    }
}

#endif
