/*
 * V6.C
 */

#include "v.h"

/*
 * JOIN
 *
 * Purpose:
 *              Join two lines in insert mode.
 * Notes:
 *              o Must be in insert mode to join two lines.
 *                Otherwise, the command reverts to backspace.
 *              o If cursor at col=0 and key=backspace, then
 *                join this line onto previous, cursor at break.
 */
void join() {
char r[1];
char *p,*q;
int i,j;
int strlen();
extern char *toprint();

  if(row) {
    j = strlen(p=line[row]);
    i = strlen(q=line[row-1]);
    if(i+j<=MAXLN) {
      strcpy(q+i,p);
      p[0] = EOS; delln(r);
      decrROW();
      curfix(i+1,0);
      xPrint(&line[row][i],row,col);
      lcount();
      return;
    }
  }
  bell();
}

/*
 * curfix
 *
 * Fix up cursor and screen offset for tabs, off-screen searches.
 * k=abs col position on line (1=first position);
 */
int curfix(k,len) int k,len; {
int j,flg;
    flg=0;
    if(k<=0) return FALSE;
    col=k-1;
    if((j=col-curoff-COLMAX)>0){/* target off screen right */
      curoff += j+len;       /* this always works */
      /* but we must show COLMAX columns */
      /* and the whole search string too! */
      if(MAXLN-1 < COLMAX+curoff) curoff = MAXLN-1-COLMAX;
      if(curoff>col) curoff=col;
      col -= curoff;
    }
    else
    if(col<curoff){     /* target off screen left */
      if(col+len < COLMAX) curoff=0; else {curoff=col; col=0;}
    }
    else {      /* target on screen, but might over-hang */
      col -= curoff;
      flg=1;
    }
    /* now have 0 <= col <= COLMAX */
    if((j=len+col-COLMAX)>0){  /* string not all on screen */
       curoff += j; col -= j; flg=0;
    }
    if(flg==0) drawall();
    return TRUE;
}

/*
 * FFIND
 *
 * Purpose:
 *              Find a string forward in file from current
 *              position plus one.
 * Returns:
 *              TRUE  - if found
 *              FALSE - not found
 * Notes:
 *              o Matching line brought to center of screen.
 *              o Cursor on first character of match. More
 *                than one match per line possible.
 *              o midstr(s) updates taberr variable.
 *                On return, ++taberr for
 *                each tab found.
 */
int ffind() {
int k,m,j,flag;
int i;
LLONG v;
int offset,off;
char far *p;
int midstr();
int Crefreshh();
extern int zero;

  flag=v = 0; offset=col+curoff; i=strlen(line[row]); j=min(offset+1,i);
  if(offset<i){
    if(zeroleft) goto dorows; /* can't match beginning of line*/
    k = midstr((char far *)line[row],&j);
    if(k){k += offset+1; goto quit; }
  }
dorows:
    /* Search the screen rows after the current row */
    for(i=row+1;i<=ROWMAX;++i) {
      j=0; ++v;
      if((k = midstr((char far *)line[i],&j)) != 0) goto quit;
    }
    m=ebuf-hbuf; p=hbuf; j=0; flag=1;
    while(j<m) { ++v; if((k = midstr(p,&j)) != 0) goto quit; }
    /* search upper blocks */
    off=finfo[curfile].nbLOW;
    while(1){
      /* don't need to copy it to search */
      if((i=getmeminfoindex(curfile,off++))== -1) break;
      m=meminfo[i].numbytes; /* don't need to copy it to search */
      p=getmemblock(meminfo[i].record);
      j=0;
      while(j<m){ ++v;
      /* DEBUG: k=0; do { sysv[k]=p[j+k]; } while(sysv[k++]);
      getnumquery(sysv,&j,0,m); */
      if((k = midstr(p, &j)) != 0) goto quit; }
    }
  goto failed;
quit:
  if(Crefreshh((LLONG)v)) {
    j=0;
    if(flag) k = midstr((char far *)line[row],&j);
    return showmatch(k);
  }
failed:
  return FALSE;
}

middle(v) LLONG v; {
LLONG x; int y;
 x=lncount-1+v; y=ROWMAX/2;
 if(x<=y) y=(LONG)x;
 return y;
}

/* Return 0 if no match otherwise nonzero */
/* Applies to current line from current position only! */
/* It is supposed to match blank lines! */
/* If the cursor is in col=0 and line="abc" and srcharg="a" */
/* then srcharg matches in a foward search at position k=1 */
/* however it does not match in a backward search. */
int findlinematch(n) int n; {
int i,j,k,sl,offset; char *p;
extern char *strchr(); extern int zeroleft,zeroright;
 sl=strlen(p=line[row]);
 j=offset=col+curoff;
 k=0;
 if(n==4) { /* Search reverse for match on same line */
  strcpy(sysv,p); sysv[offset]=0;
  if(zeroright && offset < sl-1) goto quit;
  k=findlastmatch((char far *)sysv);    /* k==0 if no match */
 }
 if(n==3) { /* Search forward for match on same line */
  if(offset>sl) goto quit;           /* Nothing here to match */
  if(zeroleft && offset>0)   /* can't match beginning of line*/
    goto quit;
  k = midstr((char far *)p,&j); /* See if srcharg[] matches */
  if(k) k += offset;
 }
quit:
  return (k? showmatch(k):FALSE);
}

/*
 * RFIND
 *
 * Purpose:
 *              Find a string in reverse direction.
 * Returns:
 *              TRUE  - if found
 *              FALSE - not found
 * Notes:
 *              o midstr() updates taberr variable.
 *                On return, ++taberr for
 *                each tab found.
 */
int rfind() {
int k,w;
int i,off,j;
LLONG v;
char far *p;
int midstr();
int Crefreshh();

  strcpy(sysv,line[v=row]); sysv[j=col+curoff]=0;
  k=findlastmatch((char far *)sysv);   /* k==0 if no match */
  if(k!=0 && k!=j+1) goto quit;        /* k==j+1 does not move cursor */
  while(--v != -1){ /* Back up on screen rows */
    if((k=findlastmatch((char far *)line[(int)v])) != 0) goto quit;
  }
  w= 1+lbuf-bufbase; p=bufbase;
  while(w>0) {
  w=startof(p,w-1);
  if((k = findlastmatch(p+w)) != 0) goto quit; --v; }
  off=finfo[curfile].nbLOW-1;
  while(off>=0){
    /* don't need to copy it to search */
    if((i=getmeminfoindex(curfile,off--))== -1) break;
    p=getmemblock(meminfo[i].record); w= meminfo[i].numbytes;
    while(w>0){
    w=startof(p,w-1);
    if((k = findlastmatch(p+w)) != 0) goto quit; --v;}
  }
  goto failed;
quit:
  if(k && Crefreshh((LLONG)(v-row))) {
    if(v<0) k=findlastmatch((char far *)line[row]);
    return showmatch(k);
  }
failed:
  return FALSE;
}

/* Find matches left-to-right, return last match, if any */
int findlastmatch(p) char far *p; {
int k,j,n; extern int zeroleft;
  n=k=0;
  if(zeroleft){j=0; return midstr(p,&j);}
  do{
    n += k; j=n;
    k = midstr(p,&j);
  } while(k);
/*    n==0 for no match at all */
/*    n>=1 means match at n-1 offset into string*/
  return n;
}

void showsearch(n) int n;{
 char v[256];
  if(nodisp) return;
  ssprintf(v,"%sSearch%s: %s",
             (n ? ((n==1 || n==3) ? "Forward " : "Backward "):""),
             (n? "" : " Failed"),srcharg);
  banner(v); if(n==0) bell();
}


search(n) int n;{
int i;
char *strchr(); extern int DETABBER,taberr;
  i=0;
  if(srcharg[0]==0) goto quit;
  if(FOLDCASE) cvupper(srcharg);
  taberr = DETABBER && (strchr(srcharg,' ') != (char *)0);
  showsearch(n);
  switch(n){
  case 1: i=ffind(); break;
  case 2: i=rfind(); break;
  case 4:
  case 3: i=findlinematch(n); break;
  }
quit:
  if(!i) showsearch(0);
  lnumcount();
  curses();
  return i;
}

/* Show string match to srcharg at col=k-1 */
int showmatch(k) int k; {
extern char *sysx;
extern int srchcount();
int c;
  if(!k) return FALSE;
  curfix(k,srchcount());
  /* Don't print off screen */
  strnzcpy(sysx,&line[row][k-1],min(srchcount(),COLMAX-col));
#if BSDunix || SYS5
  curses(); putcursor(row,col+1); strinvvideo(sysx+1); curses();
#else
  curses(); putcursor(row,col); strinvvideo(sysx); curses();
#endif
  return TRUE;
}

void unshowmatch(){
  curses(); ePrint(sysx);
}

/* Count number of real search chars in srcharg[] */
int srchcount(){
extern int zeroleft,zeroright; extern char *srcharg;
int n;
  n=strlen(srcharg)-zeroleft-zeroright;
  return (n? n : 1);
}

void setinputoutput(){
extern int intexttype,outtexttype,blockflag;
  getnumquery("(1=DOS,2=MAC,3=UNIX) Input",&intexttype,1,3);
  getnumquery("(1=DOS,2=MAC,3=UNIX) Output",&outtexttype,1,3);
  getnumquery("(0=OFF,1=prompts ON) Block read",&blockflag,0,1);
}

/* transpose 2 chars before cursor */
int chartranspose(){
unsigned char r[3];
unsigned char *p; int n;
unsigned char *curchar();
  n=curoff+col;
  if(n>1){
    p=curchar()-2;
    r[2]=0; r[0]=p[1]; r[1]=p[0]; p[0]=r[0]; p[1]=r[1];
    if(col==1) xPrint(&r[1],row,0);
    if(col>1)  xPrint(r,row,col-2);
    return 1;
  }
  return 0;
}

/* Transpose line at cursor with previous line. */
void linetranspose(){
  delln(sysvv); edup(); curses(); addln(sysvv,TRUE); eddn();
}

/* Formats a buffer for output to a printer */
/* Using global variables from printer setup */
void printformat(){
extern int PRTleft,PRTtop,PRTbot,PRTlpp,PRTff;
int i,tmp; LLONG total;
  PUSHnodisp();
  nodisp=1;
  eofile(); total=lncount;
  topfile();
  fillchr(sysvv,' ',PRTleft); sysvv[PRTleft]=0;
  while(0<total){        /* do all lines in file */
    for(i=0;i<PRTtop;++i) {addln("",FALSE); eddn();}
    for(i=0;i<PRTlpp;++i,--total){
      addstring(sysvv,line[row],FALSE,1); eddn(); homeit();
    }
    for(i=0;i<PRTbot;++i) {addln("",FALSE); eddn();}
    if(PRTff){addln("\f",FALSE); eddn();}
  }
  POPnodisp();
  topfile();
}

void printsetup(){
extern int PRTleft,PRTtop,PRTbot,PRTlpp,PRTff;
  getnumquery("Left spaces",&PRTleft,0,MAXLN-1);
  getnumquery("Top blank lines",&PRTtop,0,-1);
  getnumquery("Bottom blank lines",&PRTbot,0,-1);
  getnumquery("Text lines per page",&PRTlpp,1,-1);
  getnumquery("Use formfeeds",&PRTff,0,1);
}

/* set block marker at cursor position */
void markblock(offset) int offset;{
extern LLONG markline,lncount;
  markset=1; markcol=offset; markline=lncount;
}

LLONG getoffset(){
extern LLONG markline,lncount;
  return (markline-lncount);
}
/* Swap current cursor position with marked position. */
void swapmarker(){
extern int markcol;
extern LLONG lncount;
extern LLONG getoffset();
int ii;
LLONG x,iii; extern int Crefreshh();

  iii=getoffset();           /* iii=offset to target */
  ii=markcol;                /* ii=col of target */
  markblock(col+curoff);     /* Change marks */
  curfix(ii+1,0);            /* Put cursor on other mark */
  x=row+iii;
  if(0 <= x && x <= ROWMAX) {lncount += iii; row=(int)x;}
  else { Crefreshh(iii); }
}

/* Show marked region in inverse video. Only affects screen. */
/* flag=0 is for regular chars, flag=1 for inverse video chars */
void showregion(n,flag) int n,flag;{
int i,j,start,end,startcol,endcol;
LLONG ii;
extern LLONG lncount;
extern LLONG getoffset();

  ii=getoffset();
  if(ii>=0){
    start=row;
    end= (ROWMAX<row+ii) ? ROWMAX : (int)(row+ii);
    startcol=col;
  } else {
    start=(0>row+ii) ? 0 : (int)(row+ii);
    end=row;
    startcol=max(0,markcol-curoff);
  }
  if(start>end) return;
  if(!n) startcol=0;
  endcol=max(markcol,col+curoff);
  for(i=start;i<=end;++i){
    strcpy(sysx,line[i]);
    if(n){j=strlen(sysx); while(j<=endcol) sysx[j++]=' '; sysx[endcol+1]=0;}
    sysv[curoff+COLMAX+1]=0;
    putcursor(i,startcol);
    if(flag) strinvvideo(sysx+curoff+startcol);
    else
    ePrint(sysx+curoff+startcol);
  }
  curses();
}

/*
 * set cursor at upper left corner of rectangle,
 * mark at lower right corner of rectangle.
 */
void setcorner(){
extern LLONG getoffset();
int x;
  if(getoffset() <= 0){
    if(col+curoff<markcol){x=col+curoff; curfix(1+markcol,0); swapmarker(); curfix(1+x,0);}
    else swapmarker();
  } else {
    if(markcol<col+curoff) curfix(1+markcol,0);
    else return;
  }
  curses();
}

void badregion(){
  banner("Block not marked. On menu use: ");
  strinvvideo("Begin block set mark");
}

/* Show marked region, get query, restore region, return query char. */
int regionbanner(msg,options) char *msg; char *options; {
register int n;
int i; LLONG x;
char *endof();
char *strchr();
extern int regdef;
extern LLONG lncount;
extern LLONG getoffset();
extern LLONG absLL();

  n=regdef;
top:
  showregion(n,1);
  x=absLL(getoffset());
#if LLONGislong
#define REGIONFMT "Region: %ld %s "
#else
#define REGIONFMT "Region: %u %s "
#endif
  ssprintf(sysv,REGIONFMT,(LLONG)(x+1),(n? "by":"lines."));
  if(n)
    ssprintf(endof(sysv),"%u. ",(unsigned)(abs(markcol-col)-curoff+1));
  ssprintf(endof(sysv),"SPACE toggles view. %s: ",msg);
  if(options[0]) ssprintf(endof(sysv)-2,"? [%s]: ",options);
  banner(sysv);
  while(strchr(options,i=getupper())==(char *)0){
    if(strchr("\r\n\t ",i)!=(char *)0) break;
    bell();
  }
  showregion(n,0);
  if(i==' ') {n=1-n; goto top;}
  regdef=n;
  return i;
}

void killline(){            /* Emacs style kill line */
char r[2]; int i,j,flag; char *p;
extern int killedcrlf;  /* Flag which tells us if we killed a crlf */
  /* erase to end of line */
  /* and save a copy of the line */
  i=curoff+col;
  j=strlen(p=line[row]);
  flag=secondkill();
  if(j > i){
    if(addtolist(p+i,flag,0)){
      p[i] = EOS;
      eraseol();
      goto done;
    }
    goto failed;
  } else if(i>0 || (flag && !killedcrlf)) {
    killedcrlf=1;
    homeit(); eddn(); curses(); join();
  } else { /* i=j=0 */
    if(addtolist(p,flag,1)) delln(r); else goto failed;
  }
  goto done;
failed:
done:
  curses();
  return;
}

#if 0
/* The killerlist is an array of MAXKILLER far addresses. */
/* Each address is the base of a null-terminated string in the far data */
/* segment. The killerlist has killerlen active strings. */
/* When the list must be reused, then the strings  must be purged. */
/* This process must copy all of the string data and the structure */
/* killerlist[killerlen], then save the address of the structure and */
/* its length in the structure blocklist[]. */

typedef char far * FARSTR;      /* FARSTR data is 4 bytes */
struct bigarray {FARSTR str;};  /* killerlist points to this structure */
struct biglist {                /* Container structure to hold saves */
  struct bigarray far *
      blockbase;                /* Address of the array of far strings */
  int
      blocklen;                 /* Number of strings in that array */
};
#endif

#define NKiller 10     /* near data segment contains these arrays */
struct biglist blocklist[NKiller]={0,0,0,0,0,0,0,0,0,0};  /* 60 bytes */
int killedcrlf=0;

#define MKiller 50     /* near data segment contains these arrays */
char *srchlist[MKiller]={0,0,0,0,0,0,0,0,0,0,
                         0,0,0,0,0,0,0,0,0,0,
                         0,0,0,0,0,0,0,0,0,0,
                         0,0,0,0,0,0,0,0,0,0,
                         0,0,0,0,0,0,0,0,0,0};    /* 100 bytes */

extern struct bigarray far *killerlist;
extern int killerlen;                   /* Active length of killerlist */
extern int MAXKILLER;                   /* Maximum length of killerlist */
extern int killedcrlf;  /* Flag which tells us if we killed a crlf */

int getkillercore(){
int i;
#if MSDOS
    killerlist=farmalloc((unsigned long)MAXKILLER*sizeof(char far *));

#else
    killerlist=malloc((int)MAXKILLER*sizeof(char far *));
#endif
if(killerlist==0) bellpressany("bad farmalloc in getkillercore ");
    return ((killerlist==0) ? 0 : 1);
}

void addfailed() {
extern char *strchr();
extern void purgelist2notepad();
unsigned x;
  macabort(); repeats=1;
  bellpressany("Purging strings to NotePad");
  purgelist2notepad(1); poplist(1); killerlen=0;
  curses();
}

int addtolist(s,flag,crlf) char *s; int flag,crlf; {
FARSTR p;
  killedcrlf=crlf;
top:
  if(!flag && !savelist())
     goto failed;               /* Out of memory */
  /* Add string into list */
  if(killerlen<MAXKILLER){
    if(farsafecore(&p,s)){killerlist[killerlen++].str=p; return 1;}
  }
  if(flag){flag=0; goto top;}
failed:
  addfailed();
  return 0;                     /* failed */
}

void killblk(q,n) struct bigarray far *q; int n; {
int j;
  for(j=0;j<n;++j) farfree(q[j].str);
  farfree(q);
}


int savelist() {
struct bigarray far * q;
int i,j,n;
  q=blocklist[NKiller-1].blockbase; n=blocklist[NKiller-1].blocklen;
  if(q) killblk(q,n);
  for(i=NKiller-1;i>=1;--i){     /* Shuffle pointers */
    blocklist[i].blockbase=blocklist[i-1].blockbase;
    blocklist[i].blocklen=blocklist[i-1].blocklen;
  }
  blocklist[0].blockbase=0; blocklist[0].blocklen=0;
  if(!killerlen) return 1;
#if MSDOS
  {
  q=farmalloc((unsigned long)sizeof(struct biglist)*killerlen);}
#else
  {
  q=malloc((int)sizeof(struct biglist)*killerlen);}
#endif
  if(q){
    blocklist[0].blockbase=q; blocklist[0].blocklen=killerlen;
    for(i=0;i<killerlen;++i) q[i].str=killerlist[i].str;
    killerlen=0;
    return 1;
  }
  bellpressany("farmalloc failed in savelist");
  return 0;
}

void purgelist2notepad(flag) int flag;{
int i,j,n;
struct bigarray far *q;
 i=NKiller;while(--i>=0){
     if((q=blocklist[i].blockbase)==0) continue;
     n=blocklist[i].blocklen;
     if(flag) pushlines(0,0,q,(LLONG)n);
     killblk(q,n);
     blocklist[i].blockbase=0; blocklist[i].blocklen=0;
  }
 i=MKiller;while(--i>=0){
     freeup(srchlist[i]); srchlist[i]=(char *)0;
  }
}


/* paste the kill list at the cursor position */
/* or paste onto the end of the notepad */
int poplist(flag) int flag; {
int i,j,k,n; FARSTR p;
extern unsigned char *curchar();
   if((n=killerlen)!=0){
     k=0;
     if(flag) pushlines(0,0,killerlist,(LLONG)killerlen);
     else {
       if(*curchar() || line[row][0]==0) divide();
       fartonearstrcpy(sysv,killerlist[0].str);
       if(addstring(sysv,line[row],TRUE,0)== -1) return 0;
       if(--n){eddn(); curses();}
       k=1;
     }
     for(j=ROWMAX-row+1,i=killerlen-1;i>=k;--i) {
       p=killerlist[i].str;
       if(flag){
          farfree(p);
       } else {
         fartonearstrcpy(sysv,p);
         addln(sysv,(i<=j));
       }
     }
     if(k && n){edup(); curses();}
   }
   return killerlen;
}

/* stack for previous search arguments */
/* look through it with uparrow, downarrow on status line - see v1.c */
/* Remove last item in stack, add srcharg, copy new item to srcharg */
/* Detect special features of begin/end line matches */
/* List always contains current search arg on top */

static int firstone=0;          /* where to insert next save */
static int gotfirstone=0;

void copysrcharg(s) char *s; {
int n,m,k;
char *p;
extern int zeroleft,zeroright;
  if(s[0]==0) goto done;
  gotfirstone=1;
  for(n=0;n<MKiller;++n){
    if(srchlist[n]!=0 && strcmp(srchlist[n],s)==0)
      goto duplicate; /* already there */
  }
  freeup(srchlist[firstone]);
  safecore(&srchlist[firstone],s);              /* 0=fail,1=worked */
  firstone=circularlist(firstone,-1);           /* Next one to retire */
done:
  zeroleft=zeroright=0;                 /* No special position on line */
  if(s[0]==zero) zeroleft=1;            /* Match start of line */
  if(tailchar(s+1)==zero) zeroright=1;  /* Match end of line */
  strcpy(srcharg,s);                    /* Top of list and srcharg match */
  return;
duplicate:
  k=circularlist(firstone,1);   /* Most recent position */
  if(n==k) goto done;           /* All done. */
  p=srchlist[n];                /* Move it to the most recent position */
  while(1){
    m=circularlist(n,-1);
    srchlist[n]=srchlist[m];    /* Fold the array over on itself */
    if(m==k) break;
    n=m;
  }
  srchlist[k]=p;
  goto done;
}

int circularlist(n,flag)
int n,flag;
{
  if(flag<0){ if(++n >= MKiller) n=0;}
  else { if(--n <0) n=MKiller-1;}
  return n;
}


/* p=source string. */
/* d=direction, 1=backward, -1=forward */
/* Copy to p either srcharg or else next search string from list */
/* Works with cursor keys on status line. If UPARROW, then cycle */
/* backwards in the list until the end is reached. If DOWNARROW, */
/* then cycle forward until srcharg is reached. */
void getsrcharg(p,d) char *p; int d; {
int m,n;
static int lastone=0;           /* Marker for where to look first */

  if(gotfirstone==0){p[0]=0; return;}
  if(!p[0]) goto std;
  if(strcmp(p,srcharg)==0) { /* Always, srcharg matches top of list */
    if(d<0) return;          /* Can't move further in minus direction */
    lastone=firstone;        /* Start at top of list again */
    goto search;
  }
/* Detect when at the bottom of the list or the tp of the list */
search0:
  n=lastone;
  m=circularlist(n,d);
  if(d>0){
    if(srchlist[m]==0) return; /* partially-filled list */
    if(n==firstone) return;  /* next one wraps around */
  } else { /* d < 0 */
    if(srchlist[m]==0) goto std; /* partially-filled list */
    if(n==circularlist(firstone,1)) goto std;
  }
search:
  n=lastone;
search1:
  while(1){
    n=circularlist(n,d);
    if(srchlist[n]!=0){
      if(strcmp(srchlist[n],p)!=0) goto finished;
    }
    if(n==lastone) break;
  }
  /* All list entries are different, by construction. Two possible */
  /* events: (1) Empty list. (2) List has 1 entry and it matches p. */
  return;
std:
  n=circularlist(firstone,1);    /* index of top of list */
  strcpy(p,srchlist[n]);
  lastone=n;
  return;
finished:
  strcpy(p,srchlist[n]);
  lastone=n;
}

/* Yank all kill list members at the cursor position, interactive */
/* Yank occurs one at a time in reverse chronological order */
/* with interactive query about keep, ignore, quit */
int yanklist() {
int i,j,k,n,total; extern int regdef;
extern struct biglist blocklist[NKiller];

  for(j=ROWMAX-row+1,total=i=0;i<NKiller;++i){
   k=blocklist[i].blocklen; n=0;
   markblock(0);
   while(--k>=0){
     fartonearstrcpy(sysv,blocklist[i].blockbase[k].str);
     if(addln(sysv,(k<=j))) ++n;
   }
   if(n){
     markline += n-1; regdef=0;
     k=regionbanner("Recover","YNQ");
     if(k=='Y'){total += n; continue;}
     massdelete(0,(LLONG)n);
     if(k=='Q') goto end;
   }
  }
end:
  lcount();
  return total;
}

/*
 * ZAP
 *
 * Remove match to optbuf from current line
 */
int zap(optbuf) char *optbuf;{
int n; char *p;
  if(strlen(p=line[row])>col+curoff){
    p += (col+curoff);
    n=strlen(optbuf);
    if(strncmp(p,optbuf,n)==0){
      strcpy(p,p+n);
      xPrint(p,row,col); eraseol();
      return TRUE;
    }
  }
  return FALSE;
}

void dozap(rows,cols)
int rows,cols;
{
char *p;
int i,j,offset;
  offset=curoff+col;
  for(i=0;i<rows;++i){
    j=strlen(line[row]);
    if(j>offset){
      p = &line[row][offset];
      curses(); eraseol();
      if(offset+cols >= j) p[0]=0;
      else {strcpy(p,p+cols); ePrint(p);}
    }
    curses();
    eddn();
  }
  for(i=0;i<rows;++i) {curses(); edup();}
}

int isalpha(c) int c; {
  if(('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || ('0' <= c && c <= '9')) return 1;
  return 0;
}

/* return end of string address or address of first upper or lower char */
char *sobl(s) char *s; {
  while(s[0] && !isalpha(s[0])) ++s;
  return s;
}

/* nextword() contains KILL WORD for n==1. */
/* For n==0 it is normal word ahead and word back. */
int nextword(n) int n;{
int k,x,y,flag,wasvisible; char *p,*q;
FARSTR r;
char *sobl(); char *sob();
  y=col+curoff;
  flag=0;
  if(n && !secondkill()) addtolist("",0,0);
  if(strlen(q=line[row])>y){
    p=sob(q+y);
    if(q[y]!=' ') {
      k=0;
      while((x=p[0])!=0 && isalpha(x)){++k; ++p;} /* Skip over alpha */
      if(k==0 && p[0]!=0) ++p;            /* No alpha chars, then take one */
    } else x=p[0];
    wasvisible = (*sob(q+y)!=0);  /* wasvisible==1 if we removed visible chars*/
    k=(int)(p-q);
    if(n){      /* KILL WORD for n==1 */
      k -= y;
      if(k){strnzcpy(sysvv,q+y,k); zap(sysvv); flag=1;}
      if(killedcrlf) addtolist("",1,0);
      if(flag) {
        strcpy(sysx,sysvv);
        r=killerlist[--killerlen].str;
        fartonearstrcpy(sysvv,r); farfree(r);
        strcat(sysvv,sysx);
        addtolist(sysvv,1,0);
      }
      if(!wasvisible) goto nextline;
      goto end;
    }
    if(x){ curfix(1+k,0); goto end;}
  }
nextline:
  homeit(); eddn();
  if(n){curses(); join(); killedcrlf=flag=1;}
end:
  return flag;
}

void prevword(){
int n;
char *p,*q; char *sobl();
  n=strlen(q=line[row]);
  if(n){
    p = q + ((n>col+curoff) ? (col+curoff) : n);
    if(p==q) goto prevline;
    if(isalpha(p[0]) && p!=q && !isalpha(p[-1])) --p;
    else {
     while(p!=q && !isalpha(p[0])) --p;     /* skip over non-alpha */
     while(p!=q && isalpha(p[-1])) --p;     /* skip until non-alpha */
    }
    curfix(1+(int)(p-q),0); return;
  }
prevline:
  if(lncount==1) {homeit(); return;}
  edup(); endit();
}
