/*
 * V8.C
 *
 *              This file contains special OS functions and
 *              standard C library functions missing from the
 *              main library for the target C compiler.
 *
 *              For supported OS, see header file V.H.
 */

#include "v.h"

#define COLUMNS 1       /* using 1-column output */
/*
 * FREESP
 *
 * Purpose:
 *              Computes the number of free kilobytes on the disk.
 * Returns:
 *              Message to print on status line, in area sysv[].
 */
#if BSDunix || SYS5
VOIDS freespace() {
char buf[512];
int x,y,NCOLS;
char *p;
LONG ln;
char *endof(),*strchr();
LLONG atoiLL();

  x = ln = 0;
  if(sysv[0]) {
    if((p = strchr(sysv,',')) != 0) {
      *p = '\0'; NCOLS = (int)atoiLL(p+1);
    } else NCOLS = COLUMNS;
    strcpy(buf,sysv);
    x=dir(buf,NCOLS,0,(char *)0,(char **)0);
    ssprintf(sysv,"%u File%s",(unsigned)x,(x == 1 ? "":"s"));
  }
  return ( (x <= 0) ? 0 : 1);
}

char *getwdir(buf) char *buf; {
extern char * getcwd();
 return getcwd(buf,MAXFNAME-1);
}
#endif


#if MSDOS
int freespace() {
char buf[128];
int y,NCOLS;
char *p;
char *endof(),*strchr();
LLONG atoiLL();

  y=0;
  if(sysv[0]) {
    if((p = strchr(sysv,',')) != (char *)0) {
      *p = '\0'; NCOLS = (int)atoiLL(p+1);
    } else NCOLS = COLUMNS;
    strcpy(buf,sysv);
    y = dir(buf,NCOLS,0,(char *)0,(char **)0);   /* reads default disk */
    if(y != -1)
      ssprintf(sysv,"%u File%s",(unsigned)y,(y == 1 ? "":"s"));
  }
  return ( (y <= 0) ? 0 : 1);
}
#endif

/*
 * DIR
 *
 * Purpose:
 *              Inject disk directory in five-column format.
                Uses currently logged disk.
 * Notes:
 *
 */
#if BSDunix || SYS5

#include <sys/types.h>
#include <dirent.h>

int findfirst(fp,t,s) DIR *fp; char *t,*s; {
struct dirent *dp;
for(dp = (struct dirent *)readdir(fp);
    dp != (struct dirent *)0;
    dp = (struct dirent *)readdir(fp)){
  if(matches(t,dp->d_name)==0) {strcpy(s,dp->d_name); return 0;} /* found it */
}
strcpy(s,"");
rewinddir(fp);
return 1; /* not found */
}

matches(t,s) char *t; char *s; {
int i,status;
status=0;
while(1) {
  if(t[0]==0) break; /* end of string */
  if(s[0]==0) break; /* end of string */
  if(t[0]=='*') {
    /* Does s match a substring of t+1? */
    i=0; while(t[1+i]==s[i] && s[i]!=0) ++i;
    if(s[i]==0){ s += i; t += 1+i; break;}
    if(t[1+i]=='*' || t[1+i]=='?'){
      t += 1+i; s += i;
    } else
    /* Otherwise, leave t[0]=='*' and match current char s[0] */
    ++s;
  } else {
    if(t[0] != '?' && t[0] != s[0]) break;
    ++t; ++s;
  }
}
while(t[0]=='*') ++t;
if(t[0] != 0 || s[0]!=0) status=1;
return status;
}

int isadir(s) char *s; {
DIR *fp;
int i,j;
  j=1;
  if((i=strlen(s))>0 && s[i-1] == '/') j=s[i-1]=0;
  if((fp=(DIR *)opendir(s)) != (DIR *)0) {
    closedir(fp); if(j==0) addslash(s); return 1;
  }
  if(j==0) addslash(s);
  return 0;
}

int iswild(s) char *s; {
char *strchr();
  if(strchr(s,'*') != (char *)0) return 1;
  if(strchr(s,'?') != (char *)0) return 1;
  return 0;
}

void addslash(s) char *s; {
  if(tailchar(s) != '/') strcat(s,"/");
}

void stripstars(buf) char *buf;
{
char *p;
extern char *strstr();
  if((p=strstr(buf,"/*"))!=0 && p[2]==0) p[1]=0;
}

/* split off dir, copy to cd[] */
void dirsplit(s,cd) char *s,*cd; {
extern char *strrchr();
  if(s[0]==0 || strcmp(s,NOTEPAD)==0 || strcmp(s,UNTITLED)==0) return;
  strcpy(cd,s);
  if((s=strrchr(cd,'/'))!=(char *)0) s[1]=0; else cd[0]=0;
}

char *ffname(buf) char *buf; {
char *s;
extern char *strrchr();
  if((s=strrchr(buf,'/')) == (char *)0) s=buf;
  else ++s;
  return s;
}

char *statbyte(s) char *s; {
 return(s+strlen(s)+1);
}

int dir(buf,COLUMN,flag,dirbuff,buffer)
char *buf;      /* file name to match */
int COLUMN;     /* flag for file name completion, or # cols to output */
int flag;       /* 0=dir or lookup, 1=directory list */
char *dirbuff;  /* name of directory in lookup */
char **buffer;  /* file list, max=MAXDIRENT */
{
int n,k;
DIR *fp;
char dirbuf[MAXFNAME]; char testdir[MAXFNAME];
char  s[MAXFNAME]; char t[MAXFNAME]; char work[MAXFNAME];
char *p; char *strrchr(), *cwd(), *endof(), *getenv(), *strchr();
extern char *strstr(), *statbyte();

  if((p=strstr(buf,"//"))!=(char *)0){  /* Double slash gets rid of */
    strcpy(buf,p+1);                    /* previous string */
  }
  if(buf[0]==0) {strcpy(buf,cwd()); addslash(buf);}
  /* assume connected dir */
  if(tailchar(buf)=='/') strcat(buf,"*");
  strcpy(s,buf);
  /* Is is like ~/tmp/class/xx.c? */
  if(buf[0]=='~'){
    p=getenv("HOME");
    if(p != (char *)0){ strcpy(s,p); strcat(s,buf+1);}
  }
  else
  /* Is is like $HOME/tmp/class/xx.c? */
  if(buf[0]=='$'){
    strcpy(t,buf+1);
    p=strchr(t,'/'); if(p!=(char *)0) p[0]=0;
    p=getenv(t);
    if(p!=(char *)0){
      strcpy(s,p);
      p=strchr(buf,'/'); if(p==(char *)0) p="";
      strcat(s,p);
    }
  }
  strcpy(dirbuf,s);
  fp=(DIR *)opendir(dirbuf);
  if(fp==(DIR *)0){ /* It had a file name on the end */
    p=strrchr(s,'/'); if(p==(char *)0) {strcpy(t,s); strcpy(s,cwd());}
    else {strcpy(t,p+1); p[0]=0;}
    /* s=directory string, t=file name pattern */
    strcpy(dirbuf,s);
    fp=(DIR *)opendir(dirbuf);
    if(fp==(DIR *)0) return 0;
  } else t[0]=0; /* no file name given */
  if(t[0]==0) strcpy(t,"*"); /* no file name? then assume all files */
  strcpy(work,dirbuf);
  addslash(dirbuf);
  strcpy(buf,dirbuf); strcat(buf,t);
  n=0;
/* debug */
  k=findfirst(fp,t,s);
  if(COLUMN == -1){
    if(k) {
      strcat(t,"*");
      strcat(buf,"*");
      closedir(fp); fp=(DIR *)opendir(work);
      k=findfirst(fp,t,s);
    }
    if(k==0){
      n=1;
      strcpy(work,dirbuf); strcat(work,s); /* Save last find */
      if(findfirst(fp,t,s)==0) ++n;        /* Oops, can't find */
      else strcpy(buf,work);
    }
    goto quit;
  }
  if(flag) {
    if(dirbuf[0]) strcpy(dirbuff,dirbuf);
    else {strcpy(dirbuff,cwd()); addslash(dirbuff);}
  }
  if(k==0 && flag==0){
      addln(dirbuf,TRUE); eddn(); curses();
  }
  while(k==0) {
   if(strcmp(s,".")!=0){
    if(flag==0){
      ++n; addln(s,TRUE); eddn(); curses(); if(ccBIOS(2)) break;
    } else {
      if(!squirrel(&buffer[n],s) || n>=MAXDIRENT-1) goto quit;
      ++n;
    }
   }
   k = findfirst(fp,t,s);
  }
quit:
  closedir(fp);
  return n;
}

int squirrel(dest,src) char **dest,*src; {
int flag=0;
      if((*dest=(char *)malloc(2+strlen(src)))==(char *)0)
        goto quit;
      strcpy(*dest,src);
      *statbyte(*dest)=' ';
      flag=1;
quit:
     return flag;
}

int getbufferlist(s,last) char **s; int *last; {
register
int i,v;
char ss[MAXFNAME];

  for(v=i=0;i<maxinfo;++i){
    if(finfo[i].active) {
      if(i==curfile) last[0]=v;
      ssprintf(ss,"%s%u. %s",(i<10 ? "0":""),i,finfo[i].fname);
      if(!squirrel(&s[v],ss)) goto quit;
      ++v;
    }
  }
quit:
  return v;
}

#endif

#if MSDOS
char *ffname(buf) char *buf; {
char *s;
extern char *strrchr(),*strchr();
  s=buf; while((s=strchr(s,'/'))!=(char *)0) s[0]='\\';
  if((s=strrchr(buf,'\\')) == (char *)0){
    if((s=strchr(buf,':')) != (char *)0) ++s; else s=buf;
  }
  else ++s;
  return s;
}

/* split off dir, copy to cd[] */
void dirsplit(s,cd) char *s,*cd; {
extern char *strrchr(),*strchr();
char *p;
char buf[MAXFNAME];
  strcpy(buf,s);
  if((p=strrchr(buf,'\\'))!=(char *)0) p[1]=0;
  else
  if((p=strchr(buf,':'))!=(char *)0){p[1]=0; addslash(buf);}
  else buf[0]=0;
  strcpy(cd,buf);
}

int iswild(s) char *s; {
char *strchr();
  if(strchr(s,'*') != (char *)0 || strchr(s,'?') != (char *)0) return 1;
  return 0;
}

void addslash(s) char *s; {
  if(tailchar(s) != '\\') strcat(s,"\\");
}

#include <dir.h>
/*
int isadir(s) char *s; {
  return ((tailchar(s)=='\\') ? 1 :0);
}
*/
int isadir(s) char *s; {
struct ffblk sysfcb;
  if(tailchar(s)=='\\') return 1;
  if(getmod(s)==16) return 1;
  return 0;
}

static void putln(){
  addln(sysv,TRUE); eddn(); curses();
}

char *statbyte(s) char *s; {
 return(s+strlen(s)+1);
}

int getmod(s) char *s; {
extern void __int__();
_DX = (unsigned)s;
_AL=0; _AH=0x43; __int__(0x21);
return (_CX);
}

void pad15(ss) char *ss; {
int j;
      j=strlen(ss); while(j<15) ss[j++] = ' '; ss[j] = EOS;
}


int dir(buff,COLUMN,flag,dirbuff,buffer)
char *buff;      /* file name to match */
int COLUMN;     /* flag for file name completion, or # cols to output */
int flag;       /* 0=dir or lookup, 1=directory list */
char *dirbuff;  /* name of directory in lookup */
char **buffer;  /* file list, max=MAXDIRENT */
{
struct ffblk sysfcb;
char *s,*p;
int i,j,k,v,curdrv,destdrv;
char dirbuf[MAXFNAME],destbuf[MAXFNAME],dest[MAXFNAME],ss[MAXFNAME];
char buf[MAXFNAME];
LONG ccBIOS();
int addln(),findfirst(),findnext();
char *strchr(),*strrchr(),*ffname(),*cwd();
#define NAME sysfcb.ff_name
extern char *strstr(), *statbyte();

  v=0;
  cvupper(buff);
  ffname(buff);                  /* change '/' to '\' */
  if((s=strstr(buff,"\\\\"))!=(char *)0){  /* Double slash gets rid of */
    strcpy(buff,s+1);                      /* previous string */
  }
  curdrv=getdrive();  /* bdos(25,0,0) & 255 */
  getwdir(dirbuf);
  if(buff[0]==0) strcpy(buff,cwd());
  destdrv = ('A' <= buff[0] && buff[1]==':') ? (buff[0]-'A') : curdrv;
  if(destdrv == curdrv) strcpy(destbuf,dirbuf);
  else {
    setdrive(destdrv); /* bdos(14,destdrv,0); */
    getwdir(destbuf);
  }
  /* get file name pointer */
  if(isadir(buff)) addslash(buff);
  strcpy(buf,buff);
  k=findfirst(buf,&sysfcb,16);  /* Seek file or directory */
  s=ffname(buf);
  if(!iswild(s)){
    if(k==0 && (sysfcb.ff_attrib & 16) != 0) {addslash(buf); strcat(buf,"*.*");}
    if(k) strcat(buf,"*");
  }
  if(k && iswild(s) && strchr(s,'.')==(char *)0) strcat(buf,".*");
  strcpy(ss,buf);
  if(COLUMN == -1){ /* file name completion */
    strcpy(buff,buf);
    if(k) k=findfirst(buf,&sysfcb,16);       /* look up again */
    if(k==0){
      v=1;
      s[0]=0;
      if(buf==s || NAME[0]=='.') {
        strcpy(buf,destbuf);
        addslash(buf);
      }
      if(strcmp(NAME,".") != 0) strcat(buf,NAME);
      if((sysfcb.ff_attrib & 16) != 0) addslash(buf);
      if(findnext(&sysfcb)==0) ++v; else strcpy(buff,buf);
    } else bell();
    goto quit; /* return with file name in buff[] */
  }

  /* s=filename with path split off */
  s[0]=0; i=strlen(buf); if(i>1 && buf[i-1]=='\\') buf[i-1]=0;
  if(tailchar(buf)==':') addslash(buf);
  chdir(buf);
  getwdir(dest);
  addslash(dest);
  v=0;
  k=findfirst(ss,&sysfcb,16);       /* look up again */
  if(flag) {
    strcpy(dirbuff,dest);
    if(safecore(&buffer[0],"..\\0 ")){ ++v; *strrchr(buffer[0],'0')=0;}
  } else j=sysv[0]=0;
  while(k==0) {
    if(flag==0){
      if(j >= COLUMN) {
        putln();
        j=sysv[0]=0;
        if(ccBIOS(2)) break;
      }
      if(COLUMN==1)
       if(NAME[0] == '.' || destdrv != curdrv) strcpy(sysv,dest);
      strcpy(s,NAME);
      if((sysfcb.ff_attrib & 16) != 0) addslash(s);
      if(COLUMN>1) pad15(s);
      strcat(sysv,s); ++j; ++v;
    } else {  /* make file list into array buffer[n] */
      if(NAME[0]=='.') goto nextone;
      strcpy(ss,NAME);
      if((sysfcb.ff_attrib & 16) != 0) addslash(ss);
      if(!squirrel(&buffer[v],ss) || v>=MAXDIRENT-1) goto quit;
      ++v;
    }
nextone:
    k = findnext(&sysfcb);
  }
  if(j && flag==0) putln();
quit:
  if(curdrv != destdrv) {setdrive(destdrv); chdir(destbuf);}
  setdrive(curdrv); chdir(dirbuf);
  return v;
}

void stripstars(buf) char *buf;
{
char *p;
extern char *strstr();
  if((p=strstr(buf,"\\*.*"))!=0 && p[4]==0) p[1]=0;
}

int squirrel(dest,src) char **dest,*src; {
extern char *strrchr();
extern char *statbyte();
int flag=0;
      strcat(src,"0");   /* make space for status byte */
      if(!safecore(dest,src)) goto quit;
      *strrchr(*dest,'0')=0; *statbyte(*dest)=' ';
      flag=1;
quit:
     return flag;
}

int getbufferlist(s,last) char **s; int *last; {
register
int i,v;
char ss[MAXFNAME];

  for(v=i=0;i<maxinfo;++i){
    if(finfo[i].active) {
      if(i==curfile) last[0]=v;
      ssprintf(ss,"%s%u. %s",(i<10 ? "0":""),i,finfo[i].fname);
      pad15(ss);
      if(!squirrel(&s[v],ss)) goto quit;
      ++v;
    }
  }
quit:
  return v;
}

#endif

/*
 * ETYPE
 *
 * Purpose:
 *              Print a string to console, raw mode.
 */
VOIDS etype(s)
char *s;
{
#if BSDunix || SYS5
extern FILE *termout;
  if(nodisp) return;
  ewrite(fileno(termout),s,strlen(s));
  fflush(termout);
#endif
#if MSDOS  /* most anything else can do this */
  if(nodisp) return;
  ewrite(2,(char far *)s,strlen(s));    /* fast for turbo-c */
                                        /* no fflush needed */
#endif
}

/*
 * term setup
 *
 * Purpose:
 *              System setup - puts terminal in proper mode
 *              for binary 8-bit I/O.
 */


#if TURBOC
void termDoExit(){
   /* Happens in case of shell out to system or exit() */
    cls(); setcursortype(1); ctrlbreak(1);
#ifdef MOUSEBUILTIN
    mousereset();
#endif
}
#endif


#if BSDunix || SYS5
#if IBMAIX
#include <sys/termio.h>
#include <sys/ioctl.h>
#endif
#if ARDENT
#include <sys/43ioctl.h>
#endif
#if SUN350
#include <sys/ttold.h>
#include <sys/filio.h>
#define RAW O_RAW
#define ECHO O_ECHO
#endif
#if ALPHA
#include <sys/ioctl.h>
#include <sys/ioctl_compat.h>
#else
#if DEC3100
#include <sys/ttold.h>
#endif
#endif

#if (ARDENT || SUN350 || DEC3100)
FILE *termout=0;
static struct sgttyb tmp1;

void termcanon(){
   tmp1.sg_flags &= ~RAW;
   tmp1.sg_flags |= ECHO;
   ioctl(fileno(stdin), TIOCSETP, &tmp1);
}

void termraw(){ /* Set terminal to raw mode */
   tmp1.sg_flags |= RAW;
   tmp1.sg_flags &= ~ECHO;
   ioctl(fileno(stdin), TIOCSETP, &tmp1);
}

void termSetup(){
/* Get terminal characteristics */
  ioctl(fileno(stdin), TIOCGETP, &tmp1);
  termraw();    /* set it up the way we want it */
#if defined(SIGWINCH)
    signal(SIGWINCH, win_change);
#endif
    termout=fopen("/dev/tty","wb");
}

void termDoExit(){
    fclose(termout); termout=0;
    termcanon();          /* restore terminal settings */
#if defined(SIGWINCH)
    signal(SIGWINCH, SIG_DFL);
#endif
}
#endif

#if IBMAIX
FILE *termout=0;
static struct termio tmp1;
static struct termio tmp2;

void termcanon(){
   tmp1=tmp2;  /* restore terminal settings */
   ioctl(fileno(stdin), TCSETAF, &tmp1);
}

void termraw(){
    /* set it up the way we want it */
    tmp1.c_iflag &= ~ICRNL;
    tmp1.c_iflag &= ~IXON;
    tmp1.c_lflag &= ~ISIG;
    tmp1.c_lflag &= ~ICANON;
    tmp1.c_lflag &= ~ECHO;
#ifdef OPOST
    tmp1.c_oflag |= OPOST;     /* Enable output postprocessing */
    tmp1.c_oflag &= ~ONLCR;    /* Disable map of NL to CR-NL on output */
#endif
#ifdef NLDLY
    tmp1.c_oflag &= ~(NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
 /* No output delays */
#endif
#ifdef ASCEDIT
      tmp1.c_iflag &= ~ASCEDIT;
#endif
      tmp1.c_iflag &= ~IGNBRK;
      tmp1.c_iflag &= ~BRKINT;
#ifdef CDISABLE
      tmp1.c_cc[VERASE] = CDISABLE;       /* disable erase processing */
      tmp1.c_cc[VKILL] = CDISABLE;        /* disable kill processing */
#endif
#ifdef IEXTEN
      tmp1.c_lflag &= ~IEXTEN;      /* Disable other editing characters.  */
#endif
#if 0                           /* Don't matter */
    tmp1.c_cc[VMIN]=1;
    tmp1.c_cc[VTIME]=0;
#endif
    ioctl(fileno(stdin), TCSETAF, &tmp1);
}

termSetup(){
/* save terminal characteristics */
   ioctl(fileno(stdin), TCGETA, &tmp1);
    tmp2 = tmp1;        /* copy structure */
    termraw();          /* Set raw terminal */
#if defined(SIGWINCH)
    signal(SIGWINCH, win_change);
#endif
    termout=fopen("/dev/tty","wb");
}

void termDoExit(){
    if(termout != 0) fclose(termout); termout=0;
    ioctl(fileno(stdin), TCSETAF, &tmp2);
#if defined(SIGWINCH)
    signal(SIGWINCH, SIG_DFL);
#endif
}
#endif /* IBMAIX */
#endif /* BSDunix || SYS5 */


/*
 * ccBIOS
 *
 * Purpose:
 *              Direct bios call.
 */
#if MSDOS || BSDunix || SYS5

#if BSDunix || SYS5
#define SAVBUFSIZE 1023
#else
#define SAVBUFSIZE 511
#endif

char savbuf[1+SAVBUFSIZE];
int savcount = 0;
int curchr = 0;
int abortkeycount=0;

unsigned bufget() {
unsigned x;
LONG ccBIOS();

if(savcount==0) bufput(ccBIOS(0));
while(ccBIOS(2) != 0 && savcount < SAVBUFSIZE) bufput(ccBIOS(0));
x = (unsigned)savbuf[curchr++];
if(curchr >= savcount) curchr=savcount=0;
return x;
}

#if BSDunix || SYS5
#define SLEEP sleep(1)
#endif

void clrcon() {
LONG ccBIOS();
  abortkeycount=curchr=savcount=0;
  while(ccBIOS(2) != 0) ccBIOS(0);
}

void bufput(x) int x; {
LONG ccBIOS();
  if(savcount < SAVBUFSIZE) savbuf[savcount++] = x;
  if(x == abortkey) {   /* abortkey == ctrl-U normally */
    if(++abortkeycount <= 2) return; /* Three in a row cause abort */
    curchr=0; macabort();
    savbuf[1]=savbuf[0] = x; savcount = 2;
#ifdef SLEEP
    SLEEP;
#endif
    clrcon();
  }
  abortkeycount=0;
}

#if MSDOS
LONG ccBIOS(f)
int f;
{
static unsigned lastscancode=1;
static unsigned lastlowbyte=1;
unsigned x;
int y;
extern int oldbios;

  switch(f) {
  case 2: return bioskey(1+oldbios);
  case 0:
          if(lastlowbyte==0){
            lastlowbyte=1; return lastscancode;
          }
          while((x=bioskey(1+oldbios))==0) ;
          /* Wait for char available */
          /* handle dark plus, ast and enter keys */
          if((bioskey(2)& 0x20)) goto skipGREY; /* Shift status 00100000 */
          switch(x){                            /* i.e., numlock on */
          case 0x4e2b: /* If grey plus */
            x = GREYplus; break; /* Map to Enter key, "ALT =" */
          case 0x372a: /* if grey asterisk */
            x = GREYast; break; /* Map to Menu key, "CTRL underline" */
          case 0xe00d: /* if grey enter key, enhanced keyboard */
            x = GREYent;  /* Map to Enter key, "ALT ="  */
          }
skipGREY:
          lastscancode = (x >> 8);
          lastlowbyte = (0xFF & x);
          /* Change all extended keys to XT keys */
          if(lastlowbyte == 0xe0 && lastscancode != 0) lastlowbyte=0;
            bioskey(oldbios);  /* get char and bump keybuffer pointers */
          return lastlowbyte;
  case 3: return (255 & bufget());
/*  default: banner("ccBIOS not implemented\n"); */
  }
  return 0;
}
#endif

#if BSDunix || SYS5
LONG ccBIOS(f)
int f;
{
long ln;
  switch(f) {
  case 2:
          ioctl(fileno(stdin), FIONREAD, &ln);
          return (LONG)ln;
  case 0:
          return (LONG)(255&(unsigned)getchar());
  case 3: return (LONG)(255&(unsigned)bufget());
  default: banner("ccBIOS not implemented\n");
          return 0;
  }
}
#endif
#endif

int testkey(){
extern int savcount,curchr,halfdone;
  while(ccBIOS(2) != 0) bufput(ccBIOS(0));
  return savcount;
}

#if !HASMOVEMEM
/*
 * MOVEME
 *
 * Purpose:
 *              Memory move without overlap problems.
 */
#if TURBOC
void moveMEM(dest,source,num)
char *dest,*source;
LONG num;
{
  movmem(source,dest,num);
}
#endif

#if !TURBOC
void moveMEM(dest,source,num)
char *dest,*source;
LONG num;
{
char *d,*s;
LONG n;
  d = dest; s = source; n = num;
  if(d <= s || s+n <= d) {
#if HASMEMCPY
    memcpy(d,s,n);
#else
    while(n--) *d++ = *s++;
#endif
  }
  else {
    d += n;
    s += n;
    while(n--) *--d = *--s;
  }
}
#endif
#endif

#if !MSDOS

FILEP fopenbread(s) char *s; {
  return fopen(s,RMODE);
}

FILEP fopenbwrite(s) char *s; {
  return fopen(s,WMODE);
}

closedfile(fp) FILEP fp; {
  return fclose(fp);
}
#endif

#if MSDOS
FILEP fopenbread(s) char *s; {
  return fopen(s,RMODE);
}

FILEP fopenbwrite(s) char *s; {
  return fopen(s,WMODE);
}

closedfile(fp) FILEP fp; {
  return fclose(fp);
}
#endif

#if BSDunix || SYS5

int updateWinsize=0;

void
win_change(sig)
int     sig;
{
extern int updateWinsize;
#if defined(SIGWINCH)
        signal(SIGWINCH, SIG_IGN);
        updateWinsize=1;        /* Schedule an update */
        signal(SIGWINCH, win_change);
#else
        updateWinsize=1; bell();
#endif
}

int resetWinSize(rrow,ccol)int *rrow,*ccol; {
int Rows,Cols;
struct winsize win;
extern int MAXLN,COLMAX,ROWMAX,updateWinsize;

# if defined(TIOCGWINSZ)
    Rows=Cols=0;
    if(ioctl (0, TIOCGWINSZ, &win) == 0){
      if (win.ws_col) Cols = win.ws_col-1;
      if (win.ws_row) Rows = win.ws_row-1;
    }
#else
    Rows=Cols=0;
    banner("TIOCGWINSZ not defined"); sleep(1);
#endif
    if (Rows <= 0 || Cols <= 0){
       if(rrow[0]<=0 && ccol[0] <= 0){
         banner("Must know the screen size, assuming 24x80");
         bell(); sleep(1);
         Rows=24; Cols=80;
       } else { Rows=rrow[0]; Cols=ccol[0];}
    }
    if(MAXLN <= Cols) Cols=MAXLN-10;
    Cols--;   /* ULTRIX wraps in last column, safety factor 1 */
    Rows--;   /* don't use final line, in case it is terminal status */
                /* line (e.g. line 25 on VT100 display) */
                /* ULTRIX scrolls when cursor enters last line */

    updateWinsize=0;
    rrow[0]=Rows; ccol[0]=Cols;
}

doWinUpdate(){
  return updateWinsize;
}
#endif
