 /*
 * V9.C
 *
 */

#include "v.h"

/*
 * TERMINAL-DEPENDANT SOURCE CODE
 *
 * Conventions:
 *
 *      ROWMAX   -  The number of lines on the screen minus 1.
 *      Rows are numbered 0 to ROWMAX.
 *
 *      COLMAX   -  The number of columns on the screen minus 1.
 *      Columns are numbered 0 to COLMAX.
 *
 *      MAXLN    -  The length in characters of the longest
 *      storable string for a screen line.
 *
 */

/*
 * QUIT
 *
 * Purpose:
 *              Elegant exit
 * Notes:
 *              o Cleans up the special keypad settings.
 *              o Clears the screen.
 *              o Exits without a file save.
 */
#if TURBOC
void quit() {
  termDoExit();
  XMSfree();
  _exit();
}
#endif

#if BSDunix || SYS5
#if IBMAIX
#include <sys/ioctl.h>
#endif
#if ARDENT
#include <sys/43ioctl.h>
#endif

#if SYS5
void quit(n) int n;{
  if(n==100){termDoExit(); exit(0);} /* n=0 is init, n=1 normal, n=100 abort */
  exitterminal();
  termDoExit();
  if(n==2){
#if !defined(SIGTSTP) || !defined(SIGSTOP)
#define SIGTSTP 18      /* stop signal from tty */
#define SIGSTOP 19      /* stop signal from tty */
#endif
    kill(0, SIGSTOP);
    setup();
    initterminal();
    return;
  }
  exit(0);
}
#endif

#if BSDunix
void quit(n) int n;{
#if !defined(SIGTSTP)
#define SIGTSTP 18      /* stop signal from tty */
#endif
  exitterminal();
  termDoExit();
  /* n=0 is init, n=1 normal, n=100 abort */
  if(n==2){
    killpg(getpgrp(0), SIGTSTP);
    setup();
    initterminal();
    return;
  }
  exit(0);
}
#endif


#endif

/*
 * INITTERMINAL
 */
#if PIINILOAD && TERMCAP

#if TURBOC
void initterminal(){
  etputs(rs_termcap,1);
#ifdef MOUSEBUILTIN
  if(mousereset()) mousehook(); /* Hook mouse interrupt if hsens nonzero */
#endif
  clearscreen();
  etputs(ks_termcap,1);
  /* Fixup for bad DEC termcap (they invented the vt100?) */
  if(strcmp(ks_termcap,"\033[?1h")==0) etputs("\033=",1);
}

void exitterminal(){
 statblank=TRUE;
 setcolorsUNIX(4);
 drawall();
 clearlines(-1);
 etputs(rs_termcap,1);  etputs(ke_termcap,1);
}
#endif

#if BSDunix || SYS5
void initterminal(){
  etputs(rs_termcap,1);
  if(mouseexists() && mousekeys()) setmouse(1);
  setansicolors();                    /* Set up screen colors */
  setcolorsUNIX(1);
  clearscreen();
  etputs(ks_termcap,1);
  /* Fixup for bad DEC termcap (they invented the vt100?) */
  if(strcmp(ks_termcap,"\033[?1h")==0) etputs("\033=",1);
}

void exitterminal(){
 if(mouseexists()) setmouse(0);
/*  setcolorsUNIX(0); */
/* clearscreen(); */
 statblank=TRUE;
 setcolorsUNIX(4);
 drawall();
/* clearlines(-2); */       /* Erase one-by-one */
 clearlines(-1);
 etputs(rs_termcap,1);  etputs(ke_termcap,1);
}
#endif

#else

#if TURBOC
void initterminal(){
#ifdef MOUSEBUILTIN
  if(mousereset()) mousehook(); /* Hook mouse interrupt if hsens nonzero. */
#endif
}
void exitterminal(){}
#endif

#if BSDunix || SYS5
void initterminal(){}
void exitterminal(){}
#endif


#endif

#if BSDunix || SYS5
  /* Set background and foreground colors */
int syscolor=1;

void setcolorsUNIX(n) int n; {
syscolor=n;
  /* Reset foreground and background colors to default */
  switch(n){
  case 2:
  case 0:  etype("\033c"); etype(UNIXdefaultfgbg);
           break;
  /* Set edit screen foreground and background colors */
  case 1:  etype(UNIXtextfgbg); break;
  /* Set foreground and background colors of popup windows */
  case 3:  etype(UNIXmenufgbg); break;
  case 4: /* etype("\33[1m\33[4m\33[7m\33[0;0m"); */
    freeup(UNIXtextfgbg);
    safecore(&UNIXtextfgbg,"\33[0;0m");
    break;
  }
}

void setansicolors() {
char *s;
  if(ansimode) {
    ssprintf(sysv,"\033[0;%d;%dm",textfgcolor,textbgcolor);
    freeup(UNIXtextfgbg);
    safecore(&UNIXtextfgbg,sysv);
    ssprintf(sysv,"\033[0;%d;%dm",menufgcolor,menubgcolor);
    freeup(UNIXmenufgbg);
    safecore(&UNIXmenufgbg,sysv);
    UNIXdefaultfgbg="\033[0m";
  } else {
    UNIXtextfgbg=UNIXmenufgbg=UNIXdefaultfgbg="";
  }
}
#endif

/*
 * SETUP
 *
 */
#if TURBOC
void setup(){
  equipment();     /* find color/mono, set whole screen to black! */
  ctrlbreak(0);           /* save state, set no ctrl-break */
  nearmemorybaseinit();  /* Get near memory base, just once */
}
#endif

#if BSDunix || SYS5
void setup(){termSetup();}
#endif

/*
 * CLEARS
 *
 * Purpose:
 *              Screen clear and cursor home.
 * Notes:
 *              o An erase page takes time.
 *              o We must clear the background on the VT52
 *                to insure highlighted areas are erased.
 */
#if BSDunix || SYS5
void cleardisplay() {
  clearlines(0);
}
void clearlines(n) int n; {
int i;
int flag;
  fflush(stderr);
  setcolorsUNIX(1);
  if(cd_termcap[0]){
    putcursor(n,0);
    etputs(cd_termcap,1);
  }
  flag = (n < -1 || cd_termcap[0]==0);
  if(n<-1) n= -1;
  if(flag){ /* must erase lines one by one */
    for(i=0;i<=ROWMAX;++i){putcursor(n+i,0); eraseol();}
    putcursor(n,0);
  }
}
void clearscreen() {
  clearlines(-1);
}
#endif

/*
 * CRLF
 *
 * Purpose:
 *              Issue carriage return and linefeed.
 * Notes:
 *              o Some terminals have auto-linefeed switches.
 *                If you can't turn it off in hardware, try
 *                to fix it here.
 */
void crlf(){
  etype("\r\n");
}

/*
 * LINEDE
 *
 * Purpose:
 *              Delete line at current cursor position.
 * Notes:
 *              o Needs long delay with some terminals to insure
 *                delete command gets completed.
 */
#if 0           /* This is the prototype for old terminals */
void linedelete(){
  etype(dl_termcap);
}
#endif
#if MSDOS
void linedelete(){
  if(row == ROWMAX) { putcursor(row,0); eraseol(); }
  else dellin();
}
#endif
#if BSDunix || SYS5
void linedelete(){
int i;
extern char *toprint();
  if(row==ROWMAX) goto quit;
  if(HASDELETELINE) etputs(dl_termcap,1);
  else {
    if(row == 0 && HASSCROLL == 0) {
      fwdindex(); statblank = TRUE; lcount();
    }
    else {
      if(HASSCROLL) {
        setscroll(1+row);
        putcursor(ROWMAX,0);
        etputs(sf_termcap,ROWMAX-row);
        setscroll(0);
        curses();
      } else {
        for(i=row;i<=ROWMAX;++i) {
          putcursor(i,0); eraseol();
          if(i+1<=ROWMAX) ePrint(toprint(line[i+1]));
        }
      }
    }
  }
quit:
  putcursor(ROWMAX,0); eraseol();
  putcursor(row,col);
}

/*
 * ERASEO
 *
 * Purpose:
 *              Erase to end of line from current cursor position.
 */
void eraseol(){
  etputs(ce_termcap,1);
}

/*
 * KILLST
 *
 * Purpose:
 *             Erase the STATUS line
 */
void killstatus(){
  statblank = TRUE; putcursor(-1,0);
  if(syscolor!=1) setcolorsUNIX(1);
eraseol();
}
#endif

/*
 * BELL
 *
 * Purpose:

 *             Ring terminal bell.
 */
#if TURBOC
void bell(){
/*  etype("\7"); */
 if(!nodisp) shortclick(135,15,10000);
}
#endif

#if BSDunix || SYS5
void bell(){
  etputs(bl_termcap,1);
}

/*
 * PUTCUR
 *
 * Purpose:
 *              Place cursor on specified row and column.
 * Notes:
 *              No error check for on-screen.
 */
#if 0           /* prototype for vt52 terminal */
void putcursor(rrow,ccol) int rrow,ccol; {
char s[5]={'\033','Y','0','0',0};
  /* VT52 style custom cursor */
  s[2]=(rrow+' '+1);
  s[3]=(ccol+' ');
  etype(s);
}
#endif

void putcursor(rrow,ccol) int rrow,ccol; {
extern char *tgoto();
  etputs(tgoto(cm_termcap,ccol,rrow+1),1);
}

/*
 * HOME
 *
 * Purpose:
 *              Put cursor at home position (0,0).
 * Notes:
 *              This is the physical home position,
 *              which may not be the first text line.
 */
void home(){
  putcursor(-1,0);
}
#endif

/*
 * LINEIN
 *
 * Purpose:
 *              Insert a line with hardware command.
 */
#if MSDOS
void lineinsert(){
  if(row == ROWMAX) {
    putcursor(row,0); eraseol();
  } else
  inslin();
}
#endif

#if 0   /* VT52 prototype */
void lineinsert(){
int i;

  if(row == 0) {
    revindex();
  } else
  if(row == ROWMAX) {
    putcursor(row,0); eraseol();
  }
  else {
    etype(al_termcap);
    putcursor(row,0);
  }
}
#endif

#if BSDunix || SYS5
void lineinsert(){
int i;
extern char *toprint();

  putcursor(ROWMAX,0); eraseol();
  putcursor(row,0);
  if(row == 0) {
    revindex();
  } else {
    if(HASINSERTLINE) {
      etputs(al_termcap,1); eraseol();
    } else {
      if(HASSCROLL) {
        setscroll(1+row);
        putcursor(row,0);
        etputs(sr_termcap,ROWMAX-row);
        setscroll(0);
      } else {
        putcursor(i = row,0); eraseol();
        for(i = row+1;i<=ROWMAX;++i) {
              putcursor(i,0);
              eraseol();
              ePrint(toprint(line[i-1]));
        }
      }
    }
  }
  putcursor(row,0);
}

/*
 * SETSCR
 *
 * Purpose:
 *              Set scroll region for a VT100 terminal
 *              From row=r to ROWMAX
 *              Uses low level termcap routines (source included)
 */
void setscroll(r) int r; {
char buf[128];
extern char *tgoto();
  etputs(tgoto(cs_termcap,ROWMAX+1,r),1);
}
#endif

/*
 * REVIND
 *
 * Purpose:
 *              Performs a hardware reverse index to put a blank line
 *              on the top of the screen. Causes all current lines on
 *              screen to be shifted down one, with the last line lost.
 *              The status line is reprinted.
 */
#if MSDOS
void revindex(){
  putcursor(0,0); inslin();
  lcount();
}
#endif
#if BSDunix || SYS5
void revindex(){
int i;
extern char *toprint();

   putcursor(ROWMAX,0); eraseol();
  if(HASINSERTLINE) {
    putcursor(0,0);
    etputs(al_termcap,1); eraseol();
  } else {
    if(HASSCROLL){
       setscroll(1); putcursor(0,0);
       etputs(sr_termcap,ROWMAX-1); setscroll(0);
    } else {
      clearscreen(); putcursor(0,0);
      for(i=0;i<ROWMAX;++i) {
        crlf(); ePrint(toprint(line[i]));
      }
    }
  }
  lcount();
}

/*
 * FWDIND
 *
 * Purpose:
 *
 *              Performs a hardware forward scroll. The top line is
 *              lost and a new blank line is issued at ROWMAX.
 */
void fwdindex(){
  if(HASDELETELINE) {
    putcursor(0,0); etputs(dl_termcap,1);
    putcursor(ROWMAX,0); eraseol();
  } else {
    if(HASSCROLL){
      setscroll(1); putcursor(ROWMAX,0);
      etputs(sf_termcap,ROWMAX-1); setscroll(0);
    } else {
     killstatus(); putcursor(ROWMAX,0); crlf();
    }
  }
}

/*
 * INVVID
 *
 * Purpose:
 *              Output a byte in inverse video. This may be
 *              implemented as background or foreground on certain
 *              terminals.
 * Notes:

 *              o We assume 0 <= x < 32 so that x+'@' is printable.
 */
void invvideo(x) int x; {
char *p;
char buf[256]; char *endof();
  strcpy(buf,REVVIDEO); p=endof(buf); *p++ = x + '@';
  if(UNIXtextfgbg[0]==0) strcpy(p,NORMVIDEO);
  else
  strcat(buf,UNIXtextfgbg);
  etype(buf);
}

/*
 * ERACHA
 *
 * Purpose:
 *              Erase character left of cursor
 *
 * Notes:
 *              o Usually, "\b\40\b" will work.
 *                Termcap uses BC for the backspace char.
 */
void erachar(){
  etype(ERACHR);
}
#endif

/*
 * BANNER
 *
 * Purpose:
 *              Print a banner on the STATUS line, start col=0.
 * Notes:
 *              Erases the STATUS line first.
 *              Types the message s[] starting at the first
 *              column of the STATUS line.
 *              DOES NOT restore the cursor. Use curses().
 */
void banner(s) char *s; {
  killstatus(); ePrint(s);
}

void bbanner(s1,s2) char *s1,*s2; {
 banner(s1); ePrint(s2);
}

/*
 * NOROOM
 *
 * Purpose:
 *              Print NO ROOM message on STATUS line.
 * Notes:
 *              Restores cursor.
 *              Returns FALSE.
 */
int noroom() {
  banner("Out of room"); bell(); curses();
  return FALSE;
}

/*
 * CURSES
 *
 * Purpose:
 *             Restore cursor to its saved position.
 * Notes:
 *             Uses global variables row & column.
 */
void curses() {
  putcursor(row,col);
}
