#if 0
***************************************************************************
**                                                                       **
** VEE.C         Expression Evaluator                                    **
**                                                                       **
**                                                                       **
** AUTHOR:      Mark Morley                                              **
** COPYRIGHT:   (c) 1992 by Mark Morley                                  **
** HISTORY:     December 1991 original code.                             **
**              Jun 1992 - Updated comments and docs                     **
**              Jun 1995 - Cut out code for use in editor appl.          **
**                         Changed macros to functions.                  **
**                                                                       **
** You are free to incorporate this code into your own works, even if it **
** is a commercial application.  However, you may not charge anyone else **
** for the use of this code!  If you intend to distribute your code,     **
** I'd appreciate it if you left this message intact.  I'd like to       **
** receive credit wherever it is appropriate.  Thanks!                   **
**                                                                       **
** Please mail any bug reports/fixes/enhancments to me at:               **
**      morley@camosun.bc.ca                                             **
** or                                                                    **
**      Mark Morley                                                      **
**      3889 Mildred Street                                              **
**      Victoria, BC  Canada                                             **
**      V8Z 7G1                                                          **
**      (604) 479-7861                                                   **
**                                                                       **
**************************************************************************
#endif
#include <setjmp.h>

#define TYPE            double          /* Type of numbers to work with */

#define VARLEN          15              /* Max length of variable names */
#define MAXVARS         50              /* Max user-defined variables */
#define TOKLEN          30              /* Max token length */

#define VAR             1
#define DEL             2
#define NUM             3

typedef struct
{
   char name[VARLEN + 1];               /* Variable name */
   TYPE value;                          /* Variable value */
} VARIABLE;

typedef struct
{
   char* name;                          /* Function name */
   int   args;                          /* Number of arguments to expect */
   TYPE  (*func)();                     /* Pointer to function */
} FUNCTION;

#if 0
/* The following macros are ASCII dependant, no EBCDIC here! */
#define iswhite(c)  (c == ' ' || c == '\t')
#define isnumer(c)  ((c >= '0' && c <= '9') || c == '.')
#define isalpha(c)  ((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') \
                    || c == '_')
#define isdelim(c)  (c == '+' || c == '-' || c == '*' || c == '/' || c == '%' \
                    || c == '^' || c == '(' || c == ')' || c == ',' || c == '=')

#else
static iswhite(c) int c; {
 return(c == ' ' || c == '\t');
}

static isnumer(c) int c;{
  return((c >= '0' && c <= '9') || c == '.');
}

static isalpha(c) int c;{
  return((c >= 'a' && c <= 'z') || (c >= '0' && c <= '9')
                    || c == '_');
}

static isdelim(c) int c;{
char *strchr();
  return(strchr("+-*/%^(),=",c)!=(char *)0);
}
#endif

/* Codes returned from the evaluator */
#define E_OK           0        /* Successful evaluation */
#define E_SYNTAX       1        /* Syntax error */
#define E_UNBALAN      2        /* Unbalanced parenthesis */
#define E_DIVZERO      3        /* Attempted division by zero */
#define E_UNKNOWN      4        /* Reference to unknown variable */
#define E_MAXVARS      5        /* Maximum variables exceeded */
#define E_BADFUNC      6        /* Unrecognized function */
#define E_NUMARGS      7        /* Wrong number of arguments to function
*/
#define E_NOARG        8        /* Missing an argument to a function */
#define E_EMPTY        9        /* Empty expression */

/* These defines only happen if the values are not already defined!  You may
   want to add more precision here, if your machine supports it. */
#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif
#ifndef M_E
#define M_E     2.71828182845904523536
#endif

extern double deg();
extern double rad();
extern double oct();         /* Octal to base ten */
extern double bin();         /* Binary to base ten */

int   ERROR;                 /* The error number */
char  ERTOK[TOKLEN + 1];     /* The token that generated the error */
int   ERPOS;                 /* The offset from the start of the expression */
char* ERANC;                 /* Used to calculate ERPOS */

/*
   Add any "constants" here...  These are "read-only" values that are
   provided as a convienence to the user.  Their values can not be
   permanently changed.  The first field is the variable name, the second
   is its value.
*/
VARIABLE Consts[] =
{
   /* name, value */
   { "pi",      M_PI },
   { "e",       M_E },
   { 0 }
};

/*
   Add any math functions that you wish to recogniz here...  The first
   field is the name of the function as it would appear in an expression.
   The second field tells how many arguments to expect.  The third is
   a pointer to the actual function to use.
*/
extern double   acos(/* double x */);
extern double   asin(/* double x */);
extern double   atan(/* double x */);
extern double   atan2(/* double y, double x */);
extern double   ceil(/* double x */);
extern double   cos(/* double x */);
extern double   cosh(/* double x */);
extern double   exp(/* double x */);
extern double   fabs(/* double x */);
extern double   floor(/* double x */);
extern double   fmod(/* double x, double y */);
extern double   frexp(/* double value, int *exp */);
extern double   ldexp(/* double value, int exp */);
extern double   log(/* double x */);
extern double   log10(/* double x */);
extern double   modf(/* double value, double *iptr */);
extern double   pow(/* double x, double y */);
extern double   sin(/* double x */);
extern double   sinh(/* double x */);
extern double   sqrt(/* double x */);
extern double   tan(/* double x */);
extern double   tanh(/* double x */);
extern double   hypot();
extern double   acosh();
extern double   asinh();
extern double   atanh();
FUNCTION Funcs[] =
{
   /* name, function to call */
#if 1           /* These functions add 1.9k to the EXE file */
   { "sin",     1,    sin },
   { "cos",     1,    cos },
   { "tan",     1,    tan },
   { "asin",    1,    asin },
   { "acos",    1,    acos },
   { "atan",    1,    atan },
   { "sinh",    1,    sinh },
   { "cosh",    1,    cosh },
   { "tanh",    1,    tanh },
   { "exp",     1,    exp },
   { "ln",      1,    log },
   { "log",     1,    log },
   { "log10",   1,    log10 },
   { "sqrt",    1,    sqrt },
   { "floor",   1,    floor },
   { "ceil",    1,    ceil },
   { "hypot",   2,    hypot },
#endif
   { "abs",     1,    fabs },
   { "fmod",    2,    fmod },
   { "pow",     2,    pow },
   { "deg",     1,    deg },
   { "rad",     1,    rad },
   { "oct",     1,    oct },
   { "bin",     1,    bin },
   { "",        1,    fabs }
};

VARIABLE        Vars[MAXVARS];       /* Array for user-defined variables */
 char*  expression;          /* Pointer to the user's expression */
 char   token[TOKLEN + 1];   /* Holds the current token */
int             type;                /* Type of the current token */
jmp_buf         jb;                  /* jmp_buf for errors */


#if 0
#define ERR(n) {ERROR=n; ERPOS=expression-ERANC-1; strcpy(ERTOK,token);
longjmp(jb,1);}
#else  /* Using a function instead */
/* W A R N I N G: This function exits to processor state jb */
void ERR(n) int n; {
extern int ERROR,ERPOS;
extern char *ERANC;
extern char ERTOK[]; extern char token[];
extern char *expression;
extern jmp_buf jb;

  ERROR=n; ERPOS=(int)(expression-ERANC)-1;
  strcpy(ERTOK,token); longjmp(jb,1);
}
#endif


double
deg(x) double x;
{
   return( x * 180.0 / M_PI );
}

double
rad(x) double x;
{
   return( x * M_PI / 180.0 );
}

double
cvbase(x,b) double x; int b;  /* Base b to base ten */
{ /* x==1001101 should return (1)*b^6 + (1)*b^3 + (1)*b^2 + (1)*b^0 */
char s[256];
int i,j,k;
char flg;
  if(x<0){x = -x; flg=1;} else flg=0;
  sprintf(s,"%.0f",x);
  k=strlen(s); x=0;
  for(i=0;i<k;++i){
   j=s[i]-'0'; if(j<0 || j>= b) ERR(E_SYNTAX);
   x=b*x+j;
  }
  return (flg? -x:x);
}

double
oct(x) double x;       /* Octal to base ten */
{ /* x==314 should return oct==(3)*8^2+(1)*8+(4)*1 */
  return cvbase(x,8);
}

double
bin(x) double x;{       /* Binary to base ten */
/* x==1001101 should return bin==(1)*2^6 + (1)*2^3 + (1)*2^2 + (1)*2^0 */
  return cvbase(x,2);
}



/*************************************************************************
**                                                                       **
** GetSymbol( char* s )                                                  **
**                                                                       **
** This routine obtains a value from the program's environment.          **
** This works for DOS and unix.                                          **
**                                                                       **
 ************************************************************************/

GetSymbol(s,v) char* s; TYPE* v;
{
   char* e;
   extern char *getenv(); extern double atof();

   e = getenv( s );
   if( !e )
      return( 0 );
   *v = atof( e );
   return( 1 );
}


/*************************************************************************
**                                                                       **
** ClearAllVars()                                                        **
**                                                                       **
** Erases all user-defined variables from memory. Note that constants    **
** can not be erased or modified in any way by the user.                 **
**                                                                       **
** Returns nothing.                                                      **
**                                                                       **
 *************************************************************************/

void ClearAllVars()
{
   int i;

   for( i = 0; i < MAXVARS; i++ )
   {
      *Vars[i].name = 0;
      Vars[i].value = 0;
   }
}


/*************************************************************************
**                                                                       **
** ClearVar( char* name )                                                **
**                                                                       **
** Erases the user-defined variable that is called NAME from memory.     **
** Note that constants are not affected.                                 **
**                                                                       **
** Returns 1 if the variable was found and erased, or 0 if it didn't     **
** exist.                                                                **
**                                                                       **
 *************************************************************************/

ClearVar(name) char* name;
{
   int i;

   for( i = 0; i < MAXVARS; i++ )
      if( *Vars[i].name && ! strcmp( name, Vars[i].name ) )
      {
         *Vars[i].name = 0;
         Vars[i].value = 0;
         return( 1 );
      }
   return( 0 );
}


/*************************************************************************
**                                                                       **
** GetValue(name,value) char* name; TYPE* value;
**                                                                       **
** Looks up the specified variable (or constant) known as NAME and       **
** returns its contents in VALUE.                                        **
**                                                                       **
** First the user-defined variables are searched, then the constants are **
** searched.                                                             **
**                                                                       **
** Returns 1 if the value was found, or 0 if it wasn't.                  **
**                                                                       **
 *************************************************************************/

GetValue(name,value) char* name; TYPE* value;
{
   int i;

   /* First check for an environment variable reference... */
   if( *name == '_' )
      return( GetSymbol( name + 1, value ) );

   /* Now check the user-defined variables. */
   for( i = 0; i < MAXVARS; i++ )
      if( *Vars[i].name && ! strcmp( name, Vars[i].name ) )
      {
         *value = Vars[i].value;
         return( 1 );
      }

   /* Now check the programmer-defined constants. */
   for( i = 0; *Consts[i].name; i++ )
      if( *Consts[i].name && ! strcmp( name, Consts[i].name ) )
      {
         *value = Consts[i].value;
         return( 1 );
      }
   return( 0 );
}


/*************************************************************************
**                                                                       **
** SetValue( char* name, TYPE* value )                                   **
**                                                                       **
** First, it erases any user-defined variable that is called NAME.  Then **
** it creates a new variable called NAME and gives it the value VALUE.   **
**                                                                       **
** Returns 1 if the value was added, or 0 if there was no more room.     **
**                                                                       **
 *************************************************************************/

SetValue(name,value) char* name; TYPE* value;
{
   char b[30];
   int  i;

   ClearVar( name );
   for( i = 0; i < MAXVARS; i++ )
      if( ! *Vars[i].name )
      {
         strcpy( Vars[i].name, name );
         Vars[i].name[VARLEN] = 0;
         Vars[i].value = *value;
         return( 1 );
      }
   return( 0 );
}


/*************************************************************************
**                                                                       **
** Parse()   Internal use only                                           **
**                                                                       **
** This function is used to grab the next token from the expression that **
** is being evaluated.                                                   **
**                                                                       **
 *************************************************************************/


void Parse()
{
   char* t;

   type = 0;
   t = token;
   while( iswhite( *expression ) )
      expression++;
   if( isdelim( *expression ) )
   {
      type = DEL;
      *t++ = *expression++;
   }
   else if( isnumer( *expression ) )
   {
      type = NUM;
      while( isnumer( *expression ) )
         *t++ = *expression++;
      if(*expression == 'e' || *expression == 'd'){ /* sci notation */
        if(*expression == 'd'){ *t++ = 'e'; ++expression;} else
        *t++ = *expression++;
        if(isnumer(*expression) || *expression == '-' || *expression ==
        '+') *t++ = *expression++;
        while( isnumer(*expression)) *t++ = *expression++;
      }
   }
   else if( isalpha( *expression ) )
   {
      type = VAR;
      while( isalpha( *expression ) )
        *t++ = *expression++;
      token[VARLEN] = 0;
   }
   else if( *expression )
   {
      *t++ = *expression++;
      *t = 0;
      ERR( E_SYNTAX );
   }
   *t = 0;
   while( iswhite( *expression ) )
      expression++;
}


/*************************************************************************
**                                                                       **
** Level1( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any variable assignment operations.             **
** It returns a value of 1 if it is a top-level assignment operation,    **
** otherwise it returns 0                                                **
**                                                                       **
 *************************************************************************/


extern void Level2(),Level3(),Level4(),Level5(),Level6();

Level1(r) TYPE* r;
{
   char t[VARLEN + 1];

   if( type == VAR )
      if( *expression == '=' )
      {
         strcpy( t, token );
         Parse();
         Parse();
         if( !*token )
         {
            ClearVar( t );
            return(1);
         }
         Level2( r );
         if( ! SetValue( t, r ) )
            ERR( E_MAXVARS );
         return( 1 );
      }
   Level2( r );
   return( 0 );
}


/*************************************************************************
**                                                                       **
** Level2( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any addition and subtraction operations.        **
**                                                                       **
 *************************************************************************/


void Level2(r) TYPE* r;
{
   TYPE t = 0;
   char o;

   Level3( r );
   while( (o = *token) == '+' || o == '-' )
   {
      Parse();
      Level3( &t );
      if( o == '+' )
         *r = *r + t;
      else if( o == '-' )
         *r = *r - t;
   }
}


/*************************************************************************
**                                                                       **
** Level3( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any multiplication, division, or modulo.        **
**                                                                       **
 *************************************************************************/


void Level3(r) TYPE* r;
{
   TYPE t;
   char o;

   Level4( r );
   while( (o = *token) == '*' || o == '/' || o == '%' )
   {
      Parse();
      Level4( &t );
      if( o == '*' )
         *r = *r * t;
      else if( o == '/' )
      {
         if( t == 0 )
            ERR( E_DIVZERO );
         *r = *r / t;
      }
      else if( o == '%' )
      {
         if( t == 0 )
            ERR( E_DIVZERO );
         *r = fmod( *r, t );
      }
   }
}


/*************************************************************************
**                                                                       **
** Level4( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any "to the power of" operations.               **
**                                                                       **
 *************************************************************************/


void Level4(r) TYPE* r;
{
   TYPE t;

   Level5( r );
   if( *token == '^' )
   {
      Parse();
      Level5( &t );
      *r = pow( *r, t );
   }
}


/*************************************************************************
**                                                                       **
** Level5( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any unary + or - signs.                         **
**                                                                       **
 *************************************************************************/


void Level5(r) TYPE* r;
{
   char o = 0;

   if( *token == '+' || *token == '-' )
   {
      o = *token;
      Parse();
   }
   Level6( r );
   if( o == '-' )
      *r = -*r;
}


/*************************************************************************
**                                                                       **
** Level6( TYPE* r )   Internal use only                                 **
**                                                                       **
** This function handles any literal numbers, variables, or functions.   **
**                                                                       **
 *************************************************************************/


void Level6(r) TYPE* r;
{
   int  i;
   int  n;
   TYPE a[3];
   extern double atof();

   if( *token == '(' ) {
      Parse();
      if( *token == ')' ) ERR( E_NOARG );
      Level1( r );
      if( *token != ')' ) ERR( E_UNBALAN );
      Parse();
   }
   else
   {
      if( type == NUM ) {
         *r = (TYPE) atof( token );
         Parse();
      }
      else if( type == VAR ) {
         if( *expression == '(' ) {
            for( i = 0; *(Funcs[i].name)!=0; i++ ){
               if( ! strcmp( token, Funcs[i].name ) ) {
                  Parse();
                  n = 0;
                  do {
                     Parse();
                     if( *token == ')' || *token == ',' )
                        ERR( E_NOARG );
                     a[n] = 0;
                     Level1( &a[n] );
                     n++;
                  } while( n < 4 && *token == ',' );
                  Parse();
                  if( n != Funcs[i].args ) {
                     strcpy( token, Funcs[i].name );
                     ERR( E_NUMARGS );
                  }
                  *r = Funcs[i].func( a[0], a[1], a[2] );
                  return;
               }
            }
            if( ! *Funcs[i].name ) ERR( E_BADFUNC );
         } else
         if( ! GetValue( token, r ) ) ERR( E_UNKNOWN );
         Parse();
      } else ERR( E_SYNTAX );
   }
}

void cvlower(s) char *s; {
int x;
  while((x=s[0])!=0) *s++ = tolower(x);
}


/*************************************************************************
**                                                                       **
** Evaluate( char* e, TYPE* result, int* a )                             **
**                                                                       **
** This function is called to evaluate the expression E and return the   **
** answer in RESULT.  If the expression was a top-level assignment, a    **
** value of 1 will be returned in A, otherwise it will contain 0.        **
**                                                                       **
** Returns E_OK if the expression is valid, or an error code.            **
**                                                                       **
 *************************************************************************/

Evaluate(e,result,a) char* e; TYPE* result; int* a;
{
   if( setjmp( jb ) )
      return( ERROR );
   expression = e;
   ERANC = e;
   cvlower( expression );
   *result = 0;
   Parse();
   if( ! *token )
      ERR( E_EMPTY );
   *a = Level1( result );
   return( E_OK );
}

char* ErrMsgs[] =
{
   "Syntax error",
   "Unbalanced parenthesis",
   "Division by zero",
   "Unknown variable",
   "Maximum variables exceeded",
   "Unrecognized function",
   "Wrong number of arguments to function",
   "Missing an argument",
   "Empty expression"
};

evalexpr(src,errmsg,ans)  char *src,*errmsg; double *ans;
{
   TYPE  result;
   int   ec;
   int   a;
   static int flag=1;
   char f;

   if(flag) {   /* First time, then setup vars */
     ClearAllVars(); flag=0;
     result=0; SetValue("ans", &result );
   }
   ec=Evaluate( src, &result, &a );
   if(ec != E_OK && ec != E_EMPTY ){
   /* Display error info.  An E_EMPTY error is ignored. */
   sprintf(errmsg,"EvalExpr: %s : %s [%s]",ErrMsgs[ERROR-1],ERTOK,ERANC);
   f=2; goto quit;
   }
   /* If we didn't assign a variable, then save the result. */
   if( ec == E_OK  && ! a) {
     SetValue("ans", &result );
     ans[0]=result;  f=0;
   } else {ans[0]=errmsg[0]=0; f=1;}
quit:
   return f;
}

#if 0
#undef iswhite
#undef isnumer
#undef isalpha
#undef isdelim
#endif
